      
// #ifndef __FUNCTION_H__
// #define __FUNCTION_H__

// #include <algorithm>
// #include <iostream>
// #include <map>
// #include <set>
// #include <vector>
// #include "AsmBuilder.h"
// #include "Ast.h"
// #include "BasicBlock.h"
// #include "SymbolTable.h"

// class Unit;

// class Function
// {
//     typedef std::vector<BasicBlock *>::iterator iterator;
//     typedef std::vector<BasicBlock *>::reverse_iterator reverse_iterator;

// private:
//     std::vector<BasicBlock *> block_list;
//     std::map<BasicBlock*, MachineBlock*> map2;
//     SymbolEntry *sym_ptr;
//     BasicBlock *entry;
//     Unit *parent;

// public:
//     Function(Unit *, SymbolEntry *);
//     ~Function();
//     void insertBlock(BasicBlock *bb) { block_list.push_back(bb); };
//     BasicBlock *getEntry() { return entry; };
//     void remove(BasicBlock *bb);
//     void output() const;
//     std::vector<BasicBlock *> &getBlockList(){return block_list;};
//     iterator begin() { return block_list.begin(); };
//     iterator end() { return block_list.end(); };
//     reverse_iterator rbegin() { return block_list.rbegin(); };
//     reverse_iterator rend() { return block_list.rend(); };
//     SymbolEntry *getSymPtr() { return sym_ptr; };
//     void genMachineCode(AsmBuilder*);
//     //将所有的block块置为false
//     void clear_all_vis() {
//         for (auto it = block_list.begin(); it != block_list.end(); it++) (*it)->setvisit(false);
//     };
//     MachineBlock* getMachineBlock(BasicBlock* b){ return map2[b];};
// };

// #endif

#ifndef __FUNCTION_H__
#define __FUNCTION_H__

#include <algorithm>
#include <iostream>
#include <map>
#include <set>
#include <vector>
#include "AsmBuilder.h"
#include "Ast.h"
#include "BasicBlock.h"
#include "SymbolTable.h"

class Unit;

struct loop_item{
    BasicBlock* cond_bb=nullptr;
    BasicBlock* while_bb=nullptr;
    BasicBlock* end_bb=nullptr;
    BasicBlock* next=nullptr;
    loop_item(){}
	loop_item(BasicBlock* bb1, BasicBlock* bb2, BasicBlock* bb3) 
    :cond_bb(bb1), while_bb(bb2), end_bb(bb3){}
    loop_item(BasicBlock* bb1, BasicBlock* bb2, BasicBlock* bb3,BasicBlock* bb4) 
    :cond_bb(bb1), while_bb(bb2), end_bb(bb3),next(bb4){}
};

class Function
{
    typedef std::vector<BasicBlock *>::iterator iterator;
    typedef std::vector<loop_item *>::iterator loop_iterator;
    typedef std::vector<BasicBlock *>::reverse_iterator reverse_iterator;
    typedef std::vector<loop_item *>::reverse_iterator loop_reverse_iterator;

private:
    std::vector<loop_item*> loop_list;
    std::vector<BasicBlock *> block_list;
    std::map<BasicBlock*, MachineBlock*> map2;
    SymbolEntry *sym_ptr;
    BasicBlock *entry;
    Unit *parent;
    bool while_with_if_f = false;
    bool while_with_if_and_specialassign = false;
    bool csseflag = false;

public:

    Function(Unit *, SymbolEntry *);
    ~Function();
    void insertBlock(BasicBlock *bb) { block_list.push_back(bb); };
    BasicBlock *getEntry() { return entry; };
    std::string getFuncName(){return dynamic_cast<IdentifierSymbolEntry*>(sym_ptr)->getName();}
    void remove(BasicBlock *bb);
    void output() const;
    std::vector<loop_item *> &getLoopList(){return loop_list;};
    std::vector<BasicBlock *> &getBlockList(){return block_list;};
    iterator begin() { return block_list.begin(); };
    loop_iterator loop_begin() { return loop_list.begin(); };
    iterator end() { return block_list.end(); };
    loop_iterator loop_end() { return loop_list.end(); };
    void push_loop(loop_item* loop) { loop_list.push_back(loop); };
    reverse_iterator rbegin() { return block_list.rbegin(); };
    reverse_iterator rend() { return block_list.rend(); };
    SymbolEntry *getSymPtr() { return sym_ptr; };
    void genMachineCode(AsmBuilder*);
    //将所有的block块置为false
    void clear_all_vis() {
        for (auto it = block_list.begin(); it != block_list.end(); it++) (*it)->setvisit(false);
    };
    MachineBlock* getMachineBlock(BasicBlock* b){ return map2[b];};

    void removePred(Instruction* in);
    void addPred(Instruction* in);
    BasicBlock* getMarkBranch(BasicBlock* block);
    int essential = -1;
    int getEssential();
    std::map<Function*, std::vector<Instruction*>> preds;
    std::map<Function*, std::vector<Instruction*>>& getPreds() {
        return preds;
    };
    void setEntry(BasicBlock* block) { entry = block;}

    // for DCE
    bool clearUselessBB();

    // for SSA Destruction
    void setwhile_with_if_f(){while_with_if_f = true;}
    bool getwhile_with_if_f(){return while_with_if_f;}
    // while_with_if_and_specialassign
    void setwhile_with_if_and_specialassign(){while_with_if_and_specialassign = true;}
    bool getwhile_with_if_and_specialassign(){return while_with_if_and_specialassign;}

    // CSSE
    void setcsseflag(){csseflag = true;}
    bool getcsseflag(){return csseflag;}
};

#endif

    