#ifndef __UNIT_H__
#define __UNIT_H__

#include <vector>
#include "Function.h"
#include "SymbolTable.h"

class Unit {
    typedef std::vector<Function*>::iterator iterator;
    typedef std::vector<Function*>::reverse_iterator reverse_iterator;

   private:
    std::vector<SymbolEntry*> global_list;
    std::vector<Function*> func_list;
    std::vector<SymbolEntry*> declare_list;
    // wzr-multi:
    // 在MachineUnit记录将要被多线程化的函数表项及其栈偏移
    // 函数表项是在multithread.pass()时获取的
    // 栈偏移是在mUnit.genMachineCode()时根据“函数表项比对”获取的
    // 栈偏移最终在mUnit.output()的结尾被使用
    int target_func_unit_offset_unit = 0;
    IdentifierSymbolEntry* target_func_unit = nullptr;


   public:
    Unit() = default;
    ~Unit();
    void insertFunc(Function*);
    void removeFunc(Function*);
    void insertGlobal(SymbolEntry*);
    void removeGlobal(SymbolEntry*);
    void insertDeclare(SymbolEntry*);
    void removeDeclare(SymbolEntry*);
    void output() const;
    iterator begin() { return func_list.begin(); };
    iterator end() { return func_list.end(); };
    reverse_iterator rbegin() { return func_list.rbegin(); };
    reverse_iterator rend() { return func_list.rend(); };
    void genMachineCode(MachineUnit* munit);
    std::vector<Function* > &get_functions() {return func_list;};
    std::vector<SymbolEntry* > &get_globals() {return global_list;};
    void set_target_func_unit_offset_unit(int off){target_func_unit_offset_unit = off;}
    int get_target_func_unit_offset_unit(){return target_func_unit_offset_unit;};
    void set_target_func_unit(IdentifierSymbolEntry* tfunc){target_func_unit = tfunc;};
    IdentifierSymbolEntry* get_target_func_unit(){return target_func_unit;};

};

#endif
