#include "BasicBlock.h"
#include "Function.h"
#include <algorithm>
#include <vector>

extern FILE *yyout;

BasicBlock* BasicBlock::copybb1(Function* func)
{
    BasicBlock* new_bb = new BasicBlock(func);
    for(Instruction* ins = this->begin(); ins != this->end(); ins = ins->getNext())
    {
        new_bb->insertBack(ins->copy1());
    }
    return new_bb;
}


int BasicBlock::getInstSum(){
    int sum = 0;
    for (auto i = head->getNext(); i != head; i = i->getNext())
    sum++;

    return sum;
}

Instruction* BasicBlock::getBranchBB() {
    for (auto instr = head->getPrev(); instr != head;
         instr = instr->getPrev()) {
        if (instr->isCond())
            return instr;
    }
    return nullptr;
}

bool BasicBlock::existsInPred(BasicBlock *bb) {
    return std::find(pred.begin(), pred.end(), bb) != pred.end();
}

bool BasicBlock::existsInSucc(BasicBlock *bb) {
    return std::find(succ.begin(), succ.end(), bb) != succ.end();
}
void BasicBlock::cleanPred() {
    for (auto i : pred)
        i->removeSucc(this);
    std::vector<BasicBlock *>().swap(pred);
}
void BasicBlock::cleanSucc() {
    for (auto i : succ)
        i->removePred(this);
    std::vector<BasicBlock *>().swap(succ);
}

void BasicBlock::deleteAllRelation() {
    // std::cout<<"== deleteAllRelation "<<getNo()<<"\n";

    for (auto pred_bb : pred) {
        // std::cout<<"pred = "<<pred_bb->getNo()<<"\n";
        auto pred_bb_succ = pred_bb->getSuccBlock();
        if (std::find(pred_bb_succ.begin(), pred_bb_succ.end(), this) !=
            pred_bb_succ.end())
            pred_bb->removeSucc(this);
        // std::cout<<"pred finish \n";
    }
    for (auto succ_bb : succ) {
        // std::cout<<"succ = "<<succ_bb->getNo()<<"\n";
        auto succ_bb_pred = succ_bb->getPredBlock();
        if (std::find(succ_bb_pred.begin(), succ_bb_pred.end(), this) !=
            succ_bb_pred.end())
            succ_bb->removePred(this);
        // std::cout<<"succ finish \n";
    }
    // pred.clear();
    // succ.clear();
}

void BasicBlock::maintainAllRelation() {
    for (auto pred_bb : pred) {
        for (auto succ_bb : succ) {
            pred_bb->addSucc(succ_bb);
        }
    }
    for (auto succ_bb : succ) {
        for (auto pred_bb : pred) {
            succ_bb->addPred(pred_bb);
        }
    }
}

void BasicBlock::deleteBack(int num) {
    while (num--) {
        remove(head->getPrev());
    }
}
// void BasicBlock::deleteAllRelation() {
//     printf("deleteAllRelation1\n");
//     for(auto pred_bb:pred){
//         auto pred_bb_succ = pred_bb->getSuccBlock();
//         if(std::find(pred_bb_succ.begin(),pred_bb_succ.end(),this)!=
//         pred_bb_succ.end()) pred_bb->removeSucc(this);
//     }
//     for(auto succ_bb:succ){
//         auto succ_bb_pred = succ_bb->getPredBlock();
//         if(std::find(succ_bb_pred.begin(),succ_bb_pred.end(),this)!=
//         succ_bb_pred.end()) succ_bb->removePred(this);
//     }
//     printf("deleteAllRelation2\n");
// }

// insert the instruction to the front of the basicblock.
void BasicBlock::insertFront(Instruction *inst) {
    insertBefore(inst, head->getNext());
}

// insert the instruction to the back of the basicblock.
void BasicBlock::insertBack(Instruction *inst) { insertBefore(inst, head); }
// insert the instruction dst after src.
void BasicBlock::insertAfter(Instruction *dst, Instruction *src) {
    // if(src->getNext()==NULL)
    // //std::cout<<"src->getNext()->setPrev(dst);\n";
    dst->setNext(src->getNext());
    src->getNext()->setPrev(dst);
    dst->setPrev(src);
    src->setNext(dst);

    dst->setParent(this);
}
// insert the instruction dst before src.
void BasicBlock::insertBefore(Instruction *dst, Instruction *src) {
    src->getPrev()->setNext(dst);
    dst->setPrev(src->getPrev());
    dst->setNext(src);
    src->setPrev(dst);

    dst->setParent(this);
}

// remove the instruction from intruction list.
void BasicBlock::remove(Instruction *inst) {
    inst->getPrev()->setNext(inst->getNext());
    inst->getNext()->setPrev(inst->getPrev());
}

void BasicBlock::output() const {
    fprintf(yyout, "B%d:", no);

    if (!pred.empty()) {
        fprintf(yyout, "%*c; preds = %%B%d", 32, '\t', pred[0]->getNo());
        for (auto i = pred.begin() + 1; i != pred.end(); i++)
            fprintf(yyout, ", %%B%d", (*i)->getNo());
    }
    fprintf(yyout, "\n");
    for (auto i = head->getNext(); i != head; i = i->getNext())
        i->output();
}

void BasicBlock::addSucc(BasicBlock *bb, bool first) { succ.push_back(bb); }

// remove the successor basicclock bb.
void BasicBlock::removeSucc(BasicBlock *bb) {
    succ.erase(std::find(succ.begin(), succ.end(), bb));
}

void BasicBlock::addPred(BasicBlock *bb) { pred.push_back(bb); }

// remove the predecessor basicblock bb.
void BasicBlock::removePred(BasicBlock *bb) {
    pred.erase(std::find(pred.begin(), pred.end(), bb));
}
void BasicBlock::genMachineCode(AsmBuilder *builder) {
    // std::cout<<"bb genmachinecode\n";
    auto cur_func = builder->getFunction();
    auto cur_block = new MachineBlock(cur_func, no);
    builder->setBlock(cur_block);
    // std::cout<<"bb breakpoint2\n";
    for (auto i = head->getNext(); i != head; i = i->getNext()) {
        // std::cout<<"ins type = "<<i->getInstName()<<"\n";
        i->genMachineCode(builder);
    }
    cur_func->InsertBlock(cur_block);
    // std::cout<<"bb genmachinecode end\n";
}

BasicBlock::BasicBlock(Function *f) {
    this->no = SymbolTable::getLabel();
    this->visit = false;
    f->insertBlock(this);
    parent = f;
    head = new DummyInstruction();
    head->setParent(this);
}

BasicBlock::~BasicBlock() {
    Instruction *inst;
    inst = head->getNext();
    while (inst != head) {
        Instruction *t;
        t = inst;
        inst = inst->getNext();
        delete t;
    }
    for (auto &bb : pred)
        bb->removeSucc(this);
    for (auto &bb : succ)
        bb->removePred(this);
    parent->remove(this);
}

void BasicBlock::addAlloc(Instruction *alloc) {
    assert(alloc->isAlloc());
    if (!lastAlloc) {
        insertFront(alloc);
        lastAlloc = alloc;
    } else {
        // std::cout<<"lastAlloc\n";
        insertAfter(alloc, lastAlloc);
        lastAlloc = alloc;
    }
}

void BasicBlock::insertPhiInstruction(Operand *dst) {
    Instruction *i = new PhiInstruction(dst);
    insertFront(i);
}
// #define DEBUG0429 1
Instruction *BasicBlock::getBrorRet() {
#ifdef DEBUG0429
std::cout<<getNo()<<"\n";
    for (Instruction *it = this->begin(); it != this->end();
         it = it->getNext()) {
        std::cout << "it = " << it << "\n";
        std::cout << "it insttype = " << it->getInstType() << "\n";
    }
    std::cout << "in getBrorRet\n";
#endif
    if (this->empty())
        return nullptr;
    else {
#ifdef DEBUG0429
        std::cout << "in else branch\n";
        std::cout << "this->end() = " << this->end() << "\n";
#endif
        Instruction *ins = this->end()->getPrev();
        unsigned int insttype = ins->getInstType();
#ifdef DEBUG0429
        std::cout << "after get insttype, insttype = " << insttype << "\n";
#endif
        // std::string insttype = this->end()->getInstName();
        if (insttype == Instruction::RET) {
#ifdef DEBUG0429
            std::cout << "in ret branch ins = " << ins << "\n";
#endif
            return ins;
        } else if (insttype == Instruction::COND) {
#ifdef DEBUG0429
            std::cout << "in cond branch ins = " << ins << "\n";
#endif
            return ins;
        } else if (insttype == Instruction::UNCOND) {
#ifdef DEBUG0429
            std::cout << "in uncond branch  ins = " << ins << "\n";
#endif
            return ins;
        }
#ifdef DEBUG0429
        std::cout << "in default branch\n";
#endif
    }
    return nullptr;
}

BasicBlock* BasicBlock::copybb(Function* func)
{
    BasicBlock* new_bb = new BasicBlock(func);
    for(Instruction* ins = this->begin(); ins != this->end(); ins = ins->getNext())
    {
        ins->copy(ins, new_bb);
    }
    return new_bb;
}

