#include "ConstantUnfold.h"
#include "Instruction.h"
#include "Unit.h"
#include <unordered_set>
#define ITERBOUND 40
#define DEBUG0805

#define DEBUG_SWITCH_cu 0
#if DEBUG_SWITCH_cu
#define print(str) std::cout << str << "\n";
#else
#define print(str) //
#endif

// #define DEBUG0429

// z230812 处理重复的def
void redefineDef(BasicBlock *bb) {
    std::map<Operand *, Operand *> allDef;
    for (auto instr = bb->begin(); instr != bb->end();
         instr = instr->getNext()) {
        // std::cout << "ins = " << instr->getInstName() << "\n";
        auto def = instr->getDef();
        if (!def)
            continue;
        if (!instr->isStore()) {
            if (allDef.find(def) != allDef.end()) {
                allDef[def] = new Operand(new TemporarySymbolEntry(
                    def->getSymbolEntry()->getType(), SymbolTable::getLabel()));
                // instr->setDef(allDef[def]);
                instr->getOp()[0] = allDef[def];
                allDef[def]->setDef(instr);
                def->removeUse(instr);
                // std::cout<<"old: "<<def->toStr()<<" -> new:
                // "<<allDef[def]->toStr()<<"\n";
            } else {
                allDef[def] = def;
                def->setDef(instr);
                // instr->getOp()[0] = def;
            }
        }
        for (int i = 1; i < instr->getOp().size(); i++) {
            auto op = instr->getOp()[i];
            if (!op->getDef() || op->isConst() || op->isParam())
                continue;
            if (allDef[op] && op != allDef[op]) {
                instr->getOp()[i] = allDef[op];
                allDef[op]->addUse(instr);
                instr->getOp()[i]->removeUse(instr);
            } else {
                // std::cout << op->toStr() << " "
                //           << op->getDef()->getDef()->toStr() << "???\n";
                if (std::find(op->getUse().begin(), op->getUse().end(),
                              instr) == op->getUse().end()) {
                    op->addUse(instr);
                }
            }
        }
    }
}

void ConstantUnfold::pass() {
    detector = new LoopDetection(unit);
    detector->run();
    for (auto it = detector->begin(); it != detector->end(); it++) {

        auto oneloop = *it;
        checkLoop(oneloop, (*(oneloop->begin()))->getParent());
    }
}
void displayBB(BasicBlock *bb);
void display_instruction(Instruction *ins);
// void ConstantUnfold::pass() {
// #ifdef DEBUG0429
//     print("ConstantUnfold::pass()");
// #endif
//     std::vector<Function *> func_list = unit->get_functions();
// #ifdef DEBUG0429
//     print("func_list.size() = " << func_list.size() << "");
// #endif
//     for (auto func = func_list.begin(); func != func_list.end(); func++) {
//         checkLoop(*func);
//     }
//     // for(int )
//     // auto loops = func->loop_begin();
// }

GepInstruction *find_gep_last2(BasicBlock *bb) {
    print("find_gep_last2-1");
    for (Instruction *mins = bb->begin(); mins != bb->end();
         mins = mins->getNext()) {
        // printf("!!!mins = %s\n", mins->getInstName().c_str());
        // printf("mins = %s\n", mins->getInstName().c_str());
        if (mins->isGep() && mins->getNext() && !mins->getNext()->isGep() &&
            mins->getPrev()) {
            // getchar();
            GepInstruction *pregep =
                dynamic_cast<GepInstruction *>(mins->getPrev());
            if (pregep) {
                return pregep;
            }
            break;
        }
    }
    print("find_gep_last2-2");
    return nullptr;
}

void fix1(BasicBlock *condbb, BasicBlock *body_bb, BasicBlock *end_bb) {
    if (condbb->getPredBlock().size() == 0)
        return;
    BasicBlock *mtempbb = condbb->getPredBlock()[0];
    // getchar();
    if (mtempbb->getPredBlock().size() == 0)
        return;
    BasicBlock *mtembbb_prev = mtempbb->getPredBlock()[0];
    // Operand* mop2 = body_bb->begin()->getNext()->getOp()[2];
    Operand *mop2 = nullptr;
    // mop2 = body_bb->begin()->getOp()[2];
    for (Instruction *mins = body_bb->begin(); mins != body_bb->end();
         mins = mins->getNext()) {
        // printf("!!!mins = %s\n", mins->getInstName().c_str());
        // printf("mins = %s\n", mins->getInstName().c_str());
        if (mins->isGep() && mins->getNext() && !mins->getNext()->isGep()) {
            // getchar();
            if (mins->getPrev() && mins->getPrev()->isGep()) {
                // printf("mop2 branch1");
                mop2 = mins->getPrev()->getOp()[2];
            } else {
                // printf("mop2 branch2");
                mop2 = mins->getOp()[2];
            }
            // printf("in!!");
            // mtempbb->insertBefore(i_str, mins);
            break;
        }
    }

    print("stage7");
    // return;
    assert(mop2 != nullptr);
    // printf("mop2 = %d\n", mop2->getLabel());
    // printf("traverse end_bb, find i's addr");
    LoadInstruction *i_load = nullptr;
    Instruction *i_binary = nullptr;
    // printf("end_bb no = %d\n", end_bb->getNo());
    // displayBB(end_bb);
    // getchar();
    for (Instruction *ins = end_bb->begin(); ins != end_bb->end();
         ins = ins->getNext()) {
        // printf("ins = %d\n", ins);
        // printf("ins = %s\n", ins->getInstName().c_str());
        // printf("op size = %d\n", ins->getOp().size());
        // for(int i = 0; i < ins->getOp().size(); i ++)
        // {
        //     if(ins->getOp()[i] == nullptr)
        //     {
        //         // printf("op%d == nullptr, next\n", i);
        //         continue;
        //     }
        //     // assert(0 && "ins->getOp()[i] == nullptr");
        //     // printf("ins op%d = %d\n", i, ins->getOp()[i]->getLabel());
        // }
        // printf("after print ops");
        if (ins->isBin()) {
            // return;
            print("is binary!!");
            Operand *op0 = ins->getOp()[0];
            // printf("binary op0 = %d\n", op0->getLabel());
            // assert(op0->getDst() != nullptr);
            // printf("op0 getDst = %d\n", op0->getDst()->getLabel());
            if (op0->getDst() == nullptr) {
                break;
            }
            StoreInstruction *i_str =
                new StoreInstruction(op0->getDst(), mop2, nullptr);
            // mtempbb->insertFront(i_str);
            // printf("mtempbb no = %d\n", mtempbb->getNo());
            for (Instruction *mins = mtempbb->begin(); mins != mtempbb->end();
                 mins = mins->getNext()) {
                // printf("!!!mins = %s\n", mins->getInstName().c_str());
                if (!mins->isPhi()) {
                    // printf("in!!");
                    mtempbb->insertBefore(i_str, mins);
                    break;
                }
            }

            i_load = new LoadInstruction(op0, op0->getDst(), nullptr);

            ins->setOperand(1, op0);
        }
        if (ins->isBin()) {
            i_binary = ins;
        }
    }
    // getchar();
    print("stage8");
    // 把i_load插入end_bb中binary指令前
    if (i_binary != nullptr) {
        end_bb->insertBefore(i_load, i_binary);
    }
    return;
}

void setOperandUse(BasicBlock *newbb);
// void setOperandUse(BasicBlock* newbb)
// {
//     // printf("setOperandUse begin\n");
//     for(Instruction* ins = newbb->begin(); ins != newbb->end(); ins =
//     ins->getNext())
//     {
//         // printf("instname = %s, inst pointer = %d\n",
//         ins->getInstName().c_str(), ins); for(int i = 0; i <
//         ins->getOp().size(); i ++)
//         {
//             ins->getOp()[i]->addUse(ins);
//             if(i == 0)ins->getOp()[i]->setDef(ins);
//         }
//         if(ins->isPhi())
//         {
//             // printf("ins is phi, dst = %s\n",
//             ins->getOp()[0]->toStr().c_str()); PhiInstruction* phi =
//             dynamic_cast<PhiInstruction*>(ins); for(auto src :
//             phi->getSrcs())
//             {
//                 src.second->addUse(phi);
//             }
//         }
//     }
//     // printf("setOperandUse end\n");

//     for(Instruction* ins = newbb->begin(); ins != newbb->end(); ins =
//     ins->getNext())
//     {
//         for(int i = 0; i < ins->getOp().size(); i ++)
//         {
//             // ins->getOp()[i]->addUse(ins);
//             // if(i == 0)ins->getOp()[i]->setDef(ins);
//             // printf("op%s-->opptr = %d, use:\n",
//             ins->getOp()[i]->toStr().c_str(), ins->getOp()[i]); for(auto ope
//             : ins->getOp()[i]->getUse())
//             {
//                 // printf("%d\t", ope);
//             }
//             // printf("\n");
//             // printf("op%s-->opptr = %d, def:\n",
//             ins->getOp()[i]->toStr().c_str(), ins->getOp()[i]);
//             // printf("%d\n", ins->getOp()[i]->getDef());
//         }
//         // if(ins->isPhi())
//         // {
//         //     PhiInstruction* phi = dynamic_cast<PhiInstruction*>(ins);
//         //     for(auto src : phi->getSrcs())
//         //     {
//         //         src.second->addUse(phi);
//         //     }
//         // }
//     }
// }

void ConstantUnfold::checkLoop(blockSet *oneloop, Function *func) {
    // print("======checkLoop======= " << func->getSymPtr()->toStr()
    //                                 << " =============");

    // 一个关于cond块的后继块是不是body块的测试
    // 检查body块中是否存在call指令
    BasicBlock *body_bb = *(oneloop->begin());
    if (body_bb->getSuccBlock().size() != 2) {

        return;
    }
    bool call_exist = false;
    for (Instruction *ins = body_bb->begin(); ins != body_bb->end();
         ins = ins->getNext()) {
        if (dynamic_cast<CallInstruction *>(ins))
            call_exist = false;
    }
    if (call_exist) {
        return;
    }
    int num_phi = 0;
    BasicBlock *while_bb = *(oneloop->begin());
    for (Instruction *ins = while_bb->begin(); ins != while_bb->end();
         ins = ins->getNext()) {
        if (ins->isPhi() == true)
            num_phi++;
    }
    if (num_phi != 1) {
        return;
    }
    // wzr-multi:
    for(auto f : unit->get_functions())
    {
    print("before checkFunc2");
        if(checkFunc2(f))
        {
            return;
        }
    }
    print("before unfold");
    unfold(oneloop, func);
}

void ConstantUnfold::pass2() {
    // printf("condstantunfold pass2");
    // 接下来处理内层循环变量的终值
    // 在cond_bb中找到icmp指令
    // printf("condbb label = %d\n", condbb->getNo());
    // printf("bodybb label = %d\n", bodybb->getNo());
    // condbb.
    BasicBlock *mtempbb = condbb->getPredBlock()[0];
    // printf("mtempbb label = %d\n", mtempbb->getNo());
    for (Instruction *ins = bodybb->begin(); ins != bodybb->end();
         ins = ins->getNext()) {
        // printf("ins = %d\n", ins);
        // printf("ins = %s\n", ins->getInstName().c_str());
        if (ins->isLoad()) {
            // printf("ins op0 = %d\n", ins->getOp()[0]->getLabel());
            // printf("ins op1 = %d\n", ins->getOp()[1]->getLabel());
            // printf("ins op2 = %d\n", ins->getOp()[2]->getLabel());
            // printf("ins op3 = %d\n", ins->getOp()[3]->getLabel());
        }
    }
}
void ConstantUnfold::unfold(loop_item *oneloop, Function *func) {}
// printf("is sitofp = %d\n", ins->isSitofp());
void ConstantUnfold::unfold(blockSet *oneloop, Function *func) {
    print("============= " << func->getSymPtr()->toStr() << " =============");
    print("in unfold func first line");
    // BasicBlock* cond_bb = oneloop->cond_bb;
    // BasicBlock* body_bb = oneloop->while_bb;
    // BasicBlock* end_bb = oneloop->end_bb;

    // z0804 新的循环查找
    BasicBlock *body_bb = *(oneloop->begin());
    BasicBlock *cond_bb = detector->getCondbb(body_bb);
    BasicBlock *end_bb = detector->getEndbb(body_bb);

    this->condbb = cond_bb;
    this->bodybb = body_bb;
    this->endbb = end_bb;

    print("|| cond_bb = " << cond_bb->getNo() << "");
    print("|| body_bb = " << body_bb->getNo() << "");
    print("|| end_bb = " << end_bb->getNo() << "");
    // cond_bb->removeSucc(body_bb);
    // body_bb->removePred(cond_bb);

    int start, end;
    Instruction *cond_uncond_inbody = (body_bb->getBrorRet());
    Instruction *phi = body_bb->begin();
    if (!phi->isPhi())
        return;
    print("before assert cond bb dose not start with phi");
    assert(phi->isPhi());
    // auto operands = (dynamic_cast<PhiInstruction *>(phi))->getPhiOperands();
    if (dynamic_cast<PhiInstruction *>(phi)->getSrcs().size() != 2) {
        return;
    }
    // auto bbs = (dynamic_cast<PhiInstruction *>(phi))->getPhiBBs();
    // if (operands.size() != 2) {
    //     return;
    // }
    Operand *phi_src0_second =
        dynamic_cast<PhiInstruction *>(phi)->getSrcs().begin()->second;
    auto phiit = dynamic_cast<PhiInstruction *>(phi)->getSrcs().begin();
    phiit++;
    // wzr
    Operand *phi_src1_second = phiit->second;
    // printf("operands[0], bb = %d\n",
    // dynamic_cast<PhiInstruction*>(phi)->getSrcs().begin()->first->getNo());
    // printf("operands[0] = %s\n",
    // dynamic_cast<PhiInstruction*>(phi)->getSrcs().begin()->second->toStr().c_str());
    if (!phi_src0_second->isConst()) {
        return;
    }
    // assert(bbs[0] != body_bb && "Inheriant const from loop by phi");
    bool find_loopbb_in_phi = 1;
    for (auto src : dynamic_cast<PhiInstruction *>(phi)->getSrcs()) {
        if (src.first->getNo() == body_bb->getNo()) {
            find_loopbb_in_phi = 1;
        }
    }
    // assert((bbs[0]->getNo() == body_bb->getNo() || bbs[1]->getNo() ==
    // body_bb->getNo()) && "Inheriant const from loop by phi");
    assert(find_loopbb_in_phi);

    start = phi_src0_second->getConstVal();
    print("start1 = " << start);
    // Instruction *iter = phi_src0_second->getDef();
    Instruction *cmp_inbody = find_inst(Instruction::CMP, body_bb);
    Instruction *iter = cmp_inbody->getOp()[1]->getDef();
    Operand *myop =
        dynamic_cast<CondBrInstruction *>(cond_uncond_inbody)->getCond();
    print("myop = " << myop->toStr());
    assert(myop != nullptr);
    Instruction *cmp_inbody2 = myop->getDef();
    // std::vector<Operand*> ops = cond_uncond_inbody->getOp();
    assert(cmp_inbody2 != nullptr);
    if (!cmp_inbody2->isCmp()) {
        return;
    }
    print("debug4" << cmp_inbody2->getOp()[2]->toStr());
    // return;
    if (cmp_inbody2->getOp().size() >= 3 && cmp_inbody2->getOp()[2] &&
        !cmp_inbody2->getOp()[2]->isConst()) {
        return;
    }
    end = cmp_inbody2->getOp()[2]->getConstVal();
    // // may cause problem!!
    Operand *op1 = cmp_inbody2->getOp()[1];
    Operand *op2 = op1->getDef()->getOp()[1];
    print("op1 = " << op1->toStr() << " "
                   << op1->getDef()->getOp()[1]->toStr());
    print("phi->getDef() = " << phi->getDef()->toStr());
    if (op2 != phi->getDef()) {
#ifdef DEBUG0429_terminal
        print("its no = " << cmp_inbody2->getOp()[1]->getLabel() << "");
        print("its no = " << op2->getLabel() << "");
        print("cmp_inbody2->getOp()[0] != phi");
#endif
        return;
    }
#ifdef DEBUG0429
#endif
    print("stage2");
    // return;
    int accu = 0;
    int iter_proper = -1;
    print("iter = " << iter);
    if (iter->getOp().size() != 3 || !iter->getOp()[2]->isConst()) {
        return;
    }
    accu = iter->getOp()[2]->getConstVal();
    print("accu = " << accu << "");

    if (iter->isBin() && iter->getOpcode() == BinaryInstruction::ADD) {
        iter_proper = 0;
    } else if (iter->isBin() && iter->getOpcode() == BinaryInstruction::SUB) {
        iter_proper = 1;
        accu = -accu;
    } else {
        return;
    }
    int iternum = (end - start) / accu;
    iternum = abs(iternum);
    print("iternum  = " << iternum);
    if (iternum > ITERBOUND) {
#ifdef DEBUG0429_terminal
        print("(end - start) / accu > ITERBOUND");
#endif
        return;
    }
    if (iter_proper == -1) {
#ifdef DEBUG0429_terminal
        print("iter_proper == -1");
#endif
        return;
    }

    if (cmp_inbody2->getOpcode() != CmpInstruction::L) {
#ifdef DEBUG0429_terminal
        print("cmp_inbody2->getOpcode() != CmpInstruction::L");
#endif
        return;
    }
#ifdef DEBUG0429
#endif
    // std::cout<<"stage3\n";
    std::unordered_set<Instruction *> ins_not_process{phi, cmp_inbody2,
                                                      cond_uncond_inbody};
#ifdef DEBUG0429
    print("ignored: phi = " << phi << " cmp_inbody2 = " << compare) r
        << " cond_uncond_inbody = " << cond_uncond_inbody << "\n";
#endif
    // BasicBlock *end_bb = detector->getEndbb(body_bb);

// for(auto BB : cond_bb->getSuccBlock())
// {
//     if(BB != body_bb)end_bb = BB;
// }
#ifdef DEBUG0429
    print("end_bb.no = " << end_bb->getNo() << "");
#endif

    auto block_list0 = func->getBlockList();

    BasicBlock *new_bb = new BasicBlock(func);
    // std::cout<<"new_bb No = "<<new_bb->getNo()<<"\n";
    std::unordered_map<Instruction *, Instruction *> nounfold_unfold;
    print("start = " << start << " end = " << end << "");

    // std::cout<<"unfold begin\n";
    std::vector<Function *> func_list = unit->get_functions();

    std::vector<Instruction *> sitofps;
    for (int h = 0; h < iternum; h++) {
        nounfold_unfold.clear();

        print("below transfer body_bb for insert");

        // 找到body块中的fsub指令，确认一下是否在copy时没有拷贝上浮点数信息
        Instruction *fsub = nullptr;
        Instruction *mysitofp = nullptr;
        for (Instruction *ins = body_bb->begin(); ins != body_bb->end();
             ins = ins->getNext()) {
            if (ins->isBin() && ins->getOpcode() == 0) {
                fsub = ins;
            }

            if (ins_not_process.find(ins) != ins_not_process.end()) {
                continue;
            }
            if (ins == iter || ins->isCond() || ins->isUncond()) {
                continue;
            }
            // std::cout<<"??? instr name = "<<ins->getInstName()<<"\n";
            // Instruction *new_ins = ins->copy(ins, new_bb);
            print("insert inst pointer = " << ins << "  inst type = "
                                           << ins->getInstName());
            Instruction *new_ins = ins->copy1();
            new_bb->insertBack(new_ins);
            // std::cout<<"new>>>\n";
            nounfold_unfold.insert({ins, new_ins});
            print("nounfold_unfold[ins] = " << nounfold_unfold[ins]);
            // std::cout<<"insert>>>\n";
            if (new_ins->isSitofp()) {
                // printf("insert to sitofps");
                sitofps.push_back(new_ins);
            }
        }
        print("transfer1-1");
        setOperandUse(new_bb);
        print("transfer1-2");

        // printf("mysitofp = %d\n", mysitofp);
        // 看看fsub转换成=的指令是不是浮点指令
        // printf("fsub = %d\n", fsub);
        // printf("fsub insname 1 = %s\n", fsub->getInstName().c_str());
        // printf("fsub dst op label = %d\n", fsub->getOp()[0]->getLabel());
        // printf("fsub op0 type 1 = %s\n",
        // nounfold_unfold[fsub]->getOp()[0]->getType()->toStr().c_str());
        // printf("old fsub op0 type 1 = %s\n",
        // fsub->getOp()[0]->getType()->toStr().c_str());

        // print("below transfer body_bb for exchange operand");
        bool flag = false;

        for (Instruction *ins = body_bb->begin(); ins != body_bb->end();
             ins = ins->getNext()) {
            // print("=    "<<ins->getInstName());
            print("transfer1-2-1");
            if (ins_not_process.find(ins) != ins_not_process.end()) {
                continue;
            }
            if (ins == iter) {
                start += accu;
                // std::cout<<"start2 = "<<start<<"\n";
                continue;
            } else if (ins->isCond() || ins->isUncond()) {
                continue;
            } else if (ins->isPhi()) {
                continue;
            } else if (ins->isGep())
            // else if(ins->isCall())
            {
                // std::cout<<"isgep iscall 1\n";
                Operand *d = ins->getOp()[0];
                Type *type = TypeSystem::intType;
                auto tmp = dynamic_cast<PointerType *>(d->getType());
                d = new Operand(new TemporarySymbolEntry(
                    new PointerType(tmp->getType()), SymbolTable::getLabel()));
                nounfold_unfold[ins]->setOperand(0, d);
                // std::cout<<"isgep iscall 2\n";
            }

            print("transfer1-2-2");
            // print("ins op size = " << ins->getOp().size() << "");
            // print(ins->getPrev()->getDef()->toStr());
            // print(ins->getOp()[0]->toStr());
            // print(ins->getPrev()->getDef()->getUse().size());
            // print(ins->getPrev()->getDef()->getUse()[0]->getInstName());
            if (ins->isStore() && ins->getPrev() && ins->getPrev()->isGep() &&
                (ins->getPrev()->getDef()) == ins->getOp()[0]) {
                nounfold_unfold[ins]->setOperand(
                    0, nounfold_unfold[ins->getPrev()]->getDef());
            } else if (ins->isLoad() && ins->getPrev() &&
                       ins->getPrev()->isGep() && ins->getPrev()->getDef() &&
                       ins->getPrev()->getDef()->getUse().size() > 0 &&
                       (ins->getPrev()->getDef()->getUse()[0]) == ins) {
                // nounfold_unfold[ins]->setOperand(0,
                // nounfold_unfold[ins->getPrev()]->getDef());
                Type *tmptype = ins->getOp()[1]->getType();
                if (!tmptype->isPtr()) {
                    assert(0 && "load operand1(src) type is not PTR!");
                }
                // dynamic_cast<PointerType*>(tmptype)->getType()->isFloat();
                if ((dynamic_cast<PointerType *>(tmptype))
                        ->getType()
                        ->isInt()) {
                    tmptype = TypeSystem::intType;
                } else if ((dynamic_cast<PointerType *>(tmptype))
                               ->getType()
                               ->isFloat()) {
                    // printf("here!");
                    tmptype = TypeSystem::floatType;
                }
                Operand *newope = new Operand(
                    new TemporarySymbolEntry(tmptype, SymbolTable::getLabel()));
                Operand *oldope = nounfold_unfold[ins]->getOp()[0];
                // 拷贝def/use关系
                newope->setDef(ins);
                newope->addUse(nounfold_unfold[ins->getNext()]);

                nounfold_unfold[ins]->setOperand(0, newope);
            }
            print("transfer1-2-3");

            for (unsigned i = 1; i < ins->getOp().size(); i++) {
                print("transfer1-2-3-1");
                if (ins->isGep() && i == 1) {
                    if (ins->getOp()[1]->getDef() == nullptr)
                        continue;
                    if (ins->getOp()[1]->getDef()->isGep() &&
                        nounfold_unfold[ins->getOp()[1]->getDef()]) {
                        // 如old2new[ins->getOp()[1]->getDef()] == 0,
                        // 说明该数组的Arr作为临时变量 在别的块中被Def,
                        // 这里不管它就好 std::cout<<"gep but not the first\n";
                        print("ins->getOp()[1] = "
                              << ins->getOp()[1]->toStr().c_str());
                        print("inst pointer = " << ins << "inst type = "
                                                << ins->getInstName());
                        print("inst pointer = "
                              << ins->getOp()[1]->getDef() << "inst type = "
                              << ins->getOp()[1]->getDef()->getInstName());
                        // display_instruction(ins->getOp()[1]->getDef());
                        print("nounfold_unfold[ins->getOp()[1]->getDef()] = "
                              << nounfold_unfold[ins->getOp()[1]->getDef()]);
                        // getchar();

                        nounfold_unfold[ins]->setOperand(
                            1, nounfold_unfold[ins->getOp()[1]->getDef()]
                                   ->getOp()[0]);
                    }
                }

                print("transfer1-2-3-2");
                Instruction *ope = ins->getOp()[i]->getDef();
                if (ope == nullptr) {
                    // std::cout<<"continue\n";
                    continue;
                }
                print("transfer1-2-3-3");
                if (ins->getOp()[i] == phi->getDef() || ope == iter) {
                    print("transfer1-2-3-3-1");
                    // if(ins->isSitofp())
                    //     std::cout<<"branch 1\n";
                    // nounfold_unfold[ins]->setOperand(i, new Operand(new
                    // ConstantSymbolEntry(TypeSystem::floatType, start)));
                    nounfold_unfold[ins]->setOperand(
                        i, new Operand(new ConstantSymbolEntry(
                               TypeSystem::constIntType, start)));
                    print("transfer1-2-3-3-2");
                } else if (nounfold_unfold.find(ope) != nounfold_unfold.end() &&
                           nounfold_unfold[ope]) {
                    // print("branch nounfold_unfold.find(ope) !=
                    // nounfold_unfold.end()"); if(ins->isSitofp())
                    //     std::cout<<"branch2\n";
                    // printf("is sitofp = %d\n", ins->isSitofp());
                    // std::cout<<"setOperand 4\n";
                    print("transfer1-2-3-3-3");
                    print("nounfold_unfold[ins] = " << nounfold_unfold[ins]);
                    print("nounfold_unfold[ope] = " << nounfold_unfold[ope]);
                    nounfold_unfold[ins]->setOperand(
                        i, nounfold_unfold[ope]->getOp()[0]);
                    if (ins->isStore() && i == 1 && ins->getPrev() &&
                        ins->getPrev()->isGep()) {
                        print("transfer1-2-3-3-3-1");
                        // printf("gep before store op1 type = %s\n",
                        // ins->getPrev()->getOp()[1]->getSymbolEntry()->getType()->toStr().c_str());
                        assert(ins->getPrev()
                                   ->getOp()[1]
                                   ->getSymbolEntry()
                                   ->getType()
                                   ->isPtr() &&
                               "gep op1 is not pointer type!! will cause "
                               "nullptr fault!!");
                        // printf("isfloat? %d\n",
                        // dynamic_cast<PointerType*>(ins->getPrev()->getOp()[1]->getSymbolEntry()->getType())->getType()->isFloat());
                        PointerType *mytype =
                            dynamic_cast<PointerType *>(ins->getPrev()
                                                            ->getOp()[1]
                                                            ->getSymbolEntry()
                                                            ->getType());
                        print("transfer1-2-3-3-3-2");
                        if (mytype->isPtr()) {
                            print("transfer1-2-3-3-3-2-1");
                            Type *mytype2 = mytype->getType();
                            if (mytype2->isArray()) {
                                // printf("in array branch");
                                print("transfer1-2-3-3-3-2-2");
                                if (dynamic_cast<ArrayType *>(mytype2)
                                        ->getElementType()
                                        ->isFloat()) {
                                    // printf("in array1 branch");
                                    print("transfer1-2-3-3-3-2-3");
                                    nounfold_unfold[ins]
                                        ->getOp()[1]
                                        ->getSymbolEntry()
                                        ->setType(TypeSystem::floatType);
                                    print("transfer1-2-3-3-3-2-4");
                                } else if (dynamic_cast<ArrayType *>(mytype2)
                                               ->getElementType()
                                               ->isInt()) {
                                    // printf("in array2 branch");
                                    print("transfer1-2-3-3-3-2-5");
                                    nounfold_unfold[ins]
                                        ->getOp()[1]
                                        ->getSymbolEntry()
                                        ->setType(TypeSystem::intType);
                                    print("transfer1-2-3-3-3-2-6");
                                    // nounfold_unfold[ins]->getOp()[1]->getSymbolEntry()->setType(TypeSystem::floatType);
                                } else {
                                    printf("!!!gep before store op1 type's "
                                           "type = %s\n",
                                           dynamic_cast<PointerType *>(
                                               ins->getPrev()
                                                   ->getOp()[1]
                                                   ->getSymbolEntry()
                                                   ->getType())
                                               ->getType()
                                               ->toStr()
                                               .c_str());
                                    assert(0 &&
                                           "not int type and not float type");
                                    print("transfer1-2-3-3-3-2-7");
                                }
                            } else {

                                print("transfer1-2-3-3-3-2-8");
                                if (dynamic_cast<PointerType *>(
                                        ins->getPrev()
                                            ->getOp()[1]
                                            ->getSymbolEntry()
                                            ->getType())
                                        ->getType()
                                        ->isFloat()) {
                                    print("transfer1-2-3-3-3-2-9");
                                    nounfold_unfold[ins]
                                        ->getOp()[1]
                                        ->getSymbolEntry()
                                        ->setType(TypeSystem::floatType);
                                    print("transfer1-2-3-3-3-2-10");
                                } else if (dynamic_cast<PointerType *>(
                                               ins->getPrev()
                                                   ->getOp()[1]
                                                   ->getSymbolEntry()
                                                   ->getType())
                                               ->getType()
                                               ->isInt()) {
                                    print("transfer1-2-3-3-3-2-11");
                                    nounfold_unfold[ins]
                                        ->getOp()[1]
                                        ->getSymbolEntry()
                                        ->setType(TypeSystem::intType);
                                    // nounfold_unfold[ins]->getOp()[1]->getSymbolEntry()->setType(TypeSystem::floatType);
                                    print("transfer1-2-3-3-3-2-12");

                                } else {
                                    printf("!!!gep before store op1 type's "
                                           "type = %s\n",
                                           dynamic_cast<PointerType *>(
                                               ins->getPrev()
                                                   ->getOp()[1]
                                                   ->getSymbolEntry()
                                                   ->getType())
                                               ->getType()
                                               ->toStr()
                                               .c_str());
                                    assert(0 &&
                                           "not int type and not float type");
                                }
                            }
                            print("transfer1-2-3-3-3-3");
                        } else {
                            assert(0 && "gep op1 type is not ptr");
                        }
                        print("transfer1-2-3-3-4");
                    }
                    // nounfold_unfold[ins]->getOp()[1]->getSymbolEntry()->setType(TypeSystem::intType);
                    // std::cout<<"nounfold_unfold[ins]->getOp()[i] =
                    // "<<nounfold_unfold[ins]->getOp()[i]->getType()->toStr()<<"\n";
                    // std::cout<<"nounfold_unfold[ins]->getOp()[i] =
                    // "<<nounfold_unfold[ins]->getOp()[i]->toStr()<<"\n";
                    nounfold_unfold[ope]->getOp()[0]->addUse(
                        nounfold_unfold[ins]);
                } else {
                    // print("branch else");
                    print("transfer1-2-3-3-5");
                    nounfold_unfold[ins]->setOperand(i, ope->getOp()[0]);
                    ope->getOp()[0]->addUse(nounfold_unfold[ins]);
                    print("transfer1-2-3-3-6");
                }
                print("transfer1-2-3-4");
                // std::cout<<"small loop end\n";
            }
            print("transfer1-2-4");
        }
        print("transfer1-3");
        // print("after transfer body_bb for exchange");
        for (Instruction *ins = new_bb->begin(); ins != new_bb->end();
             ins = ins->getNext()) {
            // if(ins->isCall())
            // {
            //     std::cout<<"call dst = "<<ins->getDef()->getLabel()<<" call
            //     dst =
            //     "<<dynamic_cast<CallInstruction*>(ins)->dst->getLabel()<<"\n";
            // }
        }
    }
    // for(Instruction* ins : sitofps)
    // {
    //     printf("one sitofps ins op0 = %d, op1 = %d\n",
    //     ins->getOp()[0]->getLabel(), ins->getOp()[1]->getLabel());
    // }
    print("after while(start < end)");
#ifdef DEBUG0429
#endif
    // std::cout<<"stage4\n";
    nounfold_unfold[iter] = new BinaryInstruction(
        BinaryInstruction::ADD, iter->getDef(),
        new Operand(new ConstantSymbolEntry(TypeSystem::intType, start)),
        new Operand(new ConstantSymbolEntry(TypeSystem::intType, accu - 1)),
        new_bb);

    nounfold_unfold.clear();
    std::unordered_map<Instruction *, Instruction *> nounfold_unfold_forendbb;
#ifdef DEBUG0429
#endif
    print("before end_bb process");

    Instruction *outbb_phi = nullptr;
    Operand *outbb_phi_dst;
    for (Instruction *ins = end_bb->begin(); ins != end_bb->end();
         ins = ins->getNext()) {
        // std::cout<<"ins = "<<ins<<"\n";
        // printf("is sitofp(end_bb) = %d\n", ins->isSitofp());
        if (ins->isPhi()) {
            outbb_phi = ins;
            outbb_phi_dst = ins->getDef();
            print("outbb_phi_dst = " << outbb_phi_dst->toStr() << "");
        }
        // std::cout<<"ins.type = "<<ins->getInstName()<<"\n";
        // std::cout<<"ins.ops = "<<ins->getOp().size()<<"\n";
        for (unsigned i = 0; i < ins->getOp().size(); i++) {
            // std::cout<<"breakpoint0\n";
            if (ins->getOp()[i] == nullptr) {
                continue;
            }
            Instruction *ope = ins->getOp()[i]->getDef();
            if (ope == nullptr) {
                // std::cout<<"breakpoint1\n";
#ifdef DEBUG0429_terminal
                print("ope == nullptr");
#endif
                continue;
            }
            if (nounfold_unfold_forendbb.find(ope) !=
                nounfold_unfold_forendbb.end()) {
                // 这个分支不会进入
                // std::cout<<"breakpoint2\n";
                ins->setOperand(i, nounfold_unfold_forendbb[ope]->getOp()[0]);
            } else if (ope->isPhi()) {
                // std::cout<<"breakpoint3\n";
                // ins->setOperand(
                //     i, new Operand(new ConstantSymbolEntry(
                //            TypeSystem::constIntType, start + accu)));
            }
            // std::cout<<"breakpoint4\n";
        }
    }

    print("after end_bb process");

    // printf("bodybb = %d\n", body_bb->getNo());
    // printf("condbb label = %d\n", condbb->getNo());
    // printf("cond succ size = %d\n", cond_bb->getSuccBlock().size());
    // printf("bodybb label = %d\n", bodybb->getNo());
    // wzr!!: 这里下面的getPredBlock()直接取[0]是有可能导致问题的
    bool link = false;
    for (auto succ : cond_bb->getSuccBlock()) {
        print("succ " << succ->getNo() << " body " << body_bb->getNo());
        // std::cout<<"succ "<<succ->getNo()<<" body "<<body_bb->getNo()<<"\n";
        if (succ == body_bb)
            link = true;
    }

    print("stage6");
    // getchar();
    // fix1(cond_bb, body_bb, end_bb);
    Operand *prewhile_i = nullptr;
    GepInstruction *gep_last2 = find_gep_last2(new_bb);
    print("stage6-1");
    if (gep_last2) {
        prewhile_i = gep_last2->getOp()[2];
    }
    print("stage6-2");
    if (prewhile_i) {
        BinaryInstruction *binary_inend = dynamic_cast<BinaryInstruction *>(
            find_inst(Instruction::BINARY, end_bb));
        if (binary_inend)
            binary_inend->setOperand(1, prewhile_i);
    }
    print("stage6-3");
    if (condbb->getPredBlock().size() != 0) {
        // BasicBlock *mtempbb = condbb->getPredBlock()[0];
        // // getchar();
        // BasicBlock *mtembbb_prev = mtempbb->getPredBlock()[0];
        // // Operand* mop2 = body_bb->begin()->getNext()->getOp()[2];
        // Operand *mop2 = nullptr;
        // // mop2 = body_bb->begin()->getOp()[2];
        // for (Instruction *mins = body_bb->begin(); mins != body_bb->end();
        //      mins = mins->getNext()) {
        //     // printf("!!!mins = %s\n", mins->getInstName().c_str());
        //     // printf("mins = %s\n", mins->getInstName().c_str());
        //     if (mins->isGep() && mins->getNext() &&
        //     !mins->getNext()->isGep()) {
        //         // getchar();
        //         if (mins->getPrev() && mins->getPrev()->isGep()) {
        //             // printf("mop2 branch1");
        //             mop2 = mins->getPrev()->getOp()[2];
        //         } else {
        //             // printf("mop2 branch2");
        //             mop2 = mins->getOp()[2];
        //         }
        //         // printf("in!!");
        //         // mtempbb->insertBefore(i_str, mins);
        //         break;
        //     }
        // }

        // printf("stage7\n");
        // // return;
        // assert(mop2 != nullptr);
        // // printf("mop2 = %d\n", mop2->getLabel());
        // // printf("traverse end_bb, find i's addr");
        // LoadInstruction *i_load = nullptr;
        // Instruction *i_binary = nullptr;
        // printf("end_bb no = %d\n", end_bb->getNo());
        // displayBB(end_bb);
        // // getchar();
        // for (Instruction *ins = end_bb->begin(); ins != end_bb->end();
        //      ins = ins->getNext()) {
        //     // printf("ins = %d\n", ins);
        //     // printf("ins = %s\n", ins->getInstName().c_str());
        //     // printf("op size = %d\n", ins->getOp().size());
        //     // for(int i = 0; i < ins->getOp().size(); i ++)
        //     // {
        //     //     if(ins->getOp()[i] == nullptr)
        //     //     {
        //     //         // printf("op%d == nullptr, next\n", i);
        //     //         continue;
        //     //     }
        //     //     // assert(0 && "ins->getOp()[i] == nullptr");
        //     //     // printf("ins op%d = %d\n", i,
        //     ins->getOp()[i]->getLabel());
        //     // }
        //     // printf("after print ops");
        //     if (ins->isBin()) {
        //         // return;
        //         print("is binary!!");
        //         Operand *op0 = ins->getOp()[0];
        //         // printf("binary op0 = %d\n", op0->getLabel());
        //         // assert(op0->getDst() != nullptr);
        //         // printf("op0 getDst = %d\n", op0->getDst()->getLabel());
        //         if (op0->getDst() == nullptr) {
        //             break;
        //         }
        //         StoreInstruction *i_str =
        //             new StoreInstruction(op0->getDst(), mop2, nullptr);
        //         // mtempbb->insertFront(i_str);
        //         // printf("mtempbb no = %d\n", mtempbb->getNo());
        //         for (Instruction *mins = mtempbb->begin(); mins !=
        //         mtempbb->end();
        //              mins = mins->getNext()) {
        //             // printf("!!!mins = %s\n", mins->getInstName().c_str());
        //             if (!mins->isPhi()) {
        //                 // printf("in!!");
        //                 mtempbb->insertBefore(i_str, mins);
        //                 break;
        //             }
        //         }

        //         i_load = new LoadInstruction(op0, op0->getDst(), nullptr);

        //         ins->setOperand(1, op0);
        //     }
        //     if (ins->isBin()) {
        //         i_binary = ins;
        //     }
        // }
        // getchar();
        // print("stage8");
        // // 把i_load插入end_bb中binary指令前
        // return;
        // if (i_binary != nullptr) {
        //     end_bb->insertBefore(i_load, i_binary);
        // }
    }

    // 向end_bb里插入从i_addr中load i值的语句

    // std::cout<<"body_bb->getSuccBlock().size()1 =
    // "<<body_bb->getSuccBlock().size()<<"\n";
    //  Instruction* binaryins = new BinaryInstruction(BinaryInstruction::ADD,
    //  outbb_phi_dst, iter->getDef(),
    //                                     new Operand(new
    //                                     ConstantSymbolEntry(TypeSystem::intType,
    //                                     0)));
    //  end_bb->insertFront(binaryins);

    // std::cout<<outbb_phi->getParent()->getNo()<<"
    // "<<outbb_phi->getDef()->toStr()<<" ???????\n";
    // if (outbb_phi != nullptr && outbb_phi->getParent() == end_bb)
    //     end_bb->remove(outbb_phi);

    // clear

    // cond_bb->clearInstructions();
    // body_bb->clearInstructions();

    // return;
    //  std::cout<<"new_bb: "<<new_bb->getNo()<<"\n";
    //  std::cout<<"end_bb: "<<end_bb->getNo()<<"\n";
    if (end_bb->existsInPred(body_bb)) {
        end_bb->removePred(body_bb);
    }
    if (body_bb->existsInSucc(end_bb)) {
        body_bb->removeSucc(end_bb);
    }
    print("stage9");
    UncondBrInstruction *uncondbr = new UncondBrInstruction(end_bb, new_bb);
    new_bb->addSucc(end_bb);
    end_bb->addPred(new_bb);

    // if(end_bb->getNo() == 1186)
    // {
    //     print("check end bb");
    //     displayBB(end_bb);
    // }

    if (find_inst(Instruction::CMP, cond_bb)) {
        auto cmp = dynamic_cast<CmpInstruction *>(
            find_inst(Instruction::CMP, cond_bb));
        cmp->getParent()->remove(cmp);
    }
    if (find_inst(Instruction::COND, cond_bb)) {
        auto cond = dynamic_cast<CondBrInstruction *>(
            find_inst(Instruction::COND, cond_bb));
        cond->getParent()->remove(cond);
    }
    if (find_inst(Instruction::UNCOND, cond_bb)) {
        auto uncond = dynamic_cast<UncondBrInstruction *>(
            find_inst(Instruction::UNCOND, cond_bb));
        uncond->setBranch(new_bb);
    } else
        new UncondBrInstruction(new_bb, cond_bb);
    if (body_bb->existsInPred(cond_bb)) {
        body_bb->removePred(cond_bb);
    }
    if (cond_bb->existsInSucc(body_bb)) {
        cond_bb->removeSucc(body_bb);
    }
    if (!new_bb->existsInPred(cond_bb))
        new_bb->addPred(cond_bb);
    if (!cond_bb->existsInSucc(new_bb))
        cond_bb->addSucc(new_bb);
    //     for (BasicBlock *pred_cond : cond_bb->getPredBlock()) {
    //         print(">>>>>>>>>>>>>>>>>>>>> cond_bb: " << cond_bb->getNo());
    //         print(">>>>>>>>>>>>>>>>>>>>> pred_cond: " << pred_cond->getNo());
    //         if (pred_cond == body_bb) {
    // #ifdef DEBUG0429_terminal
    //             print("pred_cond == body_bb");
    // #endif
    //             continue;
    //         }
    //         Instruction *b = pred_cond->getBrorRet();
    //         if (link) {
    //             if (b->getInstName() == "COND") {
    //                 CondBrInstruction *condbr =
    //                     dynamic_cast<CondBrInstruction *>(b);
    //                 if (condbr->getTrueBranch() == cond_bb) {
    //                     condbr->setTrueBranch(new_bb);
    //                     pred_cond->addSucc(new_bb);
    //                     pred_cond->removeSucc(cond_bb);
    //                     // new_bb->removePred(pred_cond);
    //                     new_bb->addPred(pred_cond);
    //                 } else if (condbr->getFalseBranch() == cond_bb) {
    //                     condbr->setFalseBranch(new_bb);
    //                     pred_cond->addSucc(new_bb);
    //                     pred_cond->removeSucc(cond_bb);

    //                     new_bb->addPred(pred_cond);
    //                 }
    //             } else if (b->getInstName() == "UNCOND") {
    //                 UncondBrInstruction *uncondbr =
    //                     dynamic_cast<UncondBrInstruction *>(b);
    //                 if (uncondbr->getBranch() == cond_bb) {
    //                     uncondbr->setBranch(new_bb);
    //                     pred_cond->addSucc(new_bb);
    //                     pred_cond->removeSucc(cond_bb);
    //                     new_bb->addPred(pred_cond);
    //                 }
    //             }
    //         } else {
    //             BasicBlock *invar_bb;
    //             for (auto pred : body_bb->getPredBlock()) {
    //                 if (pred != body_bb) {
    //                     invar_bb = pred;
    //                     break;
    //                 }
    //             }
    //             if (invar_bb == nullptr)
    //                 assert(0);

    //             auto uncond_ = (UncondBrInstruction *)(find_inst(
    //                 Instruction::UNCOND, invar_bb));
    //             uncond_->setBranch(new_bb);

    //             invar_bb->addSucc(new_bb);
    //             invar_bb->removeSucc(body_bb);
    //             body_bb->removePred(invar_bb);
    //             new_bb->addPred(invar_bb);
    //             if (b->getInstName() == "COND") {
    //                 CondBrInstruction *condbr =
    //                     dynamic_cast<CondBrInstruction *>(b);
    //                 if (condbr->getTrueBranch() == cond_bb) {
    //                     condbr->setTrueBranch(invar_bb);
    //                     pred_cond->addSucc(invar_bb);
    //                     pred_cond->removeSucc(cond_bb);
    //                     // invar_bb->removePred(pred_cond);
    //                     invar_bb->addPred(pred_cond);
    //                 } else if (condbr->getFalseBranch() == cond_bb) {
    //                     condbr->setFalseBranch(invar_bb);
    //                     pred_cond->addSucc(invar_bb);
    //                     pred_cond->removeSucc(cond_bb);

    //                     invar_bb->addPred(pred_cond);
    //                 }
    //             } else if (b->getInstName() == "UNCOND") {
    //                 UncondBrInstruction *uncondbr =
    //                     dynamic_cast<UncondBrInstruction *>(b);
    //                 if (uncondbr->getBranch() == cond_bb) {
    //                     uncondbr->setBranch(invar_bb);
    //                     pred_cond->addSucc(invar_bb);
    //                     pred_cond->removeSucc(cond_bb);
    //                     invar_bb->addPred(pred_cond);
    //                 }
    //             }
    //         }
    //     }

    print("stage10");

    // std::cout<<"cond_bb = "<<cond_bb->getNo()<<"\n";
    // if (cond_bb != nullptr)
    //     cond_bb->deleteAllRelation();
    // cond_bb->getParent()->remove(cond_bb);
    if (body_bb != nullptr)
        body_bb->deleteAllRelation();
    body_bb->getParent()->remove(body_bb);
    print("after delete body_bb = " << body_bb->getNo());
    // redefineDef(new_bb);
    std::cout << "yes unfold done.\n";
    print("========= end =========");
}
