#include <iostream>
#include "Instruction.h"
#include "BasicBlock.h"
#include "Function.h"
// #include "Type.h"
#include <cmath>
#include <sstream>
extern FILE* yyout;
// extern Unit unit;
// extern MachineUnit mUnit;

class Operand;

void Instruction::setOperand2(int index, Operand *ope)
{
    // std::cout<<"index = "<<index<<" size = "<<operands.size()<<"\n";
    assert(index >= 0 && index < operands.size() &&
            "index too small or too big");
    // wzr-multi:
    if(index == 0)
    {
        ope->setDef(this);
        operands[index]->removeDef(this);
    }
    else
    {
        ope->addUse(this);
        operands[index]->removeUse(this);
    }
    operands[index] = ope;
}

void PhiInstruction::changeSrcBlock2(BasicBlock *_old, BasicBlock *_new) {
    // printf("changeSrcBlock2 start\n");
    std::vector<BasicBlock *> bbtoerase;
    Operand *temp2 = nullptr;
    for (auto src : srcs) {
        // printf("before if\n");
        // if(src == nullptr)continue;
        // printf("src = %d\n", src);
        // printf("src.first = %d\n", src.first->getNo());
        if (src.first == _old) {
            // printf("in if1\n");
            // Operand *temp = src.second;
            bbtoerase.push_back(src.first);
            temp2 = src.second;
            // srcs.erase(src.first);
            // // srcs[_new] = temp;
            // printf("in if2\n");
        }
    }
    for (int i = 0; i < bbtoerase.size(); i++) {
        srcs.erase(bbtoerase[i]);
    }
    // assert(temp2);
    srcs[_new] = temp2;
    // printf("changeSrcBlock2 end\n");
}

Instruction::Instruction(unsigned instType, BasicBlock* insert_bb) {
    opcode = -1;
    this->instType = instType;
    prev = next = this;
    if (insert_bb != nullptr) {
        insert_bb->insertBack(this);
        parent = insert_bb;
    }
}

Instruction::~Instruction() {
    parent->remove(this);
}

void Instruction::setParent(BasicBlock* bb) {
    parent = bb;
}

BasicBlock* Instruction::getParent() {
    return parent;
}

void Instruction::setNext(Instruction* inst) {
    next = inst;
}

void Instruction::setPrev(Instruction* inst) {
    prev = inst;
}

Instruction* Instruction::getNext() {
    return next;
}

Instruction* Instruction::getPrev() {
    return prev;
}

BinaryInstruction::BinaryInstruction(unsigned opcode,
                                     Operand* dst,
                                     Operand* src1,
                                     Operand* src2,
                                     BasicBlock* insert_bb)
    : Instruction(BINARY, insert_bb) {
    this->opcode = opcode;
    operands.push_back(dst);
    operands.push_back(src1); 
    operands.push_back(src2);
    dst->setDef(this);
    // std::cout << src1->getType()->toStr() << ": "<< src1->toStr() <<std::endl;
    src1->addUse(this);
    // std::cout << "addUse succeed." << std::endl;
    src2->addUse(this);
}

void BinaryInstruction::replaceUse(Operand* old, Operand* new_) {
    // std::cout<<operands[0]->toStr()<<"\n";
    if (operands[1] == old) {
        operands[1]->removeUse(this);
        operands[1] = new_;
        new_->addUse(this);
    } else if (operands[2] == old) {
        operands[2]->removeUse(this);
        operands[2] = new_;
        new_->addUse(this);
    }
    // operands.push_back(dst);
}

BinaryInstruction::~BinaryInstruction() {
    operands[0]->setDef(nullptr);
    if (operands[0]->usersNum() == 0)
        delete operands[0];
    operands[1]->removeUse(this);
    operands[2]->removeUse(this);
}

void BinaryInstruction::output() const {
    std::string s1, s2, s3, op;
    s1 = operands[0]->toStr();
    s2 = operands[1]->toStr();
    s3 = operands[2]->toStr();
    Type* type = operands[0]->getType();
    if(type->isFloat()) {
        switch (opcode) {
        case ADD:
            op = "fadd";
            break;
        case SUB:
            op = "fsub";
            break;
        case MUL:
            op = "fmul";
            break;
        case DIV:
            op = "fdiv";
            break;
        case MOD:
            op = "srem";
            break;
        default:
            break;
        }
    } else {
        switch (opcode) {
        case ADD:
            op = "add";
            break;
        case SUB:
            op = "sub";
            break;
        case MUL:
            op = "mul";
            break;
        case DIV:
            op = "sdiv";
            break;
        case MOD:
            op = "srem";
            break;
        default:
            break;
        }   
    }
    fprintf(yyout, "  %s = %s %s %s, %s\n", s1.c_str(), op.c_str(),
            type->toStr().c_str(), s2.c_str(), s3.c_str());
}


bool isPowerOf2(int num) {
    return (num & (num - 1)) ? false : true;
}

int log2ForIntpowOf2(int num) {
    int res = -1;
    while (num) {
        num >>= 1;
        res++;
    }
    return res;
}

int getMinLargerPowOf2(int num) {
    int res = 1;
    while(res < num) {
        res = res << 1;
    }
    return 256;
}

int clz(int x) { return __builtin_clz(x); }
int ctz(int x) { return __builtin_ctz(x); }

void BinaryInstruction::genMachineCode(AsmBuilder* builder) {
    auto cur_block = builder->getBlock();
    // wzr-multi:
    // if(addMov)
    // {
    //     cur_inst = new MovMInstruction(cur_block, MovMInstruction::MOV, src1, 
    //         genMachineReg(10));
    //     cur_block->InsertInst(cur_inst);
    // }
    if(operands[0]->getType()->isInt()) { //int;
        auto dst = genMachineOperand(operands[0]);
        auto src1 = genMachineOperand(operands[1]);
        auto src2 = genMachineOperand(operands[2]);
        //StrengthReduction:
        if(src1->isImm() && opcode == MUL) {  
            if(src2->isImm()) {
                MachineOperand* temp = new MachineOperand(MachineOperand::IMM, src1->getVal()*src2->getVal());
                MachineInstruction* cur_inst = new MovMInstruction(
                        cur_block, MovMInstruction::MOV, dst, temp);
                cur_block->InsertInst(cur_inst);
                return;
            }
            if(src1->getVal() == 0) {
                MachineOperand* temp = new MachineOperand(MachineOperand::IMM, 0);
                MachineInstruction* cur_inst = new MovMInstruction(
                            cur_block, MovMInstruction::MOV, dst, temp);
                cur_block->InsertInst(cur_inst);
                return;
            } else if(src1->getVal() == 1) {
                MachineInstruction* cur_inst = new MovMInstruction(
                            cur_block, MovMInstruction::MOV, dst, src2);
                cur_block->InsertInst(cur_inst);
                return;
            } else if(src1->getVal() == 2) {
                MachineOperand* temp = new MachineOperand(*src2);
                MachineInstruction* cur_inst = new BinaryMInstruction(
                            cur_block, BinaryMInstruction::ADD, dst, src2, temp);
                cur_block->InsertInst(cur_inst);
                return;
            } else if(isPowerOf2(src1->getVal())) {
                MachineOperand* pow = new MachineOperand(MachineOperand::IMM, log2ForIntpowOf2(src1->getVal()));
                MachineInstruction* cur_inst = new BinaryMInstruction(
                            cur_block, BinaryMInstruction::LSL, dst, src2, pow);
                cur_block->InsertInst(cur_inst);
                return;
            } else if(src1->getVal() >= 4  && isPowerOf2(src1->getVal() + 1)) {
                MachineOperand* pow = new MachineOperand(MachineOperand::IMM, log2ForIntpowOf2(src1->getVal() + 1));
                auto temp = genMachineVReg();
                //MachineOperand* temp = new MachineOperand(MachineOperand::VREG, SymbolTable::getLabel());
                MachineInstruction* cur_inst1 = new BinaryMInstruction(
                            cur_block, BinaryMInstruction::LSL, temp, src2, pow);
                MachineInstruction* cur_inst2 = new BinaryMInstruction(
                            cur_block, BinaryMInstruction::SUB, dst, temp, src2);
                cur_block->InsertInst(cur_inst1);
                cur_block->InsertInst(cur_inst2);
                return;
            } else if(src1->getVal() >= 4  && isPowerOf2(src1->getVal() - 1)) {
                MachineOperand* pow = new MachineOperand(MachineOperand::IMM, log2ForIntpowOf2(src1->getVal() - 1));
                auto temp = genMachineVReg();
                //MachineOperand* temp = new MachineOperand(MachineOperand::VREG, SymbolTable::getLabel());
                MachineInstruction* cur_inst1 = new BinaryMInstruction(
                            cur_block, BinaryMInstruction::LSL, temp, src2, pow);
                MachineInstruction* cur_inst2 = new BinaryMInstruction(
                            cur_block, BinaryMInstruction::ADD, dst, temp, src2);
                cur_block->InsertInst(cur_inst1);
                cur_block->InsertInst(cur_inst2);
                return;
            }
        }
        //StengthReduction end.

        if(src1->isImm()) {
            auto temp_reg = genMachineVReg();
            MachineInstruction* temp_inst = new LoadMInstruction(
                    cur_block, LoadMInstruction::LDR, temp_reg, src1);
            cur_block->InsertInst(temp_inst);
            src1 = new MachineOperand(*temp_reg);
        }

        //StrengthReduction:
        if(src2->isImm() && opcode == MUL) {  
            if(src2->getVal() == 0) {
                MachineOperand* temp = new MachineOperand(MachineOperand::IMM, 0);
                MachineInstruction* cur_inst = new MovMInstruction(
                            cur_block, MovMInstruction::MOV, dst, temp);
                cur_block->InsertInst(cur_inst);
                return;
            } else if(src2->getVal() == 1) {
                MachineInstruction* cur_inst = new MovMInstruction(
                            cur_block, MovMInstruction::MOV, dst, src1);
                cur_block->InsertInst(cur_inst);
                return;
            } else if(src2->getVal() == 2) {
                MachineOperand* temp = new MachineOperand(*src1);
                MachineInstruction* cur_inst = new BinaryMInstruction(
                            cur_block, BinaryMInstruction::ADD, dst, src1, temp);
                cur_block->InsertInst(cur_inst);
                return;
            } else if(isPowerOf2(src2->getVal())) {
                MachineOperand* pow = new MachineOperand(MachineOperand::IMM, log2ForIntpowOf2(src2->getVal()));
                MachineInstruction* cur_inst = new BinaryMInstruction(
                            cur_block, BinaryMInstruction::LSL, dst, src1, pow);
                cur_block->InsertInst(cur_inst);
                return;
            } else if(src2->getVal() >= 4  && isPowerOf2(src2->getVal() + 1)) {
                MachineOperand* pow = new MachineOperand(MachineOperand::IMM, log2ForIntpowOf2(src2->getVal() + 1));
                auto temp = genMachineVReg();
                //MachineOperand* temp = new MachineOperand(MachineOperand::VREG, SymbolTable::getLabel());
                MachineInstruction* cur_inst1 = new BinaryMInstruction(
                            cur_block, BinaryMInstruction::LSL, temp, src1, pow);
                MachineInstruction* cur_inst2 = new BinaryMInstruction(
                            cur_block, BinaryMInstruction::SUB, dst, temp, src1);
                cur_block->InsertInst(cur_inst1);
                cur_block->InsertInst(cur_inst2);
                return;
            } else if(src2->getVal() >= 4  && isPowerOf2(src2->getVal() - 1)) {
                MachineOperand* pow = new MachineOperand(MachineOperand::IMM, log2ForIntpowOf2(src2->getVal() - 1));
                auto temp = genMachineVReg();
                //MachineOperand* temp = new MachineOperand(MachineOperand::VREG, SymbolTable::getLabel());
                MachineInstruction* cur_inst1 = new BinaryMInstruction(
                            cur_block, BinaryMInstruction::LSL, temp, src1, pow);
                MachineInstruction* cur_inst2 = new BinaryMInstruction(
                            cur_block, BinaryMInstruction::ADD, dst, temp, src1);
                cur_block->InsertInst(cur_inst1);
                cur_block->InsertInst(cur_inst2);
                return;
            }
        }
        if(src2->isImm() && opcode == DIV) {
            if(src2->getVal() == 1) {
                MachineInstruction* cur_inst = new MovMInstruction(
                            cur_block, MovMInstruction::MOV, dst, src1);
                cur_block->InsertInst(cur_inst);
                return;
            }
            if((dynamic_cast<FunctionType*>(this->getParent()->getParent()->getSymPtr()->getType()))->getParamsType().size() == 1000) {           
                if(isPowerOf2(src2->getVal())) { // && src1->isImm() && src1->getVal() > 0;
                    MachineOperand* pow = new MachineOperand(MachineOperand::IMM, log2ForIntpowOf2(src2->getVal()));
                    //Both are OK:
                    // MachineInstruction* cur_inst = new MovMInstruction(
                    //             cur_block, MovMInstruction::MOVASR, dst, src1, MachineInstruction::NONE, pow);
                    MachineInstruction* cur_inst = new BinaryMInstruction(
                                cur_block, BinaryMInstruction::ASR, dst, src1, pow);
                    cur_block->InsertInst(cur_inst);
                    return;
                } else if(!isPowerOf2(src2->getVal())) { // && src1->isImm() && src1->getVal() > 0
                    //TODO : SMULL
                    int constVal = src2->getVal();
                    int s = ctz(constVal);
                    auto off = new MachineOperand(MachineOperand::IMM, 31);    
                    auto temp = new MachineOperand(MachineOperand::VREG, SymbolTable::getLabel()); 
                    int _s = 0; 
                    if(constVal % (int(1) << s) == 0) {
                        constVal = constVal / (int(1) << s);
                        _s = s;
                    }
                    int low = floor(log2(constVal));
                    long long high = ceil((((double)((long long)(1) << (low + 32))) / constVal));
                    auto m = new MachineOperand(MachineOperand::IMM, high);
                    auto l = new MachineOperand(MachineOperand::IMM, low + _s);
                    auto rh = new MachineOperand(MachineOperand::VREG, SymbolTable::getLabel());
                    auto rl = new MachineOperand(MachineOperand::VREG, SymbolTable::getLabel());
                    MachineInstruction* inst1 = new LoadMInstruction(nullptr, LoadMInstruction::LDR, temp, m);
                    MachineInstruction* inst2 = new SmullMInstruction(nullptr, rl, rh, src1, temp);
                    MachineInstruction* inst3 = new MovMInstruction(nullptr, MovMInstruction::MOVASR,
                                        dst, rh, MachineInstruction::NONE, l);
                    MachineInstruction* inst4 = new MovMInstruction(nullptr, MovMInstruction::MOVLSR,
                                        temp, rh, MachineInstruction::NONE, off);
                    MachineInstruction* inst5 = new BinaryMInstruction(nullptr, BinaryMInstruction::ADD,
                                        dst, dst, temp);
                
                    cur_block->InsertInst(inst1);
                    cur_block->InsertInst(inst2);
                    cur_block->InsertInst(inst3);
                    cur_block->InsertInst(inst4);
                    cur_block->InsertInst(inst5);
                }
            }
        }
        if(src1->isImm() && src2->isImm() && opcode == MOD) {
            if(isPowerOf2(src2->getVal()) && (src1->getVal()>=0)) {
                MachineOperand* temp = new MachineOperand(MachineOperand::IMM, src2->getVal() - 1);
                MachineInstruction* cur_inst = new BinaryMInstruction(
                            cur_block, BinaryMInstruction::AND, dst, src1, temp);
                cur_block->InsertInst(cur_inst);
                return;
            } 
        }
        //StengthReduction end.


        if(src2->isImm() && (src2->getVal() >= 256 || src2->getVal() <= -255)) {
            auto temp_reg = genMachineVReg();
            MachineInstruction* temp_inst = new LoadMInstruction(
                    cur_block, LoadMInstruction::LDR, temp_reg, src2);
            cur_block->InsertInst(temp_inst);
            src2 = new MachineOperand(*temp_reg);
        }
        if(src2->isImm() && opcode >= MUL) {
            auto temp_reg = genMachineVReg();
            MachineInstruction* temp_inst = new LoadMInstruction(
                    cur_block, LoadMInstruction::LDR, temp_reg, src2);
            cur_block->InsertInst(temp_inst);
            src2 = new MachineOperand(*temp_reg);
        }
        MachineInstruction* cur_inst = nullptr;
        switch (opcode) {
            case ADD:
                if(src2->isImm() && src2->getVal() == 0) {
                    cur_inst = new MovMInstruction(cur_block, MovMInstruction::MOV, dst, src1);
                } else {
                    cur_inst = new BinaryMInstruction(cur_block, BinaryMInstruction::ADD, dst, src1, src2);
                }
                break;
            case SUB:
                if(src2->isImm() && src2->getVal() == 0) {
                    cur_inst = new MovMInstruction(cur_block, MovMInstruction::MOV, dst, src1);
                } else {
                    cur_inst = new BinaryMInstruction(cur_block, BinaryMInstruction::SUB, dst, src1, src2);
                }
                break;
            case AND:
                cur_inst = new BinaryMInstruction(cur_block, BinaryMInstruction::AND, dst, src1, src2);
                break;
            case OR:
                cur_inst = new BinaryMInstruction(cur_block, BinaryMInstruction::OR, dst, src1, src2);
                break;
            case MUL:
                cur_inst = new BinaryMInstruction(cur_block, BinaryMInstruction::MUL, dst, src1, src2);
                break;
            case DIV:
                cur_inst = new BinaryMInstruction(cur_block, BinaryMInstruction::DIV, dst, src1, src2);
                break;
            case MOD: {
                auto dst1 = genMachineVReg();
                cur_inst = new BinaryMInstruction(cur_block, BinaryMInstruction::DIV, dst1, src1, src2);
                src1 = new MachineOperand(*src1);
                src2 = new MachineOperand(*src2);
                auto temp = new MachineOperand(*dst1);
                cur_block->InsertInst(cur_inst);
                auto dst2 = genMachineVReg();
                cur_inst = new BinaryMInstruction(cur_block, BinaryMInstruction::MUL, dst2, temp, src2);
                cur_block->InsertInst(cur_inst);
                dst2 = new MachineOperand(*dst2);
                cur_inst = new BinaryMInstruction(cur_block, BinaryMInstruction::SUB, dst, src1, dst2);
                break;
            }
            default:
                break;
        }
        cur_block->InsertInst(cur_inst);
    } else if(operands[0]->getType()->toStr() == "float") {
        auto dst = genMachineFloatOperand(operands[0]);
        auto src1 = genMachineFloatOperand(operands[1]);
        auto src2 = genMachineFloatOperand(operands[2]);
        if(src1->isImm()) {
            auto temp_freg = genMachineVReg(true);
            auto temp_reg = genMachineVReg();
            MachineInstruction* temp_inst = new LoadMInstruction(cur_block, LoadMInstruction::LDR, temp_reg, src1);
            cur_block->InsertInst(temp_inst);
            MachineInstruction* temp_inst2 = new MovMInstruction(cur_block, MovMInstruction::VMOV, temp_freg, temp_reg);
            cur_block->InsertInst(temp_inst2);
            src1 = new MachineOperand(*temp_freg);
        }
        if(src2->isImm()) {
            auto temp_freg = genMachineVReg(true);
            auto temp_reg = genMachineVReg();
            MachineInstruction* temp_inst = new LoadMInstruction(cur_block, LoadMInstruction::LDR, temp_reg, src2);
            cur_block->InsertInst(temp_inst);
            MachineInstruction* temp_inst2 = new MovMInstruction(
            cur_block, MovMInstruction::VMOV, temp_freg, temp_reg);
            cur_block->InsertInst(temp_inst2);
            src2 = new MachineOperand(*temp_freg);
        }
        MachineInstruction* cur_inst = nullptr;
        switch (opcode) {
            case ADD:
                if(src2->isImm() && src2->getFVal() == 0) {
                    cur_inst = new MovMInstruction(cur_block, MovMInstruction::VMOVF32, dst, src1);
            // printf("3\n");printf("vmov operand1 3 = %d\n", cur_inst->getUse()[0]->getReg());
                } else {
                    cur_inst = new BinaryMInstruction(cur_block, BinaryMInstruction::VADD, dst, src1, src2);
                }
                break;
            case SUB:
                if(src2->isImm() && src2->getFVal() == 0) {
                    cur_inst = new MovMInstruction(cur_block, MovMInstruction::VMOVF32, dst, src1);
            // printf("4\n");printf("vmov operand1 4 = %d\n", cur_inst->getUse()[0]->getReg());
                } else {
                    cur_inst = new BinaryMInstruction(cur_block, BinaryMInstruction::VSUB, dst, src1, src2);
                }
                break;
            case MUL:
                cur_inst = new BinaryMInstruction(
                    cur_block, BinaryMInstruction::VMUL, dst, src1, src2);
                break;
            case DIV:
                cur_inst = new BinaryMInstruction(
                        cur_block, BinaryMInstruction::VDIV, dst, src1, src2);
                break;
            case MOD: {
                break;
            }
            case AND: {
                break;
            }
            case OR: {
                break;
            }
            default:
                break;
        }
        cur_block->InsertInst(cur_inst);
    }
}

CmpInstruction::CmpInstruction(unsigned opcode,
                               Operand* dst,
                               Operand* src1,
                               Operand* src2,
                               BasicBlock* insert_bb)
    : Instruction(CMP, insert_bb) {
    this->opcode = opcode;
    operands.push_back(dst);
    operands.push_back(src1);
    operands.push_back(src2);
    dst->setDef(this);
    src1->addUse(this);
    src2->addUse(this);
}

Instruction* CmpInstruction::copy(Instruction* ins, BasicBlock* new_bb)
{
    CmpInstruction* cmpins = dynamic_cast<CmpInstruction*>(ins);
    unsigned mopcode = cmpins->getOpcode();
    SymbolEntry* dst_se = cmpins->getOp()[0]->getSymbolEntry();
    SymbolEntry* src1_se = cmpins->getOp()[1]->getSymbolEntry();
    SymbolEntry* src2_se = cmpins->getOp()[2]->getSymbolEntry();
    return new CmpInstruction(mopcode, new Operand(dst_se), new Operand(src1_se), new Operand(src2_se), new_bb);
}

void CmpInstruction::replaceUse(Operand* old, Operand* new_) {
    if (operands[1] == old) {
        operands[1]->removeUse(this);
        operands[1] = new_;
        new_->addUse(this);
    } else if (operands[2] == old) {
        operands[2]->removeUse(this);
        operands[2] = new_;
        new_->addUse(this);
    }
}

CmpInstruction::~CmpInstruction() {
    operands[0]->setDef(nullptr);
    if (operands[0]->usersNum() == 0)
        delete operands[0];
    operands[1]->removeUse(this);
    operands[2]->removeUse(this);
}

void CmpInstruction::output() const {
    std::string s1, s2, s3, op;
    s1 = operands[0]->toStr();
    s2 = operands[1]->toStr();
    s3 = operands[2]->toStr();
    Type* type = operands[1]->getType();
    switch (opcode) {
        case E:
            op = "eq";
            break;
        case NE:
            op = "ne";
            break;
        case L:
            op = "slt";
            break;
        case LE:
            op = "sle";
            break;
        case G:
            op = "sgt";
            break;
        case GE:
            op = "sge";
            break;
        default:
            op = "";
            break;
    }

    fprintf(yyout, "  %s = icmp %s %s %s, %s\n", s1.c_str(), op.c_str(),
            type->toStr().c_str(), s2.c_str(), s3.c_str());
}

void CmpInstruction::genMachineCode(AsmBuilder *builder) {
    auto cur_block = builder->getBlock();

    if (operands[1]->getType()->isFloat()) {
        auto src1 = genMachineFloatOperand(operands[1]);
        auto src2 = genMachineFloatOperand(operands[2]);
        MachineInstruction *cur_inst = nullptr;
        if (src1->isImm()) {
            auto tmp_reg = genMachineVReg(true);
            auto internal_reg = genMachineVReg();
            cur_inst = new LoadMInstruction(cur_block, LoadMInstruction::LDR,
                                            internal_reg, src1);
            cur_block->InsertInst(cur_inst);
            internal_reg = new MachineOperand(*internal_reg);
            cur_inst = new MovMInstruction(cur_block, MovMInstruction::VMOV,
                                           tmp_reg, internal_reg);
            // printf("5\n");printf("vmov operand1 5 = %d\n",
            // cur_inst->getUse()[0]->getReg());
            cur_block->InsertInst(cur_inst);
            src1 = new MachineOperand(*tmp_reg);
        }
        if (src2->isImm()) {
            auto tmp_reg = genMachineVReg(true);
            auto internal_reg = genMachineVReg();
            cur_inst = new LoadMInstruction(cur_block, LoadMInstruction::LDR,
                                            internal_reg, src2);
            cur_block->InsertInst(cur_inst);
            internal_reg = new MachineOperand(*internal_reg);
            cur_inst = new MovMInstruction(cur_block, MovMInstruction::VMOV,
                                           tmp_reg, internal_reg);
            // printf("6\n");printf("vmov operand1 6 = %d\n",
            // cur_inst->getUse()[0]->getReg());
            cur_block->InsertInst(cur_inst);
            src2 = new MachineOperand(*tmp_reg);
        }
        cur_inst = new CmpMInstruction(cur_block, CmpMInstruction::VCMP, src1,
                                       src2, opcode);
        cur_block->InsertInst(cur_inst);
        cur_inst = new VmrsMInstruction(cur_block);
        cur_block->InsertInst(cur_inst);

        if (opcode >= CmpInstruction::L && opcode <= CmpInstruction::GE) {
            auto dst = genMachineOperand(operands[0]);
            auto trueOperand = genMachineImm(1);
            auto falseOperand = genMachineImm(0);
            cur_inst = new MovMInstruction(cur_block, MovMInstruction::MOV, dst,
                                           trueOperand, opcode);
            cur_block->InsertInst(cur_inst);
            cur_inst = new MovMInstruction(cur_block, MovMInstruction::MOV, dst,
                                           falseOperand, 7 - opcode);
            cur_block->InsertInst(cur_inst);
        } else if (opcode == CmpInstruction::E) {
            auto dst = genMachineOperand(operands[0]);
            auto trueOperand = genMachineImm(1);
            auto falseOperand = genMachineImm(0);
            cur_inst = new MovMInstruction(cur_block, MovMInstruction::MOV, dst,
                                           trueOperand, E);
            cur_block->InsertInst(cur_inst);
            cur_inst = new MovMInstruction(cur_block, MovMInstruction::MOV, dst,
                                           falseOperand, NE);
            cur_block->InsertInst(cur_inst);

        } else if (opcode == CmpInstruction::NE) {
            auto dst = genMachineOperand(operands[0]);
            auto trueOperand = genMachineImm(1);
            auto falseOperand = genMachineImm(0);
            cur_inst = new MovMInstruction(cur_block, MovMInstruction::MOV, dst,
                                           trueOperand, NE);
            cur_block->InsertInst(cur_inst);
            cur_inst = new MovMInstruction(cur_block, MovMInstruction::MOV, dst,
                                           falseOperand, E);
            cur_block->InsertInst(cur_inst);
        }

    } else {
        auto src1 = genMachineOperand(operands[1]);
        auto src2 = genMachineOperand(operands[2]);
        MachineInstruction *cur_inst = nullptr;
        // wzr-multi:
        if(addMov)
        {
            cur_inst = new MovMInstruction(cur_block, MovMInstruction::MOV, src1, 
                genMachineReg(10));
            cur_block->InsertInst(cur_inst);
        }
        if (src1->isImm()) {
            auto internal_reg = genMachineVReg();
            cur_inst = new LoadMInstruction(cur_block, LoadMInstruction::LDR,
                                            internal_reg, src1);
            cur_block->InsertInst(cur_inst);
            src1 = new MachineOperand(*internal_reg);
        }
        if (src2->isImm() &&
            ((ConstantSymbolEntry *)(operands[2]->getEntry()))->getValue() >
                255) {
            auto internal_reg = genMachineVReg();
            cur_inst = new LoadMInstruction(cur_block, LoadMInstruction::LDR,
                                            internal_reg, src2);
            cur_block->InsertInst(cur_inst);
            src2 = new MachineOperand(*internal_reg);
        }
        cur_inst = new CmpMInstruction(cur_block, CmpMInstruction::CMP, src1,
                                       src2, opcode);
        cur_block->InsertInst(cur_inst);
        if (opcode >= CmpInstruction::L && opcode <= CmpInstruction::GE) {
            auto dst = genMachineOperand(operands[0]);
            auto trueOperand = genMachineImm(1);
            auto falseOperand = genMachineImm(0);
            cur_inst = new MovMInstruction(cur_block, MovMInstruction::MOV, dst,
                                           trueOperand, opcode);
            cur_block->InsertInst(cur_inst);
            cur_inst = new MovMInstruction(cur_block, MovMInstruction::MOV, dst,
                                           falseOperand, 7 - opcode);
            cur_block->InsertInst(cur_inst);
        } else if (opcode == CmpInstruction::E) {
            auto dst = genMachineOperand(operands[0]);
            auto trueOperand = genMachineImm(1);
            auto falseOperand = genMachineImm(0);
            cur_inst = new MovMInstruction(cur_block, MovMInstruction::MOV, dst,
                                           trueOperand, E);
            cur_block->InsertInst(cur_inst);
            cur_inst = new MovMInstruction(cur_block, MovMInstruction::MOV, dst,
                                           falseOperand, NE);
            cur_block->InsertInst(cur_inst);

        } else if (opcode == CmpInstruction::NE) {
            auto dst = genMachineOperand(operands[0]);
            auto trueOperand = genMachineImm(1);
            auto falseOperand = genMachineImm(0);
            cur_inst = new MovMInstruction(cur_block, MovMInstruction::MOV, dst,
                                           trueOperand, NE);
            cur_block->InsertInst(cur_inst);
            cur_inst = new MovMInstruction(cur_block, MovMInstruction::MOV, dst,
                                           falseOperand, E);
            cur_block->InsertInst(cur_inst);
        }
    }
}

UncondBrInstruction::UncondBrInstruction(BasicBlock* to, BasicBlock* insert_bb)
    : Instruction(UNCOND, insert_bb) {
    branch = to;
}

Instruction* UncondBrInstruction::copy(Instruction* ins, BasicBlock* new_bb)
{
    UncondBrInstruction* uncondbrins = dynamic_cast<UncondBrInstruction*>(ins);
    BasicBlock* dst_bb = uncondbrins->getBranch();
    return new UncondBrInstruction(dst_bb, new_bb);
}

void UncondBrInstruction::output() const {
    fprintf(yyout, "  br label %%B%d\n", branch->getNo());
}

void UncondBrInstruction::genMachineCode(AsmBuilder* builder) {
    std::stringstream str;
    str << ".L" << branch->getNo();
    MachineOperand* dst = new MachineOperand(str.str());

    auto cur_block = builder->getBlock();
    auto cur_inst = new BranchMInstruction(cur_block, BranchMInstruction::B, dst);
    cur_block->InsertInst(cur_inst);
}

void UncondBrInstruction::setBranch(BasicBlock* bb) {
    branch = bb;
}

BasicBlock* UncondBrInstruction::getBranch() {
    return branch;
}

CondBrInstruction::CondBrInstruction(BasicBlock* true_branch,
                                     BasicBlock* false_branch,
                                     Operand* cond,
                                     BasicBlock* insert_bb)
    : Instruction(COND, insert_bb) {
    this->true_branch = true_branch;
    this->false_branch = false_branch;
    cond->addUse(this);
    this->cond = cond;
    operands.push_back(cond);
    //std::cout<<"new condbr, this = "<<this<<"\n";
}

Instruction* CondBrInstruction::copy(Instruction* ins, BasicBlock* new_bb)
{
    CondBrInstruction* condbrins = dynamic_cast<CondBrInstruction*>(ins);
    BasicBlock* dst_truebb = condbrins->getTrueBranch();
    BasicBlock* dst_falsebb = condbrins->getFalseBranch();
    SymbolEntry* se = condbrins->getOp()[0]->getSymbolEntry();
    return new CondBrInstruction(dst_truebb, dst_falsebb, new Operand(se), new_bb);
}

void CondBrInstruction::replaceUse(Operand* old, Operand* new_) {
    if (operands[0] == old) {
        operands[0]->removeUse(this);
        operands[0] = new_;
        new_->addUse(this);
    }
}

CondBrInstruction::~CondBrInstruction() {
    operands[0]->removeUse(this);
}

void CondBrInstruction::output() const {
    std::string cond = operands[0]->toStr();
    int true_label = true_branch->getNo();
    int false_label = false_branch->getNo();
    Type* type = operands[0]->getType();
    fprintf(yyout, "  br %s %s, label %%B%d, label %%B%d\n", type->toStr().c_str(),
            cond.c_str(), true_label, false_label);
}

void CondBrInstruction::genMachineCode(AsmBuilder* builder) {
    std::stringstream str1;
    str1 << ".L" << true_branch->getNo();
    MachineOperand* dst = new MachineOperand(str1.str());
    auto cur_block = builder->getBlock();
    auto cur_inst = new BranchMInstruction(cur_block, BranchMInstruction::B, dst, cur_block->getCmpCond());
    cur_block->InsertInst(cur_inst);

    std::stringstream str2;
    str2.str("");
    str2 << ".L" << false_branch->getNo();
    dst = new MachineOperand(str2.str());
    cur_block = builder->getBlock();
    cur_inst = new BranchMInstruction(cur_block, BranchMInstruction::B, dst);
    cur_block->InsertInst(cur_inst);
}

void CondBrInstruction::setFalseBranch(BasicBlock* bb) {
    false_branch = bb;
}

BasicBlock* CondBrInstruction::getFalseBranch() {
    return false_branch;
}

void CondBrInstruction::setTrueBranch(BasicBlock* bb) {
    true_branch = bb;
}

BasicBlock* CondBrInstruction::getTrueBranch() {
    return true_branch;
}

RetInstruction::RetInstruction(Operand* src, BasicBlock* insert_bb)
    : Instruction(RET, insert_bb) {
    if (src != nullptr) {
        operands.push_back(src);
        // std::cout<<"return"<<src->toStr()<<std::endl;
        src->addUse(this);
    }
}

Instruction* RetInstruction::copy(Instruction* ins, BasicBlock* new_bb)
{
    RetInstruction* retins = dynamic_cast<RetInstruction*>(ins);
    SymbolEntry* se = retins->getOp()[0]->getSymbolEntry();
    return new RetInstruction(new Operand(se), new_bb);
}

void RetInstruction::replaceUse(Operand* old, Operand* new_) {
    if (operands.size() && operands[0] == old) {
        operands[0]->removeUse(this);
        operands[0] = new_;
        new_->addUse(this);
        // std::cout<<"returtru"<<new_->toStr()<<std::endl;
    }
}

RetInstruction::~RetInstruction() {
    if (!operands.empty())
        operands[0]->removeUse(this);
}

void RetInstruction::output() const {
    if (operands.empty()) {
        fprintf(yyout, "  ret void\n");
    } else {
        std::string ret = operands[0]->toStr();
        Type* type = operands[0]->getType();
        fprintf(yyout, "  ret %s %s\n", type->toStr().c_str(), ret.c_str());
    }
}

void RetInstruction::genMachineCode(AsmBuilder* builder) {
    auto cur_block = builder->getBlock();
    if (!operands.empty()) {
        MachineOperand* dst;
        MachineOperand* src;
        MachineInstruction* cur_inst;

        if (operands[0]->getType()->toStr() == "float") {
            src = genMachineFloatOperand(operands[0]);
            dst = new MachineOperand(MachineOperand::REG, 16, true);
            if (src->isImm()) {
                auto temp_reg = genMachineVReg();
                cur_inst = new LoadMInstruction(cur_block, LoadMInstruction::LDR, temp_reg, src);
                cur_block->InsertInst(cur_inst);
                src = temp_reg;
            }
            cur_inst = new MovMInstruction(cur_block, MovMInstruction::VMOV, dst, src);
            // printf("7\n");printf("vmov operand1 7 = %d\n", cur_inst->getUse()[0]->getReg());
        } else {
            src = genMachineOperand(operands[0]);
            dst = new MachineOperand(MachineOperand::REG, 0);
            if (operands[0]->isConst()) {
                auto val = operands[0]->getConstVal();
                if (val > 255 || val <= -255) {
                    auto temp_Op = new MachineOperand(MachineOperand::REG, 0);
                    cur_block->InsertInst(new LoadMInstruction(cur_block, LoadMInstruction::LDR, temp_Op, src));
                    src = temp_Op;
                }
            }
            cur_inst = new MovMInstruction(cur_block, MovMInstruction::MOV, dst, src);
        }
        cur_block->InsertInst(cur_inst);
    }
    auto temp_Op = new MachineOperand(MachineOperand::REG, 13);

    auto cur_func = builder->getFunction();
    auto size = new MachineOperand(MachineOperand::IMM, cur_func->AllocSpace(0));

    auto cur_inst = new BinaryMInstruction(cur_block, BinaryMInstruction::ADD, temp_Op, temp_Op, size);
    cur_block->InsertInst(cur_inst);

    auto temp_Op_ = new MachineOperand(MachineOperand::REG, 14);
    auto cur_inst2 = new BranchMInstruction(cur_block, BranchMInstruction::BX, temp_Op_);
    cur_block->InsertInst(cur_inst2);
}

AllocaInstruction::AllocaInstruction(Operand* dst,
                                     SymbolEntry* se,
                                     BasicBlock* insert_bb)
    : Instruction(ALLOCA, insert_bb) {
    operands.push_back(dst);
    dst->setDef(this);
    this->se = se;
}

Instruction* AllocaInstruction::copy(Instruction* ins, BasicBlock* new_bb)
{
    AllocaInstruction* allocains = dynamic_cast<AllocaInstruction*>(ins);
    SymbolEntry* opSE = allocains->getOp()[0]->getSymbolEntry();
    SymbolEntry* se = allocains->getSymbolEntry();
    return new AllocaInstruction(new Operand(opSE), se, new_bb);
}

AllocaInstruction::~AllocaInstruction() {
    operands[0]->setDef(nullptr);
    if (operands[0]->usersNum() == 0)
        delete operands[0];
}

void AllocaInstruction::output() const {
    std::string dst = operands[0]->toStr();
    Type* type;
    if (se->getType()->isInt() || se->getType()->toStr() == "float") {
        type = se->getType();
        fprintf(yyout, "  %s = alloca %s, align 4\n", dst.c_str(), type->toStr().c_str());
    } else if (se->getType()->isArray()) {
        type = se->getType();
        // type = operands[0]->getSymbolEntry()->getType()->toStr();
        fprintf(yyout, "  %s = alloca %s, align 4\n", dst.c_str(), type->toStr().c_str());
    }
}

void AllocaInstruction::genMachineCode(AsmBuilder* builder) {
    auto cur_func = builder->getFunction();
    long long size = se->getType()->getSize();
    if (size <= 0) {
        size = 32;
    }
    int offset = 0 - cur_func->AllocSpace(size / 8);
    dynamic_cast<TemporarySymbolEntry*>(operands[0]->getEntry())->setOffset(offset);
}

LoadInstruction::LoadInstruction(Operand* dst,
                                 Operand* src_addr,
                                 BasicBlock* insert_bb)
    : Instruction(LOAD, insert_bb) {
    operands.push_back(dst);
    operands.push_back(src_addr);
    dst->setDef(this);
    src_addr->addUse(this);
}

Instruction* LoadInstruction::copy(Instruction* ins, BasicBlock* new_bb)
{
    LoadInstruction* loadins = dynamic_cast<LoadInstruction*>(ins);
    SymbolEntry* dst_se = loadins->getOp()[0]->getSymbolEntry();
    SymbolEntry* src_se = loadins->getOp()[1]->getSymbolEntry();
    return new LoadInstruction(new Operand(dst_se), new Operand(src_se), new_bb);
}

void LoadInstruction::replaceUse(Operand* old, Operand* new_) {
    if (operands[1] == old) {
        operands[1]->removeUse(this);
        operands[1] = new_;
        new_->addUse(this);
    }
}

LoadInstruction::~LoadInstruction() {
    operands[0]->setDef(nullptr);
    if (operands[0]->usersNum() == 0)
        delete operands[0];
    operands[1]->removeUse(this);
}

void LoadInstruction::output() const {
    std::string dst = operands[0]->toStr();
    std::string src = operands[1]->toStr();
    Type* dst_type = operands[0]->getType();
    Type* src_type = operands[1]->getType();
    fprintf(yyout, "  %s = load %s, %s %s, align 4\n", dst.c_str(),
            dst_type->toStr().c_str(), src_type->toStr().c_str(), src.c_str());
}

void LoadInstruction::genMachineCode(AsmBuilder* builder) {
    auto cur_block = builder->getBlock();
    MachineInstruction* cur_inst = nullptr;
    if (operands[1]->getEntry()->isVariable() 
        && dynamic_cast<IdentifierSymbolEntry*>(operands[1]->getEntry())->isGlobal()) {

        auto src = genMachineOperand(operands[1]);
        auto internal_reg1 = genMachineVReg();
        auto internal_reg2 = new MachineOperand(*internal_reg1);
        if (operands[0]->getType()->toStr() == "float") {
            auto dst = genMachineFloatOperand(operands[0]);

            cur_inst = new LoadMInstruction(cur_block, LoadMInstruction::LDR, internal_reg1, src);
            cur_block->InsertInst(cur_inst);

            cur_inst = new LoadMInstruction(cur_block, LoadMInstruction::VLDR, dst, internal_reg2);
            cur_block->InsertInst(cur_inst);
        } else {
            auto dst = genMachineOperand(operands[0]);

            cur_inst = new LoadMInstruction(cur_block, LoadMInstruction::LDR, internal_reg1, src);
            cur_block->InsertInst(cur_inst);

            cur_inst = new LoadMInstruction(cur_block, LoadMInstruction::LDR, dst, internal_reg2);
            cur_inst->setisVectorization(getAsmType());
            cur_block->InsertInst(cur_inst);
        }
    } else if (operands[1]->getEntry()->isTemporary() && operands[1]->getDef() && operands[1]->getDef()->isAlloc()) {
        if (operands[0]->getType()->toStr() == "float") {
            auto dst = genMachineFloatOperand(operands[0]);
            auto src1 = genMachineReg(11);
            int offset = dynamic_cast<TemporarySymbolEntry*>(operands[1]->getEntry()) ->getOffset();
            auto src2 = genMachineImm(offset);
            if (offset > 255 || offset < -255) {
                auto operand1 = genMachineVReg();
                cur_block->InsertInst((new LoadMInstruction(cur_block, LoadMInstruction::LDR, operand1, src2)));
                src2 = new MachineOperand(*operand1);

                auto operand2 = genMachineVReg();
                cur_block->InsertInst((new BinaryMInstruction(cur_block, BinaryMInstruction::ADD, operand2, src1, src2)));
                cur_inst = new LoadMInstruction(cur_block, LoadMInstruction::VLDR, dst, new MachineOperand(*operand2));
            } else {
                cur_inst = new LoadMInstruction(cur_block, LoadMInstruction::VLDR, dst, src1, src2);
            }
            cur_block->InsertInst(cur_inst);
        } else {
            int offset = dynamic_cast<TemporarySymbolEntry*>(operands[1]->getEntry())->getOffset();
            auto src1 = genMachineReg(11);
            auto src2 = genMachineImm(offset);
            if (offset > 255 || offset < -255) {
                auto operand = genMachineVReg();
                cur_block->InsertInst((new LoadMInstruction(cur_block, LoadMInstruction::LDR, operand, src2)));
                src2 = operand;
            }
            auto dst = genMachineOperand(operands[0]);
            cur_inst = new LoadMInstruction(cur_block, LoadMInstruction::LDR, dst, src1, src2);
            cur_inst->setisVectorization(getAsmType());
            cur_block->InsertInst(cur_inst);
        }
    } else { 
        MachineOperand* dst = nullptr;
        MachineOperand* src = nullptr;
        if (operands[0]->getType()->toStr() == "float") {
            dst = genMachineFloatOperand(operands[0]);
        } else {
            dst = genMachineOperand(operands[0]);
        }

        src = genMachineOperand(operands[1]);

        if (operands[0]->getType()->toStr() == "float" || operands[1]->getType()->toStr() == "float") {
            cur_inst = new LoadMInstruction(cur_block, LoadMInstruction::VLDR, dst, src);
            cur_block->InsertInst(cur_inst);
        } else {
            cur_inst = new LoadMInstruction(cur_block, LoadMInstruction::LDR, dst, src);
            cur_inst->setisVectorization(getAsmType());
            cur_block->InsertInst(cur_inst);
        }
    }
}

StoreInstruction::StoreInstruction(Operand* dst_addr,
                                   Operand* src,
                                   BasicBlock* insert_bb,bool assign)
    : Instruction(STORE, insert_bb) {
    operands.push_back(dst_addr);
    operands.push_back(src);
    dst_addr->addUse(this);
    src->addUse(this);
    if(!assign){
    this->push_back_opse(dst_addr->getSymbolEntry());
    // std::cout<<"=============== "<<dst_addr->getSymbolEntry()->toStr()<<"\n";
    // this->setIdentse((IdentifierSymbolEntry*)dst_addr->getSymbolEntry());
    }

}

void StoreInstruction::replaceUse(Operand* old, Operand* new_) {
    if (operands[0] == old) {
        operands[0]->removeUse(this);
        operands[0] = new_;
        new_->addUse(this);
    } else if (operands[1] == old) {
        operands[1]->removeUse(this);
        operands[1] = new_;
        new_->addUse(this);
    }
    //std::cout << "hhhhhh: " << new_->toStr() << std::endl;
}

Instruction* StoreInstruction::copy(Instruction* ins, BasicBlock* new_bb)
{
    StoreInstruction* storeins = dynamic_cast<StoreInstruction*>(ins);
    SymbolEntry* dst_se = storeins->getOp()[0]->getSymbolEntry();
    SymbolEntry* src_se = storeins->getOp()[1]->getSymbolEntry();
    Operand* dst = new Operand(dst_se);
    Operand* src = new Operand(src_se);
    src->setDst(dst);
    return new StoreInstruction(dst, src, new_bb);
}

StoreInstruction::~StoreInstruction() {
    operands[0]->removeUse(this);
    operands[1]->removeUse(this);
}

void StoreInstruction::output() const {
    std::string dst = operands[0]->toStr();
    std::string src = operands[1]->toStr();
    Type* dst_type = operands[0]->getType();
    Type* src_type = operands[1]->getType();

    fprintf(yyout, "  store %s %s, %s %s, align 4\n", src_type->toStr().c_str(),
            src.c_str(), dst_type->toStr().c_str(), dst.c_str());
}

void StoreInstruction::genMachineCode(AsmBuilder* builder) {
    auto cur_block = builder->getBlock();

    MachineOperand* dst = genMachineOperand(operands[0]);;
    MachineOperand* src = nullptr;
    if (operands[1]->getType()->toStr() == "float") {
        src = genMachineFloatOperand(operands[1]);
    } else {
        src = genMachineOperand(operands[1]);
    }

    MachineInstruction* cur_inst = nullptr;
    if (operands[1]->getEntry()->isConstant()) {
        auto dst1 = genMachineVReg(operands[1]->getType()->toStr() == "float");
        if (operands[1]->getType()->toStr() == "float") {
            auto internal_reg = genMachineVReg();
            cur_inst = new LoadMInstruction(cur_block, LoadMInstruction::LDR,  internal_reg, src);
            cur_block->InsertInst(cur_inst);
            
            internal_reg = new MachineOperand(*internal_reg);
            cur_inst = new MovMInstruction(cur_block, MovMInstruction::VMOV, dst1, internal_reg);
            cur_block->InsertInst(cur_inst);
            // printf("8\n");printf("vmov operand1 8 = %d\n", cur_inst->getUse()[0]->getReg());
        } else {
            cur_inst = new LoadMInstruction(cur_block, LoadMInstruction::LDR, dst1, src);
            cur_inst->setisVectorization(getAsmType());
            cur_block->InsertInst(cur_inst);
        }
        src = new MachineOperand(*dst1);
    }

    if (operands[0]->getEntry()->isTemporary() && operands[0]->getDef() && operands[0]->getDef()->isAlloc()) {
        auto src1 = genMachineReg(11);
        int offset = dynamic_cast<TemporarySymbolEntry*>(operands[0]->getEntry())->getOffset();
        auto src2 = genMachineImm(offset);
        if (offset > 255 || offset < -255) {
            auto operand = genMachineVReg();
            MachineInstruction* inst = new LoadMInstruction(cur_block, LoadMInstruction::LDR, operand, src2);
            cur_block->InsertInst(inst);
            inst->setisVectorization(getAsmType());
            src2 = operand;
        }
        if (operands[1]->getType()->toStr() == "float") {
            if (offset > 255 || offset < -255) {
                auto reg = genMachineVReg();
                cur_inst = new BinaryMInstruction(cur_block, BinaryMInstruction::ADD, reg, src1, src2);
                cur_block->InsertInst(cur_inst);
                // printf("new vstr3\n");
                cur_inst = new StoreMInstruction(cur_block, StoreMInstruction::VSTR, src, new MachineOperand(*reg));
            } else {
                // printf("new vstr4\n");
                cur_inst = new StoreMInstruction(cur_block, StoreMInstruction::VSTR, src, src1, src2);
            }
        } else {
            cur_inst = new StoreMInstruction(cur_block, StoreMInstruction::STR, src, src1, src2);
            cur_inst->setisVectorization(getAsmType());
        }
        cur_block->InsertInst(cur_inst);
    } else if (operands[0]->getEntry()->isVariable() && dynamic_cast<IdentifierSymbolEntry*>(operands[0]->getEntry())->isGlobal()) {
        auto internal_reg1 = genMachineVReg();
        if (operands[1]->getType()->toStr() == "float") {
            cur_inst = new LoadMInstruction(cur_block, LoadMInstruction::LDR, internal_reg1, dst);
            cur_block->InsertInst(cur_inst);
            // printf("new vstr5\n");
            cur_inst = new StoreMInstruction(cur_block, StoreMInstruction::VSTR, src, internal_reg1);
            cur_block->InsertInst(cur_inst);
        } else {
            cur_inst = new LoadMInstruction(cur_block, LoadMInstruction::LDR, internal_reg1, dst);
            cur_inst->setisVectorization(getAsmType());
            cur_block->InsertInst(cur_inst);
            cur_inst = new StoreMInstruction(cur_block, StoreMInstruction::STR, src, internal_reg1);
            cur_inst->setisVectorization(getAsmType());
            cur_block->InsertInst(cur_inst);
        }
    } else if (operands[0]->getType()->isPtr()) {
        if (operands[1]->getType()->toStr() == "float") {
                    // printf("new vstr6\n");
            cur_inst = new StoreMInstruction(cur_block, StoreMInstruction::VSTR, src, dst);
            cur_block->InsertInst(cur_inst);
        } else {
            cur_inst = new StoreMInstruction(cur_block, StoreMInstruction::STR, src, dst);
            cur_inst->setisVectorization(getAsmType());
            cur_block->InsertInst(cur_inst);
        }
    }
}


MachineOperand* Instruction::genMachineOperand(Operand* ope) {
    MachineOperand* mope = nullptr;
    auto se = ope->getEntry();

    IdentifierSymbolEntry* id_se;
    if(se->isConstant()) {
        mope = new MachineOperand(MachineOperand::IMM, (int)dynamic_cast<ConstantSymbolEntry*>(se)->getValue());
    } else if(se->isTemporary()) {
        mope = new MachineOperand(MachineOperand::VREG, dynamic_cast<TemporarySymbolEntry*>(se)->getLabel());
    } else if(se->isVariable()) {
        id_se = dynamic_cast<IdentifierSymbolEntry*>(se);
        if (id_se->isGlobal()) {
            mope = new MachineOperand(id_se->toStr().c_str());
        } else if (id_se->isParam()) {
            if (id_se->getParamNo() < 4) {
                mope = new MachineOperand(MachineOperand::REG, id_se->getParamNo());
            } else {
                mope = new MachineOperand(MachineOperand::REG, 3);
                mope->setParam();
                mope->setParamNo(id_se->getParamNo());
            }
            mope->setAllParamNo(id_se->getStackParamNo());
        } else {
            exit(0);
        }
    }
    return mope;
}
MachineOperand* Instruction::genMachineFloatOperand(Operand* ope) {
    MachineOperand* mope = nullptr;
    auto se = ope->getEntry();

    if (!(se->getType()->toStr() == "float")) {
        return genMachineOperand(ope);
    }

    IdentifierSymbolEntry* id_se;  
    if(se->isConstant()) {
        mope = new MachineOperand(MachineOperand::IMM, (float)dynamic_cast<ConstantSymbolEntry*>(se)->getValue());
    } else if(se->isTemporary()) {
        mope = new MachineOperand(MachineOperand::VREG, dynamic_cast<TemporarySymbolEntry*>(se)->getLabel(), true);
    } else if(se->isVariable()) {
        id_se = dynamic_cast<IdentifierSymbolEntry*>(se);
        if (id_se->isGlobal()) {
            mope = new MachineOperand(id_se->toStr().c_str());
        } else if (id_se->isParam()) {
            if (id_se->getParamNo() < 4) {
                mope = new MachineOperand(MachineOperand::REG, id_se->getParamNo() + 16, true);
            } else {
                mope = new MachineOperand(MachineOperand::REG, 20, true);
                mope->setParam();
                mope->setParamNo(id_se->getParamNo());
            }
            mope->setAllParamNo(id_se->getStackParamNo());
        }
    }
    return mope;
}

MachineOperand* Instruction::genMachineReg(int reg) {
    return new MachineOperand(MachineOperand::REG, reg);
}

MachineOperand* Instruction::genMachineFReg(int freg) {
    return new MachineOperand(MachineOperand::REG, freg + 16, true);
}

MachineOperand* Instruction::genMachineVReg(bool fpu) {
    return new MachineOperand(MachineOperand::VREG, SymbolTable::getLabel(), fpu);
}

MachineOperand* Instruction::genMachineImm(int val) {
    return new MachineOperand(MachineOperand::IMM, val);
}

MachineOperand* Instruction::genMachineLabel(int block_no) {
    std::ostringstream buf;
    buf << ".L" << block_no;
    std::string label = buf.str();
    return new MachineOperand(label);
}


CallInstruction::CallInstruction(Operand* dst,
                                 SymbolEntry* func,
                                 std::vector<Operand*> params,
                                 BasicBlock* insert_bb)
    : Instruction(CALL, insert_bb), func(func), dst(dst) {
    operands.push_back(dst);
    if (dst)
        dst->setDef(this);
    for (auto param : params) {
        operands.push_back(param);
        param->addUse(this);
        //j0425添加传参顺序，方便内联：

    }
}

Instruction* CallInstruction::copy(Instruction* ins, BasicBlock* new_bb)
{
    CallInstruction* callins = dynamic_cast<CallInstruction*>(ins);
    SymbolEntry* func_se = callins->getFuncSymbolEntry();
    std::vector<Operand*> ps = callins->getParams();
    if(callins->getOp()[0]==nullptr){
        Instruction* res = new CallInstruction(nullptr, func_se, ps, new_bb);
        return res;
    }
    SymbolEntry* dst_se = callins->getOp()[0]->getSymbolEntry();
    Instruction* res = new CallInstruction(new Operand(dst_se), func_se, ps, new_bb);
    return res;
}

void CallInstruction::replaceUse(Operand* old, Operand* new_) {
    for (int i = 1; i < (int)operands.size(); i++)
        if (operands[i] == old) {
            operands[i]->removeUse(this);
            operands[i] = new_;
            new_->addUse(this);
        }
}

void CallInstruction::output() const {
    fprintf(yyout, "  ");
    if (operands[0]) {
        fprintf(yyout, "%s = ", operands[0]->toStr().c_str());
    }
    FunctionType* type = (FunctionType*)(func->getType());
    fprintf(yyout, "call %s %s(", type->getRetType()->toStr().c_str(),
            func->toStr().c_str());
    for (long unsigned int i = 1; i < operands.size(); i++) {
        if (i != 1) fprintf(yyout, ", ");
        fprintf(yyout, "%s %s", operands[i]->getType()->toStr().c_str(),
                operands[i]->toStr().c_str());
    }
    fprintf(yyout, ")\n");
}
BitcastInstruction::BitcastInstruction(Operand* dst,
                                       Operand* src,
                                       BasicBlock* insert_bb)
    : Instruction(BITCAST, insert_bb), dst(dst), src(src) {
    operands.push_back(dst);
    operands.push_back(src);
    dst->setDef(this);
    src->addUse(this);
    flag = false;
}

Instruction* BitcastInstruction::copy(Instruction* ins, BasicBlock* new_bb)
{
    BitcastInstruction* bitcastins = dynamic_cast<BitcastInstruction*>(ins);
    SymbolEntry* dst_se = bitcastins->getOp()[0]->getSymbolEntry();
    SymbolEntry* src_se = bitcastins->getOp()[1]->getSymbolEntry();
    return new BitcastInstruction(new Operand(dst_se), new Operand(src_se), new_bb);
}

void BitcastInstruction::replaceUse(Operand* old, Operand* new_) {
    if (operands[1] == old) {
        operands[1]->removeUse(this);
        operands[1] = new_;
        new_->addUse(this);
    }
}

void BitcastInstruction::output() const {
    Operand* dst = operands[0];
    Operand* src = operands[1];
    fprintf(yyout, "  %s = bitcast %s %s to %s\n", dst->toStr().c_str(),
            src->getType()->toStr().c_str(), src->toStr().c_str(),
            dst->getType()->toStr().c_str());
}

BitcastInstruction::~BitcastInstruction() {
    operands[0]->setDef(nullptr);
    if (operands[0]->usersNum() == 0)
        delete operands[0];
    operands[1]->removeUse(this);
}

void BitcastInstruction::genMachineCode(AsmBuilder* builder) {
    auto block = builder->getBlock();
    auto type = ((PointerType*)(dst->getType()))->getType();
    if (!(type->isInt() && type->getSize() == 8)) {
        auto dst = genMachineOperand(operands[0]);
        auto src = genMachineOperand(operands[1]);
        auto zero = genMachineImm(0);
        auto in = new BinaryMInstruction(block, BinaryMInstruction::ADD, dst,src, zero);
        block->InsertInst(in);
    }
}

void CallInstruction::genMachineCode(AsmBuilder* builder) {
    auto cur_block = builder->getBlock();
    dst = this->getDef();
    MachineOperand* operand; 
    MachineInstruction* cur_inst;
    size_t idx;
    auto funcSE = (IdentifierSymbolEntry*)func;
    if (funcSE->getName() == "llvm.memset.p0.i32") {
        auto r0 = genMachineReg(0);
        auto r1 = genMachineReg(1);
        auto r2 = genMachineReg(2);
        if (!((BitcastInstruction*)(operands[1]->getDef()))->getFlag()) {
            int offset = ((TemporarySymbolEntry*)(((BitcastInstruction*)(operands[1]->getDef()))->getUse()[0]->getEntry()))->getOffset();
            auto fp = genMachineReg(11);
            if (offset > -255 && offset < 255) {
                cur_block->InsertInst(new BinaryMInstruction(cur_block, BinaryMInstruction::ADD, r0, fp, genMachineImm(offset)));
            } else {
                operand = genMachineVReg();
                cur_inst = new LoadMInstruction(cur_block, LoadMInstruction::LDR, operand, genMachineImm(offset));
                operand = new MachineOperand(*operand);
                cur_block->InsertInst(cur_inst);
                cur_block->InsertInst(new BinaryMInstruction(cur_block, BinaryMInstruction::ADD, r0, fp, operand));
            }
        } else {
            cur_block->InsertInst(new MovMInstruction(cur_block, MovMInstruction::MOV, r0, genMachineOperand(((BitcastInstruction*)(operands[1]->getDef()))->getUse()[0])));
        }
        cur_block->InsertInst(new MovMInstruction(cur_block, MovMInstruction::MOV, r1, genMachineImm(0)));

        operand = genMachineOperand(operands[3]);
        if (genMachineOperand(operands[3])->isImm() && genMachineOperand(operands[3])->getVal() > 255) {
            operand = genMachineVReg();
            cur_inst = new LoadMInstruction(cur_block, LoadMInstruction::LDR, operand, genMachineOperand(operands[3]));
            operand = new MachineOperand(*operand);
            cur_block->InsertInst(cur_inst);
        }
        cur_block->InsertInst(new MovMInstruction(cur_block, MovMInstruction::MOV, r2, operand));
        cur_block->InsertInst(new BranchMInstruction(cur_block, BranchMInstruction::BL, new MachineOperand("@memset")));
        return;
    }

    int stk_count = 0;
    std::vector<MachineOperand*> vec;

    int float_num = 0, int_num = 0;
    for (size_t i = 1; i < operands.size(); i++) {
        if (operands[i]->getType()->toStr() == "float") {
            float_num++;
        } else {
            int_num++;
        }
    }

    int push_num = 0;
    if (int_num > 4) {
        push_num += int_num - 4;
    }
    if (float_num > 4) {
        push_num += float_num - 4;
    }

    bool if_aligned = true;
    if (push_num % 2 == 0) {
        if_aligned = false;
    }
     
    int gpreg_count = 1;
    for (idx = 1; idx < operands.size(); idx++) {
        if (gpreg_count == 5)
            break;
        if (operands[idx]->getType()->toStr() == "float") {
            continue;
        }
        
        operand = genMachineReg(gpreg_count - 1);
        auto src = genMachineOperand(operands[idx]);
        if (src->isImm() && src->getVal() > 255 || src->isImm() && src->getVal() < -255 ) {
            cur_inst = new LoadMInstruction(cur_block, LoadMInstruction::LDR, operand, src);
        } else {
            cur_inst = new MovMInstruction(cur_block, MovMInstruction::MOV, operand, src);
        }
        cur_block->InsertInst(cur_inst);
        gpreg_count++;
    }

    int fpreg_count = 1;
    size_t int_idx = idx;
    for (idx = 1; idx < operands.size(); idx++) {
        if (fpreg_count == 5 && !if_aligned) {
            break;
        }
        if (fpreg_count == 6 && if_aligned) {
            break;
        }
        if (!(operands[idx]->getType()->toStr() == "float")) {
            continue;
        }
        operand = genMachineFReg(fpreg_count - 1);
        auto src = genMachineFloatOperand(operands[idx]);
        if (src->isImm()) {
            auto internal_reg = genMachineVReg();
            cur_inst = new LoadMInstruction(cur_block, LoadMInstruction::LDR, internal_reg, src);
            cur_block->InsertInst(cur_inst);
            internal_reg = new MachineOperand(*internal_reg);
            cur_inst = new MovMInstruction(cur_block, MovMInstruction::VMOV, operand, internal_reg);
        } else {
            cur_inst = new MovMInstruction(cur_block, MovMInstruction::VMOV, operand, src);
        }
        cur_block->InsertInst(cur_inst);
        fpreg_count++;
    }

    size_t float_idx = idx;
    for (size_t i = operands.size() - 1; i >= std::min(float_idx, int_idx); i--) {
        if (operands[i]->getType()->toStr() == "float" && i >= float_idx) {
            operand = genMachineFloatOperand(operands[i]);
            auto dst = genMachineVReg(true);
            auto internal_reg = genMachineVReg();
            if (operand->isImm()) {
                cur_inst = new LoadMInstruction(cur_block, LoadMInstruction::LDR, internal_reg, operand);
                cur_block->InsertInst(cur_inst);
                internal_reg = new MachineOperand(*internal_reg);
                cur_inst = new MovMInstruction(cur_block, MovMInstruction::VMOV, dst, internal_reg);
                cur_block->InsertInst(cur_inst);
                operand = new MachineOperand(*dst);
            }
            cur_inst = new StackMInstruction(cur_block, StackMInstruction::VPUSH, vec, operand);
            cur_block->InsertInst(cur_inst);
            stk_count++;
        } else if (!(operands[i]->getType()->toStr() == "float") && i >= int_idx) {
            operand = genMachineOperand(operands[i]);
            if (operand->isImm()) {
                auto dst = genMachineVReg();
                if (operand->getVal() < 256) {
                    cur_inst = new MovMInstruction(cur_block, MovMInstruction::MOV, dst, operand);
                } else {
                    cur_inst = new LoadMInstruction(cur_block, LoadMInstruction::LDR, dst, operand);
                }
                cur_block->InsertInst(cur_inst);
                operand = new MachineOperand(*dst);
            }
            cur_inst = new StackMInstruction(cur_block, StackMInstruction::PUSH,vec, operand);
            cur_block->InsertInst(cur_inst);
            stk_count++;
        }
    }
    // std::cout<<"stk_count1 = "<<stk_count<<"\n";

    auto label = new MachineOperand(func->toStr().c_str());
    cur_inst = new BranchMInstruction(cur_block, BranchMInstruction::BL, label);
    cur_block->InsertInst(cur_inst);
    
    if ((gpreg_count >= 5 || fpreg_count >= 5) && stk_count != 0) {
        auto sp = new MachineOperand(MachineOperand::REG, 13);
        cur_inst = new BinaryMInstruction(cur_block, BinaryMInstruction::ADD, 
                sp, sp, genMachineImm(stk_count * 4));
        cur_block->InsertInst(cur_inst);
    }
    if (dst) {
        if (dst->getType()->toStr() == "float") {
            operand = genMachineFloatOperand(dst);
            auto s0 = new MachineOperand(MachineOperand::REG, 16, true);
            cur_inst = new MovMInstruction(cur_block, MovMInstruction::VMOV, operand, s0);
            cur_block->InsertInst(cur_inst);
        } else {
            operand = genMachineOperand(dst);
            auto r0 = new MachineOperand(MachineOperand::REG, 0);
            cur_inst = new MovMInstruction(cur_block, MovMInstruction::MOV, operand, r0);
            cur_block->InsertInst(cur_inst);
        }
    }
    // std::cout<<"stk_count3 = "<<stk_count<<"\n";
    // if(addMov)
    // {
    //     cur_inst = new MovMInstruction(cur_block, MovMInstruction::MOV, 
    //         genMachineOperand(), new MachineOperand(MachineOperand::REG, 10));
    // }
}

CallInstruction::~CallInstruction() {
    if (operands[0]) {
        operands[0]->setDef(nullptr);
        if (operands[0]->usersNum() == 0)
            delete operands[0];
    }
    for (long unsigned int i = 1; i < operands.size(); i++) {
        operands[i]->removeUse(this);
    }
}


FptosiInstruction::FptosiInstruction(Operand* dst,
                                     Operand* src,
                                     BasicBlock* insert_bb)
    : Instruction(FPTOSI, insert_bb), dst(dst), src(src) {
    operands.push_back(dst);
    operands.push_back(src);
    dst->setDef(this);
    src->addUse(this);
}

Instruction* FptosiInstruction::copy(Instruction* ins, BasicBlock* new_bb)
{
    FptosiInstruction* fptosiins = dynamic_cast<FptosiInstruction*>(ins);
    SymbolEntry* dst_se = fptosiins->getOp()[0]->getSymbolEntry();
    SymbolEntry* src_se = fptosiins->getOp()[1]->getSymbolEntry();
    return new FptosiInstruction(new Operand(dst_se), new Operand(src_se), new_bb);
}

void FptosiInstruction::output() const {
    Operand* dst = operands[0];
    Operand* src = operands[1];
    fprintf(yyout, "  %s = fptosi %s %s to %s\n", dst->toStr().c_str(),
            src->getType()->toStr().c_str(), src->toStr().c_str(),
            dst->getType()->toStr().c_str());
}

void FptosiInstruction::replaceUse(Operand* old, Operand* new_) {
    if (operands[1] == old) {
        operands[1]->removeUse(this);
        operands[1] = new_;
        new_->addUse(this);
        src = new_;
    }
}

FptosiInstruction::~FptosiInstruction() {
    operands[0]->setDef(nullptr);
    if (operands[0]->usersNum() == 0)
        delete operands[0];
    operands[1]->removeUse(this);
}

void FptosiInstruction::genMachineCode(AsmBuilder* builder) {
    auto src_operand = genMachineFloatOperand(src);
    auto cur_block = builder->getBlock();
    if (src_operand->isImm()) {
        auto temp = genMachineVReg(true);
        auto internal_reg = genMachineVReg();
        MachineInstruction* cur_inst1 = new LoadMInstruction(cur_block, LoadMInstruction::LDR, internal_reg, src_operand);

        internal_reg = new MachineOperand(*internal_reg);
        MachineInstruction* cur_inst2 = new MovMInstruction(cur_block, MovMInstruction::VMOV, temp, internal_reg);
        
        src_operand = temp;

        cur_block->InsertInst(cur_inst1);
        cur_block->InsertInst(cur_inst2);
    }
    auto vcvtDst = genMachineVReg(true);
    MachineInstruction* cur_inst3 = new VcvtMInstruction(cur_block, VcvtMInstruction::F2S, vcvtDst, src_operand);
    cur_block->InsertInst(cur_inst3);

    auto dst_operand = genMachineOperand(dst);
    MachineInstruction* cur_inst4 = new MovMInstruction(cur_block, MovMInstruction::VMOV, dst_operand, new MachineOperand(*vcvtDst));
    cur_block->InsertInst(cur_inst4);
}



SitofpInstruction::SitofpInstruction(Operand* dst,
                                     Operand* src,
                                     BasicBlock* insert_bb)
    : Instruction(SITOFP, insert_bb), dst(dst), src(src) {
    operands.push_back(dst);
    operands.push_back(src);
    dst->setDef(this);
    src->addUse(this);
}

Instruction* SitofpInstruction::copy(Instruction* ins, BasicBlock* new_bb)
{
    SitofpInstruction* sitifpins = dynamic_cast<SitofpInstruction*>(ins);
    SymbolEntry* dst_se = sitifpins->getOp()[0]->getSymbolEntry();
    SymbolEntry* src_se = sitifpins->getOp()[1]->getSymbolEntry();
    return new SitofpInstruction(new Operand(dst_se), new Operand(src_se), new_bb);
}

void SitofpInstruction::output() const {
    Operand* dst = operands[0];
    Operand* src = operands[1];
    fprintf(yyout, "  %s = sitofp %s %s to %s\n", dst->toStr().c_str(),
            src->getType()->toStr().c_str(), src->toStr().c_str(),
            dst->getType()->toStr().c_str());
}


SitofpInstruction::~SitofpInstruction() {
    operands[0]->setDef(nullptr);
    if (operands[0]->usersNum() == 0)
        delete operands[0];
    operands[1]->removeUse(this);
}

void SitofpInstruction::replaceUse(Operand* old, Operand* new_) {
    if (operands[1] == old) {
        operands[1]->removeUse(this);
        operands[1] = new_;
        new_->addUse(this);
        src = new_;
    }
}


void SitofpInstruction::genMachineCode(AsmBuilder* builder) {
    auto src_operand = genMachineOperand(operands[1]);
    auto cur_block = builder->getBlock();
    if (src_operand->isImm()) {
        auto tmp = genMachineVReg();
        MachineInstruction* cur_inst = new LoadMInstruction(cur_block, LoadMInstruction::LDR, tmp, src_operand);
        src_operand = new MachineOperand(*tmp);
        cur_block->InsertInst(cur_inst);
    }
    auto movDst = genMachineVReg(true);
    MachineInstruction* cur_inst2 = new MovMInstruction(cur_block, MovMInstruction::VMOV, movDst, src_operand);
    MachineInstruction* cur_inst3 = new VcvtMInstruction(cur_block, VcvtMInstruction::S2F, genMachineFloatOperand(dst), new MachineOperand(*movDst));
    cur_block->InsertInst(cur_inst2);
    cur_block->InsertInst(cur_inst3);
}


ZextInstruction::ZextInstruction(Operand* dst,
                                 Operand* src,
                                 BasicBlock* insert_bb)
    : Instruction(ZEXT, insert_bb) {
    operands.push_back(dst);
    operands.push_back(src);
    dst->setDef(this);
    src->addUse(this);
}

Instruction* ZextInstruction::copy(Instruction* ins, BasicBlock* new_bb)
{
    ZextInstruction* zextins = dynamic_cast<ZextInstruction*>(ins);
    SymbolEntry* dst_se = zextins->getOp()[0]->getSymbolEntry();
    SymbolEntry* src_se = zextins->getOp()[1]->getSymbolEntry();
    return new ZextInstruction(new Operand(dst_se), new Operand(src_se), new_bb);
}

void ZextInstruction::output() const {
    Operand* dst = operands[0];
    Operand* src = operands[1];
    fprintf(yyout, "  %s = zext %s %s to i32\n", dst->toStr().c_str(),
            src->getType()->toStr().c_str(), src->toStr().c_str());
}

ZextInstruction::~ZextInstruction() {
    operands[0]->setDef(nullptr);
    if (operands[0]->usersNum() == 0)
        delete operands[0];
    operands[1]->removeUse(this);
}

XorInstruction::XorInstruction(Operand* dst,
                               Operand* src,
                               BasicBlock* insert_bb)
    : Instruction(XOR, insert_bb) {
    operands.push_back(dst);
    operands.push_back(src);
    dst->setDef(this);
    src->addUse(this);
    this->push_back_opse(src->getSymbolEntry());
}

Instruction* XorInstruction::copy(Instruction* ins, BasicBlock* new_bb)
{
    XorInstruction* xorins = dynamic_cast<XorInstruction*>(ins);
    SymbolEntry* dst_se = xorins->getOp()[0]->getSymbolEntry();
    SymbolEntry* src_se = xorins->getOp()[1]->getSymbolEntry();
    return new XorInstruction(new Operand(dst_se), new Operand(src_se), new_bb);
}

void XorInstruction::output() const {
    Operand* dst = operands[0];
    Operand* src = operands[1];
    fprintf(yyout, "  %s = xor %s %s, true\n", dst->toStr().c_str(),
            src->getType()->toStr().c_str(), src->toStr().c_str());
}

XorInstruction::~XorInstruction() {
    operands[0]->setDef(nullptr);
    if (operands[0]->usersNum() == 0)
        delete operands[0];
    operands[1]->removeUse(this);
}

GepInstruction::GepInstruction(Operand* dst,
                               Operand* arr,
                               Operand* idx,
                               BasicBlock* insert_bb,
                               bool paramFirst)
    : Instruction(GEP, insert_bb), paramFirst(paramFirst) {

    operands.push_back(dst);
    operands.push_back(arr);
    operands.push_back(idx);
    
    dst->setDef(this);
    arr->addUse(this);
    idx->addUse(this);
    first = false;
    init = nullptr;
    last = false;
    noAsm = false;

    if (!paramFirst)
    {
        this->push_back_opse(arr->getSymbolEntry());
    }
    else
    {
        this->push_back_opse(dst->getSymbolEntry());
    }
    if (idx != nullptr)
        this->push_back_opse(idx->getSymbolEntry());

    ArrayType* type =
            (ArrayType*)(((PointerType*)(operands[1]->getType()))->getType());
    Type* elementType = type->getElementType();
}

Instruction* GepInstruction::copy(Instruction* ins, BasicBlock* new_bb)
{
    GepInstruction* gepins = dynamic_cast<GepInstruction*>(ins);
    SymbolEntry* dst_se = gepins->getOp()[0]->getSymbolEntry();
    SymbolEntry* arr_se = gepins->getOp()[1]->getSymbolEntry();
    SymbolEntry* index_se = gepins->getOp()[2]->getSymbolEntry();
    GepInstruction* res = new GepInstruction(new Operand(dst_se), new Operand(arr_se), new Operand(index_se), new_bb);


    if(gepins->getparamFirst() == true)res->setparamFirst();
    if(gepins->getFirst() == true)res->setFirst();
    if(gepins->getLast() == true)res->setLast();
    Operand* init = gepins->getInit();
    if(init)
    {
        SymbolEntry* init_se = init->getSymbolEntry();
        res->setInit1(new Operand(init_se), 0);
    }
    res->setNoAsm(gepins->hasNoAsm());
    
    res->setOff(gepins->getOff());

    return res;
}

void GepInstruction::output() const {
    Operand* dst = operands[0];
    Operand* arr = operands[1];
    Operand* idx = operands[2];

    std::string arrType = arr->getType()->toStr();
    ArrayType* type = (ArrayType*)(((PointerType*)(operands[1]->getType()))->getType());
    Type* elementType = type->getElementType();
    
    // wzr-multi:
    fprintf(yyout, "paramFirst = %d, first = %d, last = %d, off = %d, noAsm = %d\n", 
        paramFirst, first, last, off, noAsm);

    if (paramFirst) {
        fprintf(yyout, "  %s = getelementptr inbounds %s, %s %s, i32 %s\n",
                dst->toStr().c_str(),
                arrType.substr(0, arrType.size() - 1).c_str(), arrType.c_str(),
                arr->toStr().c_str(), idx->toStr().c_str());
    }
    else {
        fprintf(yyout, "  %s = getelementptr inbounds %s, %s %s, i32 0, i32 %s\n",
            dst->toStr().c_str(), arrType.substr(0, arrType.size() - 1).c_str(),
            arrType.c_str(), arr->toStr().c_str(), idx->toStr().c_str());
    }
}

GepInstruction::~GepInstruction() {
    operands[0]->setDef(nullptr);
    if (operands[0]->usersNum() == 0)
        delete operands[0];
    operands[1]->removeUse(this);
    operands[2]->removeUse(this);

    ArrayType* type =
            (ArrayType*)(((PointerType*)(operands[1]->getType()))->getType());
    Type* elementType = type->getElementType();
}

void GepInstruction::replaceUse(Operand* old, Operand* new_) {
    if (operands[1] == old) {
        ArrayType* type =
                (ArrayType*)(((PointerType*)(operands[1]->getType()))->getType());
        Type* elementType = type->getElementType();

        operands[1]->removeUse(this);
        operands[1] = new_;
        new_->addUse(this);
        type = (ArrayType*)(((PointerType*)(operands[1]->getType()))->getType());
        elementType = type->getElementType();

    } else if (operands[2] == old) {
        operands[2]->removeUse(this);
        operands[2] = new_;
        new_->addUse(this);
    }
}

std::vector<Operand*> GepInstruction::getUse() {
    ArrayType* type = (ArrayType*)(((PointerType*)(operands[1]->getType()))->getType());
    Type* elementType = type->getElementType();

    return std::vector<Operand*>({operands[1], operands[2]});
}

void GepInstruction::genMachineCode(AsmBuilder* builder) {
    ArrayType* type_0 = (ArrayType*)(((PointerType*)(operands[1]->getType()))->getType());
    auto cur_block = builder->getBlock();
    MachineInstruction* cur_inst;
    auto dst = genMachineOperand(operands[0]);
    auto idx = genMachineOperand(operands[2]);
    if (init) {
        if(!last) {
            noAsm = true;
            return;
        } else {
            auto base = genMachineOperand(init);
            MachineOperand* imm = genMachineImm(off + 4);
            int off = this->off;
            off += 4;
            if (off > 255) {
                MachineOperand* temp = genMachineVReg();
                cur_block->InsertInst(new LoadMInstruction(cur_block, LoadMInstruction::LDR, temp, imm));
                imm = temp;
            }
            cur_inst = new BinaryMInstruction(cur_block, BinaryMInstruction::ADD, dst, base, imm);
            cur_block->InsertInst(cur_inst);
            return;
        }
    }
    MachineOperand* base = nullptr;
    auto idx1 = genMachineVReg();
    if (idx->isImm()) {
        if (idx->getVal() < 255) {
            cur_inst = new MovMInstruction(cur_block, MovMInstruction::MOV, idx1, idx);
        } else {
            cur_inst = new LoadMInstruction(cur_block, LoadMInstruction::LDR, idx1, idx);
        }
        idx = new MachineOperand(*idx1);
        cur_block->InsertInst(cur_inst);
    }
    long long size = 0;
    if (paramFirst) {
        size = ((PointerType*)(operands[1]->getType()))->getType()->getSize();
        size /= 8;
        
        ArrayType* type = (ArrayType*)(((PointerType*)(operands[1]->getType()))->getType());
        Type* elementType = type->getElementType();
    } else {
        if (first) {
            //std::cout<<"first  == true\n";
            ////std::cout<<"breakpoint 4\n";
            base = genMachineVReg();
            if (operands[1]->getEntry()->isVariable() && ((IdentifierSymbolEntry*)(operands[1]->getEntry()))->isGlobal()) {
                auto src = genMachineOperand(operands[1]);
                cur_inst = new LoadMInstruction(cur_block, LoadMInstruction::LDR, base, src);
                
                //std::cout<<"check operands[1] 7\n";
                ArrayType* type = (ArrayType*)(((PointerType*)(operands[1]->getType()))->getType());
                //std::cout<<"origin ayyaytype type = "<<type->toStr()<<"\n";
                Type* elementType = type->getElementType();
                //std::cout<<"origin elementtype = "<<elementType<<"\n";
            } else {
                int offset = ((TemporarySymbolEntry*)(operands[1]->getEntry()))->getOffset();

                //std::cout<<"check operands[1] 8\n";
                ArrayType* type = (ArrayType*)(((PointerType*)(operands[1]->getType()))->getType());
                //std::cout<<"origin ayyaytype type = "<<type->toStr()<<"\n";
                Type* elementType = type->getElementType();
                //std::cout<<"origin elementtype = "<<elementType<<"\n";

                if (offset > -255 && offset < 255) {
                    cur_inst = new MovMInstruction(cur_block, MovMInstruction::MOV, base, genMachineImm(offset));
                } else {
                    cur_inst = new LoadMInstruction(cur_block, LoadMInstruction::LDR, base, genMachineImm(offset));
                }
            }
            cur_block->InsertInst(cur_inst);
        }

        //std::cout<<"check operands[1] 9\n";
        ArrayType* type = (ArrayType*)(((PointerType*)(operands[1]->getType()))->getType());
        // std::cout<<"ayyaytype type1 = "<<((PointerType*)(operands[1]->getType()))->toStr()<<"\n";
        // std::cout<<"ayyaytype type2 = "<<type->toStr()<<"\n";
        Type* elementType = type->getElementType();
        // std::cout<<"elementtype = "<<elementType<<"\n";
        assert(elementType != nullptr && "gep genMachineCode() arraytype->getElementType() == nullptr!!!");
        size = elementType->getSize() / 8;//cause segement fault
        //std::cout<<"breakpoint3\n";
    }
    auto size_ = genMachineVReg();
    if(size <= -255 || size >=255) {
        cur_inst = new LoadMInstruction(cur_block, LoadMInstruction::LDR, size_, genMachineImm(size));
        cur_block->InsertInst(cur_inst);
    } else {
        cur_inst = new MovMInstruction(cur_block, MovMInstruction::MOV, size_, genMachineImm(size));
        cur_block->InsertInst(cur_inst);
    }

    
    size_ = new MachineOperand(*size_);
    auto off = genMachineVReg();
    cur_inst = new BinaryMInstruction(cur_block, BinaryMInstruction::MUL, off, idx, size_);
    off = new MachineOperand(*off);
    cur_block->InsertInst(cur_inst);
    if (paramFirst || !first) {
        //std::cout<<"branch if\n"<<(int)first;
        //std::cout<<"paramFirst || !first  == true\n";
        auto in = operands[1]->getDef();

        //std::cout<<"check operands[1] 10\n";
        ArrayType* type = (ArrayType*)(((PointerType*)(operands[1]->getType()))->getType());
        //std::cout<<"origin ayyaytype type = "<<type->toStr()<<"\n";
        Type* elementType = type->getElementType();
        //std::cout<<"origin elementtype = "<<elementType<<"\n";

        if (in && in->isGep()) {
            auto gepinst = (GepInstruction*)in;
            if (gepinst->hasNoAsm()) {
                gepinst->setInit1(nullptr, 0);
                gepinst->genMachineCode(builder);
            }
        }
        cur_inst = new BinaryMInstruction(cur_block, BinaryMInstruction::ADD, dst, genMachineOperand(operands[1]), off);
        cur_block->InsertInst(cur_inst);
    } else {
        auto addr = genMachineVReg();
        auto base1 = new MachineOperand(*base);
        if (operands[1]->getEntry()->isVariable() &&
            ((IdentifierSymbolEntry*)(operands[1]->getEntry()))->isGlobal()) {
            ArrayType* type = (ArrayType*)(((PointerType*)(operands[1]->getType()))->getType());
            //std::cout<<"origin ayyaytype type = "<<type->toStr()<<"\n";
            Type* elementType = type->getElementType();
            //std::cout<<"origin elementtype = "<<elementType<<"\n";

            cur_inst = new BinaryMInstruction(cur_block, BinaryMInstruction::ADD, addr, base1, off);
            cur_block->InsertInst(cur_inst);
            addr = new MachineOperand(*addr);
            cur_inst = new MovMInstruction(cur_block, MovMInstruction::MOV, dst, addr);
            cur_block->InsertInst(cur_inst);
        } else {
            //std::cout<<"branch-branch else\n";
            cur_inst = new BinaryMInstruction(cur_block, BinaryMInstruction::ADD, addr, genMachineReg(11), base1);
            cur_block->InsertInst(cur_inst);
            addr = new MachineOperand(*addr);
            cur_inst = new BinaryMInstruction(cur_block, BinaryMInstruction::ADD, dst, addr, off);
            cur_block->InsertInst(cur_inst);
        }
    }
    //std::cout<<"gepinstruction genmachinecode end\n";
}


void ZextInstruction::genMachineCode(AsmBuilder* builder) {
    auto cur_block = builder->getBlock();
    auto cur_inst =
        new MovMInstruction(cur_block, 
            MovMInstruction::MOV, 
            genMachineOperand(operands[0]), 
            genMachineOperand(operands[1]));
    cur_block->InsertInst(cur_inst);
}

void ZextInstruction::replaceUse(Operand* old, Operand* new_) {
    if (operands[1] == old) {
        operands[1]->removeUse(this);
        operands[1] = new_;
        new_->addUse(this);
    }
}

void XorInstruction::genMachineCode(AsmBuilder* builder) {
    auto dst = genMachineOperand(operands[0]);
    auto cur_block = builder->getBlock();
    auto cur_inst1 = new MovMInstruction(cur_block, MovMInstruction::MOV, dst, genMachineImm(1), MachineInstruction::EQ);
    auto cur_inst2 = new MovMInstruction(cur_block, MovMInstruction::MOV, dst, genMachineImm(0), MachineInstruction::NE);
    cur_block->InsertInst(cur_inst1);
    cur_block->InsertInst(cur_inst2);
}

void XorInstruction::replaceUse(Operand* old, Operand* new_) {
    if (operands[1] == old) {
        operands[1]->removeUse(this);
        operands[1] = new_;
        new_->addUse(this);
    }
}

PhiInstruction::PhiInstruction(Operand* dst, BasicBlock* insert_bb)
    : Instruction(PHI, insert_bb) {
    operands.emplace_back(dst);
    this->dst = dst;
    this->originDef = dst;
    dst->setDef(this);
}

Instruction* PhiInstruction::copy(Instruction* ins, BasicBlock* new_bb)
{
    PhiInstruction* phiins = dynamic_cast<PhiInstruction*>(ins);
    SymbolEntry* dst_se = phiins->getOp()[0]->getSymbolEntry();
    // for hoist
    auto inst = new PhiInstruction(new Operand(dst_se), new_bb);
    inst->setIdentse((IdentifierSymbolEntry*)ident); // for dce
    inst->push_back_opse(op_se[0]); // for cse
    for (auto src : srcs) {
        inst->addSrc(src.first, src.second);
    }
    return inst;
}

PhiInstruction::~PhiInstruction() {
    dst->setDef(nullptr);
    if (dst->usersNum() == 0)
        delete dst;
    for (auto it : srcs)
        it.second->removeUse(this);
}

void PhiInstruction::output() const {
    // dst = operands[0];
    fprintf(yyout, "  %s = phi %s", operands[0]->toStr().c_str(),
            operands[0]->getType()->toStr().c_str());
    bool first = true;
    for (auto it = srcs.begin(); it != srcs.end(); it++) {
        if (!first)
            fprintf(yyout, ", ");
        else
            first = false;
        fprintf(yyout, "[ %s , %%B%d ]", it->second->toStr().c_str(),
                it->first->getNo());
    }
    fprintf(yyout, "\n");
}

Instruction* BinaryInstruction::copy(Instruction* ins, BasicBlock* new_bb)
{
    BinaryInstruction* binaryins = dynamic_cast<BinaryInstruction*>(ins);
    unsigned mopcode = binaryins->getOpcode();
    SymbolEntry* dst_se = binaryins->getOp()[0]->getSymbolEntry();
    SymbolEntry* dst_se2 = this->getOp()[0]->getSymbolEntry();
    assert(dst_se->getType()->toStr() == dst_se2->getType()->toStr());
    // if(dst_se->getLabel() == 31)
    //     printf("in copy vsub dst_se type = %s\n", (this->getOp()[0]->getSymbolEntry())->getType()->toStr().c_str());
    SymbolEntry* src1_se = binaryins->getOp()[1]->getSymbolEntry();
    SymbolEntry* src2_se = binaryins->getOp()[2]->getSymbolEntry();
    return new BinaryInstruction(mopcode, new Operand(dst_se), new Operand(src1_se), new Operand(src2_se), new_bb);
}

void BinaryInstruction::replaceDef(Operand* new_) {
    operands[0]->removeDef(this);
    operands[0] = new_;
    new_->setDef(this);
}


void CmpInstruction::replaceDef(Operand* new_) {
    operands[0]->removeDef(this);
    operands[0] = new_;
    new_->setDef(this);
}

void RetInstruction::replaceDef(Operand* new_) {
    if (operands.size()) {
        operands[0]->removeDef(this);
        operands[0] = new_;
        new_->setDef(this);
    }
}

void BitcastInstruction::replaceDef(Operand* new_) {
    operands[0]->removeDef(this);
    operands[0] = new_;
    new_->setDef(this);
    dst = new_;
}


void AllocaInstruction::replaceDef(Operand* new_) {
    operands[0]->removeDef(this);
    operands[0] = new_;
    new_->setDef(this);
}

void LoadInstruction::replaceDef(Operand* new_) {
    operands[0]->removeDef(this);
    operands[0] = new_;
    new_->setDef(this);
}

void StoreInstruction::replaceDef(Operand* new_) {
    operands[0]->removeDef(this);
    operands[0] = new_;
    new_->addUse(this); //G2L的数组替换；
}


void CallInstruction::replaceDef(Operand* new_) {
    if (dst) {
        operands[0]->removeDef(this);
        operands[0] = new_;
        new_->setDef(this);
        dst = new_;
    }
}


void ZextInstruction::replaceDef(Operand* new_) {
    operands[0]->removeDef(this);
    operands[0] = new_;
    new_->setDef(this);
}


void XorInstruction::replaceDef(Operand* new_) {
    operands[0]->removeDef(this);
    operands[0] = new_;
    new_->setDef(this);
}


void GepInstruction::replaceDef(Operand* new_) {
    operands[0]->removeDef(this);
    operands[0] = new_;
    new_->setDef(this);
}


void PhiInstruction::replaceUse(Operand* old, Operand* new_) {
    for (auto& it : srcs) {
        if (it.second == old) {
            it.second->removeUse(this);
            it.second = new_;
            new_->addUse(this);
        }
    }
    for (auto it = operands.begin() + 1; it != operands.end(); it++)
        if (*it == old)
            *it = new_;
}

void PhiInstruction::replaceDef(Operand* new_) {
    dst->removeDef(this);
    dst = new_;
    new_->setDef(this);
    // this->dst = operands[0];
    operands[0] = new_;
}

void PhiInstruction::addSrc(BasicBlock* block, Operand* src) {
    operands.emplace_back(src);
    srcs.insert(std::make_pair(block, src));
    src->addUse(this);
    phiOperands.push_back(src);
    phiBBs.push_back(block);
}



void FptosiInstruction::replaceDef(Operand* new_) {
    operands[0]->removeDef(this);
    operands[0] = new_;
    new_->setDef(this);
    dst = new_;
}

void SitofpInstruction::replaceDef(Operand* new_) {
    operands[0]->removeDef(this);
    operands[0] = new_;
    new_->setDef(this);
    dst = new_;
}



void PhiInstruction::genMachineCodeAfter(AsmBuilder* builder){
    // 0408m
    // fix: 由于每次生成机器代码是一个block一个block生成的, 因此在当前phi指令的block当中可能还没有定义来源路径block
    // 可能的一种解决方案是在整体机器代码生成结束进行Phi resolution
    auto f = builder->getFunction();
    auto mdst = genMachineOperand(operands[0]);
    // auto tempop = new TemporarySymbolEntry(TypeSystem::intType, SymbolTable::getLabel());
    MachineBlock* block = builder->getBlock();
    // MachineOperand* temp = new MachineOperand(MachineOperand::REG, SymbolTable::getLabel());
    // std::cout<<this->getDef()->toStr()<<"\t";
    for (auto src : srcs){
        MachineBlock* b = src.first->getParent()->getMachineBlock(src.first);
        // std::cout<<src.first->getNo()<<"\t"<<b<<"\t";
        auto mop = genMachineOperand(src.second);
        auto inst = new MovMInstruction(b, MovMInstruction::MOV, mdst, mop);
        auto endInst = b->end();
        
        b->InsertInstBeforeBr(inst);
        
    }
    // std::cout<<"\n";
    // new MovMInstruction(block, MovMInstruction::MOV, mdst, temp);
}

void PhiInstruction::output_to_console() const {
    printf("  %s = phi %s", dst->toStr().c_str(),
           dst->getType()->toStr().c_str());
    bool first = true;
    for (auto it = srcs.begin(); it != srcs.end(); it++) {
        if (!first)
            printf(", ");
        else
            first = false;
        printf("[ %s , %%B%d ]", it->second->toStr().c_str(),
               it->first->getNo());
    }
    printf("\n");
}

