#include "InstructionsSchedule.h"

#define DEBUG_SWITCH_ische 0
#if DEBUG_SWITCH_ische
#define printISC(str) std::cout << str << "\n";
#else
#define printISC(str) //
#endif



//ͬͻָdelayֵοCortex-A72 Software Optimization Guide
int InstructionsScheduling::getInstDelay(MachineInstruction* inst) {
    switch (inst->getType())
    {
    case MachineInstruction::VNEG:
        printISC("    " << "    " << "getInstDelay: MachineInstruction::VNEG");
        return 3;
    case MachineInstruction::FUSE:
        printISC("    " << "    " << "getInstDelay: MachineInstruction::FUSE");
        return 3;
    case MachineInstruction::LOAD:
        printISC("    " << "    " << "getInstDelay: MachineInstruction::LOAD");
        return 4;
    case MachineInstruction::STORE:
        printISC("    " << "    " << "getInstDelay: MachineInstruction::STORE");
        return 1;
    case MachineInstruction::BRANCH:
        printISC("    " << "    " << "getInstDelay: MachineInstruction::BRANCH");
        return 1;
    case MachineInstruction::VCVT:
        printISC("    " << "    " << "getInstDelay: MachineInstruction::VCVT");
        return 3;
    case MachineInstruction::VMRS:
        printISC("    " << "    " << "getInstDelay: MachineInstruction::VMRS");
        return 1;
    case MachineInstruction::BINARY:
        printISC("    " << "    " << "getInstDelay: MachineInstruction::BINARY");
        switch (inst->getOp())
        {
        case BinaryMInstruction::ADD:
            return 1;
        case BinaryMInstruction::SUB:
            return 1;
        case BinaryMInstruction::MUL:
            return 3;
        case BinaryMInstruction::DIV:
            return 4;
        case BinaryMInstruction::AND:
            return 1;
        case BinaryMInstruction::OR:
            return 1;
        case BinaryMInstruction::VADD:
            return 4;
        case BinaryMInstruction::VSUB:
            return 4;
        case BinaryMInstruction::VMUL:
            return 4;
        case BinaryMInstruction::VDIV:
            return 6;
        default:
            return 4;
        }
    case MachineInstruction::MOV:
        printISC("    " << "    " << "getInstDelay: MachineInstruction::MOV");
        switch (inst->getOp())
        {
        case MovMInstruction::VMOV:
            return 5;
        case MovMInstruction::VMOVF32:
            return 3;
        default:
            return 1;
        }
    case MachineInstruction::CMP:
        printISC("    " << "    " << "getInstDelay: MachineInstruction::CMP");
        switch (inst->getOp())
        {
        case CmpMInstruction::VCMP:
            return 3;
        default:
            return 1;
        }
    case MachineInstruction::STACK:
        printISC("    " << "    " << "getInstDelay: MachineInstruction::STACK");
        switch (inst->getOp())
        {
        case StackMInstruction::POP:
            return 3 + ((StackMInstruction*)inst)->getUse().size();
        case StackMInstruction::VPOP:
            return 4 + ((StackMInstruction*)inst)->getUse().size();
        default:
            return ((StackMInstruction*)inst)->getUse().size();
        }
    default:
        printISC("    " << "    " << "getInstDelay: default");
        return 6;
    }
}

void InstructionsScheduling::calculate_priority(int index) {
    Node* node = Nodelist[index];
    if (node->pred.empty()) { //ûԼָ;
        node->priority = 0; 
        // printISC("    " << "    " << "    " << "node->pred is empty.");
        printISC("    " << "    " << "    " << "node->priority: " << node->priority);
        return;
    }
    for (auto i: node->pred) {
        if (Nodelist[i]->priority == -1) { //ݹԼҵָȼ;
            calculate_priority(i);
        }
    }
    for (auto i: node->pred) {
        int latency = 0;
        for(auto succ: Nodelist[i]->succ) {
            if(succ.first == index) {
                latency = succ.second;
            }
        }
        node->priority = std::max(node->priority, Nodelist[i]->priority + latency); //DPµǰȼֻܽ
    }
    printISC("    " << "    " << "    " << "node->priority: " << node->priority);
}

void InstructionsScheduling::pass() {

    printISC("InstructionsScheduling::pass begin: ");
    for (auto func_iter = mUnit->begin(); func_iter != mUnit->end(); func_iter++) {
        auto func = *func_iter;
        for (auto block_iter = func->begin(); block_iter != func->end(); block_iter++) {
            //MblockΪλŻ:
            auto block = *block_iter; 
            printISC("    " << "Process Mblock: " << block->getNo());

            if (block->getInsts().empty()) {
                continue;
            }
            auto Instlist = block->getInsts();
            if(Instlist.size() >= 3000) {
                return;
            }

            //ݽṹNodelistָӦdelay;
            Nodelist.clear();
            for(auto inst: Instlist) {
                Node* node = new Node(inst);
                int delay = this->getInstDelay(inst);
                node->delay = delay;
                Nodelist.push_back(node);
                // printISC("    " << "    " << "inst: " << inst->getDef()[0]->toStr());
            }

            printISC("    " << "    " << "Instlist.size: " << Instlist.size());

            //Լϵ:
            for(int i = 0; i < Nodelist.size(); i++) {
                Node* node = Nodelist[i];
                if(node->inst->getType() == MachineInstruction::STACK) {
                    for(int j = i; j < Nodelist.size(); j++) {
                        node->succ.push_back(std::pair(j, node->delay));
                        if (Nodelist[j]->inst->isBranch()) { 
                            break;
                        }
                    }
                    continue;
                }
                if(node->inst->getType() == MachineInstruction::BRANCH) {
                    for(int j = i; j < Nodelist.size(); j++) {
                        node->succ.push_back(std::pair(j, node->delay));
                    }
                    continue;
                }
                if(node->inst->isSpecial()) {
                    for(int j = i; j < Nodelist.size(); j++) {
                        node->succ.push_back(std::pair(j, node->delay));
                    }
                    continue;
                }

                //general situation:
                for(int j = i; j < Nodelist.size(); j++) {
                    auto prev_defs = node->inst->getDef();
                    auto prev_uses = node->inst->getUse();
                    auto succ_defs = Nodelist[j]->inst->getDef();
                    auto succ_uses = Nodelist[j]->inst->getUse();

                    //ждϵĲͬʱ:
                    if ((node->inst->isLoad() && Nodelist[j]->inst->isStore()) ||
                        (node->inst->isStore() && Nodelist[j]->inst->isLoad()) ||
                        (node->inst->isStore() && Nodelist[j]->inst->isStore())) {
                        node->succ.push_back(std::pair(j, node->delay));
                    }
                    for (auto prev_def : prev_defs) {  // RAW
                        for (auto succ_use : succ_uses) {
                            if (*prev_def == *succ_use) {
                                node->succ.push_back(std::pair(j, node->delay));
                            }
                        }
                    }
                    for (auto prev_use : prev_uses) { // WAR
                        for (auto succ_def : succ_defs) {
                            if (*prev_use == *succ_def) {
                                node->succ.push_back(std::pair(j, node->delay));
                            }
                        }
                    }
                    for (auto prev_def : prev_defs) { //WAW;
                        for (auto succ_def : succ_defs) {
                            if (*prev_def == *succ_def) {
                                node->succ.push_back(std::pair(j, node->delay));
                            }
                        }
                    }

                    //
                    if (node->inst->isCmp() && Nodelist[j]->inst->isCondMov()) { 
                        node->succ.push_back(std::pair(j, node->delay));
                    }
                    if (Nodelist[j]->inst->isBranch() 
                        || Nodelist[j]->inst->isVMRS()
                        || Nodelist[j]->inst->isSpecial()) {
                        node->succ.push_back(std::pair(j, node->delay));
                    }
                }
                printISC("    " << "    " << "Instlist: " << i << " general"); 
            }


            printISC("    " << "    " << "delete Nodes. ");

            //ɾ:
            for(int i = 0; i < Nodelist.size(); i++) {
                if(i % 1000 == 0) printISC("    " << "    " << "Instlist: " << i);
                for(int j = 0; j < Nodelist[i]->succ.size(); j++) {
                    if(Nodelist[i]->succ[j].first == i) {
                        Nodelist[i]->succ.erase(Nodelist[i]->succ.begin() + j);
                        j--;
                        continue;
                    }
                    if(j < Nodelist[i]->succ.size() - 1) {
                        if(Nodelist[i]->succ[j].first == Nodelist[i]->succ[j + 1].first) {
                            Nodelist[i]->succ.erase(Nodelist[i]->succ.begin() + j);
                            j--;
                        }
                    }
                }
            }
            printISC("    " << "    " << "delete Nodes success. ");

            for(int i = 0; i < Nodelist.size(); i++) {
                Node* node = Nodelist[i];
                for(auto succ: node->succ) {
                    int j = succ.first;
                    Nodelist[j]->pred.push_back(i);
                    Nodelist[j]->degree++;
                }
            }
            printISC("    " << "    " << "build CFG success. ");

            //ӡǷл:
            for(int i = 0; i < Nodelist.size(); i++) {
                printISC("    " << "    " << "    " << "Node " << i << "(degree "
                        << Nodelist[i]->degree << "): ")
                for(int j = 0; j < Nodelist[i]->succ.size(); j++) {
                   printISC("    " << "    " << "    " << "    " 
                                    << "succ: " << Nodelist[i]->succ[j].first 
                                    << " dalay: " << Nodelist[i]->succ[j].second) 
                }
            }

            //ȼ:
            printISC("    " << "    " << "calculate_priority begin: ");
            for (size_t i = 0; i < Nodelist.size(); i++) {
                if(Nodelist[i]->priority == -1) {
                    printISC("    " << "    " << "    " << "calculate_priority: " << i);
                    calculate_priority(i);
                }
            }

            printISC("    " << "    " << "calculate_priority success. ");
            

            //ɵ:
            printISC("    " << "    " << "schedule begin. ");
            std::vector<Node*> inWaitlist;
            std::vector<Node*> inPipeline;
            std::vector<Node*> scheduledList;
            for (size_t i = 0; i < Nodelist.size(); i++) {
                if(Nodelist[i]->pred.empty()) {
                    inWaitlist.push_back(Nodelist[i]);
                }
            }

            printISC("    " << "    " << "inWaitlist.size: " << inWaitlist.size());

            int cpuCycle = 0;
            std::map<Node*, int> inTime;
            while(!(inWaitlist.empty() && inPipeline.empty())) {
                printISC("    " << "    " << "    " << "cpuCycle: " << cpuCycle);
                std::sort(inWaitlist.begin(), inWaitlist.end(), 
                            [&](Node* a, Node*b) -> bool {return a->priority < b->priority;}); //
                for(auto node = inWaitlist.begin(); node != inWaitlist.end(); ) {
                    inPipeline.push_back(*node);
                    inTime[*node] = cpuCycle;
                    scheduledList.push_back(*node);
                    printISC("    " << "    " << "    " << "    " << "inPipeline.push_back: " << getIndex(*node))
                    node = inWaitlist.erase(node);
                }

                for(auto node = inPipeline.begin(); node != inPipeline.end(); ) {
                    if((inTime[*node] + (*node)->delay) <= cpuCycle + 1) {
                        printISC("    " << "    " << "    " << "    " <<  getIndex(*node) << ": " << inTime[*node] + (*node)->delay << " <= " << cpuCycle + 1)                        
                        for(auto succ : (*node)->succ) {
                            printISC("    " << "    " << "    " << "    " << "    " << "Nodelist[" << succ.first << "]->degree --");
                            Nodelist[succ.first]->degree --;
                            if(Nodelist[succ.first]->degree <= 0) {
                                inWaitlist.push_back(Nodelist[succ.first]);
                                printISC("    " << "    " << "    " << "    " << "    " << "    " << "inWaitlist.push_back " << succ.first);
                            }
                        }
                        node = inPipeline.erase(node);
                        printISC("    " << "    " << "    " << "    " << "inPipeline.erase: " << getIndex(*node));
                    } else {
                        node++;
                    }
                }
                cpuCycle ++;
            }

            printISC("    " << "    " << "schedule succes. ");


            //ʹָԭָ:
            printISC("    " << "    " << "replace begin. ");
            std::vector<MachineInstruction*> scheduledInstlist;
            
            for (size_t i = 0; i < scheduledList.size(); i++) {
                printISC("    " << "    " << "    " << "scheduledInstlist.push_back: " << getIndex(scheduledList[i]));
                scheduledInstlist.push_back(scheduledList[i]->inst);
            }
            block->getInsts() = scheduledInstlist;

            printISC("    " << "    " << "replace success. ");

            /*
            */
        }
    }
}