#include "LoopDetection.h"

using blockSet = std::vector<BasicBlock *>;

#define MAXNEST 2

#define DEBUG_SWITCH_ld 0
#if DEBUG_SWITCH_ld
#define print(str) std::cout << str << "\n";
#else
#define print(str) //
#endif

void LoopDetection::clear() {
    loops.clear();
    coreBBs.clear();
}

void LoopDetection::printloop(blockSet *loop) {
    for (auto bb : *loop) {
        print(" | bb:" << bb->getNo());
    }
}

BasicBlock *LoopDetection::getCondbb(BasicBlock *loopbb) {
    // if((loopbb->getPredBlock()).size()!=2)assert(0);
    for (auto pred : loopbb->getPredBlock()) {
        // print("== pred: "<<pred->getNo());
        if (pred != loopbb)
            return pred;
    }
}

BasicBlock *LoopDetection::getOutbb(BasicBlock *condbb) {
    auto head = condbb->end();
    CondBrInstruction *cond;
    for (auto instr = head->getNext(); instr != head;
         instr = instr->getNext()) {
        if (instr->isCond()) {
            cond = dynamic_cast<CondBrInstruction *>(instr);
            break;
        }
    }
    return cond->getFalseBranch();
}

BasicBlock *LoopDetection::getEndbb(BasicBlock *loopbb) {
    // if((loopbb->gketSuccBlock()).size()!=1)assert(0);
    // if((loopbb->getPredBlock()).size()!=2)assert(0);
    for (auto succ : loopbb->getSuccBlock()) {
        // print("== succ: "<<succ->getNo());
        if (succ != loopbb)
            return succ;
    }
}

void test(Unit *unit) {
    auto detector = new LoopDetection(unit);
    print("start run");

    detector->run();
    print("run finish");
    for (auto it = detector->begin(); it != detector->end(); it++) {
        // if (!detector->isBaseLoop(*it))
        //     continue;
        auto loop = *it;
        print("size = " << loop->size());
        detector->printloop(loop);
        BasicBlock *loopbb;
        BasicBlock *condbb;
        BasicBlock *endbb;
        // if (loop->size() == 1) {
        //     BasicBlock *loopbb = *(loop->begin());
        //     condbb = detector->getCondbb(loopbb);
        //     endbb = detector->getEndbb(loopbb);
        // } else {
        //     condbb = detector->getLoopCond(loop);
        //     endbb = detector->getOutbb(condbb);
        // }
        print("-- condbb = " << condbb->getNo()
                             << " endbb = " << endbb->getNo());
    }
}

void LoopDetection::run() {

    for (auto func : unit->get_functions()) {
        for (auto bb : func->getBlockList()) {
            for (Instruction *ins = bb->begin(); ins != bb->end();
                 ins = ins->getNext()) {
                if (ins->isCond() &&
                    dynamic_cast<CondBrInstruction *>(ins)->getTrueBranch() ==
                        bb) {
                    // std::cout<<"??????? " << bb->getNo()<<"\n";
                    auto bset = new blockSet;
                    bset->push_back(bb);
                    loops.insert(bset);
                }
            }
        }
    }
}

Instruction *findInst(int insttype, BasicBlock *bb) {
    for (Instruction *ins = bb->begin(); ins != bb->end();
         ins = ins->getNext()) {
        if (ins->getInstType() == insttype) {
            return ins;
        }
    }
    return nullptr;
}

#define DEBUG_SWITCH_nest 0
#if DEBUG_SWITCH_nest
#define printn(str) std::cout << str << "\n";
#else
#define printn(str) //
#endif

void LoopDetection::findNestLoop(BasicBlock *start,
                                 CondBrInstruction *cond_start) {
    printn("============= findNestLoop ==============");
    auto bset = new blockSet;

    CondBrInstruction *cond = cond_start;
    UncondBrInstruction *uncond = nullptr;
    BasicBlock *lastCondbb = start;
    if (!cond_start->getisLoopCond()) {
        start = cond->getTrueBranch();
        cond = dynamic_cast<CondBrInstruction *>(
            findInst(Instruction::COND, start));
        uncond = dynamic_cast<UncondBrInstruction *>(
            findInst(Instruction::UNCOND, start));
    }
    BasicBlock *nextbb = start;
    printn("start->getNumOfPred():"<<start->getNumOfPred());
    if(start->getNumOfPred()<2)return;
    bset->push_back(start);
    std::map<BasicBlock *, bool> branchSwitch;
    printn("startbb = " << start->getNo());
    // getchar();
    int cnt = 0;
    while (true) {
        if(cnt>MAXNEST)return;
        if (branchSwitch.find(nextbb) == branchSwitch.end()) {
            branchSwitch[nextbb] = 0;
        }else{
            branchSwitch[nextbb] = 1;
            cnt++;
        }
        if (cond != nullptr) {
            if (cond->getTrueBranch() == start) {
                loops.insert(bset);
                printn("get a loop1 size = " << bset->size());
                break;
            }
            printn("cond->getTrueBranch() = "<<cond->getTrueBranch()->getNo());
            if (lastCondbb == cond->getTrueBranch() || branchSwitch[nextbb])
                nextbb = cond->getFalseBranch();
            else
                nextbb = cond->getTrueBranch();
        } else if (uncond != nullptr) {
            if (uncond->getBranch() == lastCondbb)
                branchSwitch[nextbb] = 1;
            else
                branchSwitch[nextbb] = 0;
            nextbb = uncond->getBranch();
        }
        visitedbb.insert(nextbb);
        bset->push_back(nextbb);
        printn("nextbb = " << nextbb->getNo()
                           << " lastCondbb = " << lastCondbb->getNo());
        if (nextbb == start) {
            loops.insert(bset);
            printn("get a loop2 size = " << bset->size());
            break;
        }
        if (findInst(Instruction::COND, nextbb)) {
            cond = dynamic_cast<CondBrInstruction *>(
                findInst(Instruction::COND, nextbb));
            if (!cond->getisLoop())
                break; // 不做循环内分支，嘶，没卡掉
            lastCondbb = nextbb;
            uncond = nullptr;
        } else if (findInst(Instruction::UNCOND, nextbb)) {
            uncond = dynamic_cast<UncondBrInstruction *>(
                findInst(Instruction::UNCOND, nextbb));
            cond = nullptr;
        } else {
            printn("no cond & uncond");
            // 先不做循环内ret
            break;
        }
    }
    printn("============= findNestLoop END ==============");
}

void LoopDetection::run2() {

    for (auto func : unit->get_functions()) {
        for (auto bb : func->getBlockList()) {
            if (visitedbb.find(bb) != visitedbb.end()) {
                visitedbb.insert(bb);
            }
            if (findInst(Instruction::COND, bb) == nullptr)
                continue;
            auto cond = dynamic_cast<CondBrInstruction *>(
                findInst(Instruction::COND, bb));
            printn("startbb = " << bb->getNo()
                                << " condisloop:" << cond->getisLoop());
            if (!cond->getisLoop())
                continue;
            findNestLoop(bb, cond);
        }
    }
}

LoopDetection::~LoopDetection() {
    for (auto set : loops) {
        if (set)
            delete set;
    }
}
