#include "Multithread.h"
#include "Instruction.h"
#include <unordered_set>
#include "Unit.h"
#define ITERBOUND 40

#define DEBUG_SWITCH_VEC 0
#if DEBUG_SWITCH_VEC
#define debugmul(str) std::cout << str << "\n";
#else
#define debugmul(str) //
#endif

// 以下是一些方便调试的工具函数
void display_instruction(Instruction* ins);

void displayBB(BasicBlock* bb);

// 返回块中第一条具有指定类型的指令
Instruction* find_inst(int insttype, BasicBlock* bb);

std::vector<Instruction*> find_insts(int insttype, BasicBlock* bb);

bool find_bb(BasicBlock* bb, Function* func);

// 先找一条store指令，确定存入的值是0后，找到这个store对应的第一条gep指令（里面有数组首地址）
GepInstruction* find_gep(BasicBlock* body_bb);

Instruction* get_lastInstruction(BasicBlock* bb);

void changeOperandLabel(BasicBlock* newbb, BasicBlock* oldbb);

void changeOperandLabel2(BasicBlock* newbb, BasicBlock* oldbb);

void setBranch_trueBranch(BasicBlock* branch_truebranch, BasicBlock* bb_toprocess);

void setOperandUse(BasicBlock* newbb);

void Multithread::pass()
{
    for (auto it = unit->begin(); it != unit->end(); it++) {

        Function* func = *it;
        checkFunc(func);
        checkFunc3(func, true);
    }
}

void Multithread::checkFunc(Function *func) {
    debugmul("multithread::checkfunc begin");
    // TODO: 判断该函数中是否包含矩阵乘法，并将相关块指针交给doMultithread
    // 1: cond1中有cmp指令，从true分支跳入后两步之内遇到cond2
    BasicBlock* entry = func->getEntry();
    CondBrInstruction* cond_inentry = dynamic_cast<CondBrInstruction*>(find_inst(Instruction::COND, entry));
    if(!cond_inentry)return;
    BasicBlock* cond1 = cond_inentry->getFalseBranch();
    if(cond1->getSuccBlock().size()<1)return;
    if(find_inst(Instruction::CMP, cond1) == nullptr)
    {
        cond1 = cond1->getSuccBlock()[0];
    }
    if(find_inst(Instruction::CMP, cond1) == nullptr)return;
    debugmul("cond1 no = "<<cond1->getNo());
    CondBrInstruction* cond_incond1 = dynamic_cast<CondBrInstruction*>(find_inst(Instruction::COND, cond1));
    if(!cond_incond1)return;
    BasicBlock* cond2 = cond_incond1->getTrueBranch();
    // assert(cond2);
    if(!cond2)return;
    debugmul("cond2 no1 = "<<cond2->getNo());
    if(find_inst(Instruction::CMP, cond2) == nullptr)
    {
        if(cond2->getSuccBlock().size() <= 0)return;
        cond2 = cond2->getSuccBlock()[0];
    }
    if(!cond2)return;
    // getchar();
    if(find_inst(Instruction::CMP, cond2) == nullptr)return;
    // debugmul("cond2 no = "<<cond2->getNo());
    debugmul("cond2 no2 = ???");

    // 2: cond2中有cmp指令，从true分支跳入后遇到if块
    CondBrInstruction* cond_incond2 = dynamic_cast<CondBrInstruction*>(find_inst(Instruction::COND, cond2));
    if(!cond_incond2)return;
    BasicBlock* ifbb = cond_incond2->getTrueBranch();
    debugmul("ifbb no = "<<ifbb->getNo());

    // 3：if块中有一条比eq的cmp, 从true分支跳入后遇到有iter的块，从false分支跳入后遇到cond3
    CmpInstruction* cmp_inifbb = dynamic_cast<CmpInstruction*>(find_inst(Instruction::CMP, ifbb));
    if(!cmp_inifbb)return;
    if(cmp_inifbb->getOpcode() != CmpInstruction::E)return;
    CondBrInstruction* cond_inifbb = dynamic_cast<CondBrInstruction*>(find_inst(Instruction::COND, ifbb));
    if(!cond_inifbb)return;
    BasicBlock* if_truebb = cond_inifbb->getTrueBranch();
    BasicBlock* if_falsebb = cond_inifbb->getFalseBranch();
    if(find_inst(Instruction::BINARY, if_truebb) == nullptr)return;
    debugmul("if_truebb->getSuccBlock()[0] no = "<<if_truebb->getSuccBlock()[0]->getNo());
    if(if_truebb->getSuccBlock()[0] != cond2)return;
    BasicBlock* cond3 = if_falsebb;

    // 4: cond3中有cmp指令，从true分支进入body块
    CmpInstruction* cmp_incond3 = dynamic_cast<CmpInstruction*>(find_inst(Instruction::CMP, cond3));
    if(!cmp_incond3)return;
    CondBrInstruction* cond_incond3 = dynamic_cast<CondBrInstruction*>(find_inst(Instruction::COND, cond3));
    if(!cond_incond3)return;
    BasicBlock* bodybb = cond_incond3->getTrueBranch();

    // 5：body块中有三组gep指令，后两条的dst相乘后与第一个gep的dst相加，存回第四条gep的dst
    LoadInstruction* load_aftergep1 = nullptr;
    LoadInstruction* load_aftergep2 = nullptr;
    LoadInstruction* load_aftergep3 = nullptr;
    StoreInstruction* store_aftergep4 = nullptr;
    BinaryInstruction* mul = nullptr;
    BinaryInstruction* add_aftermul = nullptr;

    std::vector<Instruction*> geps = find_insts(Instruction::GEP, bodybb);
    if(geps.size() != 8)return;
    load_aftergep1 = dynamic_cast<LoadInstruction*>(geps[1]->getNext());
    assert(load_aftergep1);
    load_aftergep2 = dynamic_cast<LoadInstruction*>(geps[3]->getNext());
    assert(load_aftergep2);
    load_aftergep3 = dynamic_cast<LoadInstruction*>(geps[5]->getNext());
    assert(load_aftergep3);
    store_aftergep4 = dynamic_cast<StoreInstruction*>(geps[7]->getNext());
    assert(store_aftergep4);
    std::vector<Instruction*> binarys = find_insts(Instruction::BINARY, bodybb);
    // 找到那条mul指令
    for(auto bin = binarys.begin(); bin != binarys.end(); bin ++)
    {
        if((*bin)->getOpcode() == BinaryInstruction::MUL)
        {
            mul = dynamic_cast<BinaryInstruction*>(*bin);
        }
    }
    assert(mul);
    add_aftermul = dynamic_cast<BinaryInstruction*>(mul->getNext());
    assert(add_aftermul);
    // 判断乘加关系
    debugmul("mul->getOp()[1] label = "<<mul->getOp()[1]->toStr());
    debugmul("load_aftergep2->getOp()[0] label = "<<load_aftergep2->getOp()[0]->toStr());
    debugmul("mul->getOp()[2] label = "<<mul->getOp()[2]->toStr());
    debugmul("load_aftergep3->getOp()[0] label = "<<load_aftergep3->getOp()[0]->toStr());
    if(!((mul->getOp()[1] == load_aftergep2->getOp()[0] && mul->getOp()[2] == load_aftergep3->getOp()[0])
        || (mul->getOp()[1] == load_aftergep3->getOp()[0] && mul->getOp()[2] == load_aftergep2->getOp()[0])))
        return;
    debugmul("add_aftermul->getOp()[1] label = "<<add_aftermul->getOp()[1]->toStr());
    debugmul("mul->getOp()[0] label = "<<mul->getOp()[0]->toStr());
    debugmul("add_aftermul->getOp()[2] label = "<<add_aftermul->getOp()[2]->toStr());
    debugmul("load_aftergep1->getOp()[0] label = "<<load_aftergep1->getOp()[0]->toStr());
    if(!((add_aftermul->getOp()[1] == mul->getOp()[0] && add_aftermul->getOp()[2] == load_aftergep1->getOp()[0]) 
        || (add_aftermul->getOp()[1] == load_aftergep1->getOp()[0] && add_aftermul->getOp()[2] == mul->getOp()[0])))
        return;
    
    // 获取其他关键块
    BasicBlock* iter21 = if_truebb;
    CondBrInstruction* cond_inbody = dynamic_cast<CondBrInstruction*>(find_inst(Instruction::COND, bodybb));
    assert(cond_inbody);
    BasicBlock* iter22 = cond_inbody->getFalseBranch();
    if(find_inst(Instruction::CMP, iter22) == nullptr)
    {
        iter22 = iter22->getSuccBlock()[0];
    }
    if(find_inst(Instruction::CMP, iter22) == nullptr)return;

    CondBrInstruction* cond_initer22 = dynamic_cast<CondBrInstruction*>(find_inst(Instruction::COND, iter22));
    assert(cond_initer22);
    BasicBlock* iter1 = cond_initer22->getFalseBranch();
    if(find_inst(Instruction::CMP, iter1) == nullptr)
    {
        iter1 = iter1->getSuccBlock()[0];
    }
    if(find_inst(Instruction::CMP, iter1) == nullptr)return;
    std::vector<BasicBlock*> blks;
    blks.push_back(cond1);
    blks.push_back(cond2);
    blks.push_back(ifbb);
    blks.push_back(cond3);
    blks.push_back(bodybb);
    blks.push_back(iter21);
    blks.push_back(iter22);
    blks.push_back(iter1);
    doMultithread(blks, func);
    // 向全局传递将要被多线程化的函数表项
    // target_func = dynamic_cast<IdentifierSymbolEntry*>(func->getSymPtr());
    unit->set_target_func_unit(dynamic_cast<IdentifierSymbolEntry*>(func->getSymPtr()));
    debugmul("unit->get_target_func_unit() = "<<unit->get_target_func_unit()->toStr());
    debugmul("multithread::checkfunc end");
    printf("checkfunc end\n");
}

bool checkFunc2(Function *func) {
    debugmul("multithread::checkfunc2 begin");
    // TODO: 判断该函数中是否包含矩阵乘法，并将相关块指针交给doMultithread
    // 1: cond1中有cmp指令，从true分支跳入后两步之内遇到cond2
    BasicBlock* entry = func->getEntry();
    CondBrInstruction* cond_inentry = dynamic_cast<CondBrInstruction*>(find_inst(Instruction::COND, entry));
    if(!cond_inentry)return false;
    BasicBlock* cond1 = cond_inentry->getFalseBranch();
    if(cond1->getSuccBlock().size()<1)return false;
    if(find_inst(Instruction::CMP, cond1) == nullptr)
    {
        cond1 = cond1->getSuccBlock()[0];
    }
    if(find_inst(Instruction::CMP, cond1) == nullptr)return false;
    debugmul("cond1 no = "<<cond1->getNo());
    CondBrInstruction* cond_incond1 = dynamic_cast<CondBrInstruction*>(find_inst(Instruction::COND, cond1));
    if(!cond_incond1)return false;
    BasicBlock* cond2 = cond_incond1->getTrueBranch();
    if(find_inst(Instruction::CMP, cond2) == nullptr)
    {
        if(cond2->getSuccBlock().size()<=0)return false;
        cond2 = cond2->getSuccBlock()[0];
    }
    if(find_inst(Instruction::CMP, cond2) == nullptr)return false;
    debugmul("cond2 no = "<<cond2->getNo());

    // 2: cond2中有cmp指令，从true分支跳入后遇到if块
    CondBrInstruction* cond_incond2 = dynamic_cast<CondBrInstruction*>(find_inst(Instruction::COND, cond2));
    if(!cond_incond2)return false;
    BasicBlock* ifbb = cond_incond2->getTrueBranch();
    debugmul("ifbb no = "<<ifbb->getNo());

    // 3：if块中有一条比eq的cmp, 从true分支跳入后遇到有iter的块，从false分支跳入后遇到cond3
    CmpInstruction* cmp_inifbb = dynamic_cast<CmpInstruction*>(find_inst(Instruction::CMP, ifbb));
    if(!cmp_inifbb)return false;
    if(cmp_inifbb->getOpcode() != CmpInstruction::E)return false;
    CondBrInstruction* cond_inifbb = dynamic_cast<CondBrInstruction*>(find_inst(Instruction::COND, ifbb));
    if(!cond_inifbb)return false;
    BasicBlock* if_truebb = cond_inifbb->getTrueBranch();
    BasicBlock* if_falsebb = cond_inifbb->getFalseBranch();
    if(find_inst(Instruction::BINARY, if_truebb) == nullptr)return false;
    debugmul("if_truebb->getSuccBlock()[0] no = "<<if_truebb->getSuccBlock()[0]->getNo());
    if(if_truebb->getSuccBlock()[0] != cond2)return false;
    BasicBlock* cond3 = if_falsebb;

    // 4: cond3中有cmp指令，从true分支进入body块
    CmpInstruction* cmp_incond3 = dynamic_cast<CmpInstruction*>(find_inst(Instruction::CMP, cond3));
    if(!cmp_incond3)return false;
    CondBrInstruction* cond_incond3 = dynamic_cast<CondBrInstruction*>(find_inst(Instruction::COND, cond3));
    if(!cond_incond3)return false;
    BasicBlock* bodybb = cond_incond3->getTrueBranch();

    // 5：body块中有三组gep指令，后两条的dst相乘后与第一个gep的dst相加，存回第四条gep的dst
    LoadInstruction* load_aftergep1 = nullptr;
    LoadInstruction* load_aftergep2 = nullptr;
    LoadInstruction* load_aftergep3 = nullptr;
    StoreInstruction* store_aftergep4 = nullptr;
    BinaryInstruction* mul = nullptr;
    BinaryInstruction* add_aftermul = nullptr;

    std::vector<Instruction*> geps = find_insts(Instruction::GEP, bodybb);
    if(geps.size() != 8)return false;
    load_aftergep1 = dynamic_cast<LoadInstruction*>(geps[1]->getNext());
    assert(load_aftergep1);
    load_aftergep2 = dynamic_cast<LoadInstruction*>(geps[3]->getNext());
    assert(load_aftergep2);
    load_aftergep3 = dynamic_cast<LoadInstruction*>(geps[5]->getNext());
    assert(load_aftergep3);
    store_aftergep4 = dynamic_cast<StoreInstruction*>(geps[7]->getNext());
    assert(store_aftergep4);
    std::vector<Instruction*> binarys = find_insts(Instruction::BINARY, bodybb);
    // 找到那条mul指令
    for(auto bin = binarys.begin(); bin != binarys.end(); bin ++)
    {
        if((*bin)->getOpcode() == BinaryInstruction::MUL)
        {
            mul = dynamic_cast<BinaryInstruction*>(*bin);
        }
    }
    assert(mul);
    add_aftermul = dynamic_cast<BinaryInstruction*>(mul->getNext());
    assert(add_aftermul);
    // 判断乘加关系
    debugmul("mul->getOp()[1] label = "<<mul->getOp()[1]->toStr());
    debugmul("load_aftergep2->getOp()[0] label = "<<load_aftergep2->getOp()[0]->toStr());
    debugmul("mul->getOp()[2] label = "<<mul->getOp()[2]->toStr());
    debugmul("load_aftergep3->getOp()[0] label = "<<load_aftergep3->getOp()[0]->toStr());
    if(!((mul->getOp()[1] == load_aftergep2->getOp()[0] && mul->getOp()[2] == load_aftergep3->getOp()[0])
        || (mul->getOp()[1] == load_aftergep3->getOp()[0] && mul->getOp()[2] == load_aftergep2->getOp()[0])))
        return false;
    debugmul("add_aftermul->getOp()[1] label = "<<add_aftermul->getOp()[1]->toStr());
    debugmul("mul->getOp()[0] label = "<<mul->getOp()[0]->toStr());
    debugmul("add_aftermul->getOp()[2] label = "<<add_aftermul->getOp()[2]->toStr());
    debugmul("load_aftergep1->getOp()[0] label = "<<load_aftergep1->getOp()[0]->toStr());
    if(!((add_aftermul->getOp()[1] == mul->getOp()[0] && add_aftermul->getOp()[2] == load_aftergep1->getOp()[0]) 
        || (add_aftermul->getOp()[1] == load_aftergep1->getOp()[0] && add_aftermul->getOp()[2] == mul->getOp()[0])))
        return false;
    
    // 获取其他关键块
    BasicBlock* iter21 = if_truebb;
    CondBrInstruction* cond_inbody = dynamic_cast<CondBrInstruction*>(find_inst(Instruction::COND, bodybb));
    assert(cond_inbody);
    BasicBlock* iter22 = cond_inbody->getFalseBranch();
    if(find_inst(Instruction::CMP, iter22) == nullptr)
    {
        iter22 = iter22->getSuccBlock()[0];
    }
    if(find_inst(Instruction::CMP, iter22) == nullptr)return false;

    CondBrInstruction* cond_initer22 = dynamic_cast<CondBrInstruction*>(find_inst(Instruction::COND, iter22));
    assert(cond_initer22);
    BasicBlock* iter1 = cond_initer22->getFalseBranch();
    if(find_inst(Instruction::CMP, iter1) == nullptr)
    {
        iter1 = iter1->getSuccBlock()[0];
    }
    if(find_inst(Instruction::CMP, iter1) == nullptr)return false;
    // printf("checkfunc2 end\n");
    return true;
}

bool Multithread::checkFunc3(Function *func, bool call_multithread) {
    debugmul("multithread::checkfunc3 begin");
    // TODO: 判断该函数中是否包含矩阵乘法，并将相关块指针交给doMultithread
    // 1：判断转置循环结构
    // 1-1：tbody块中有四条gep指令
    // 1-1：并将前两条gep得到的地址中的值store到后两条gep得到的地址中
    // 1-1：且他们的下标恰好相反，并记录这两个数组的基地址operand
    // 1-1：这个tbody块false后继的false后继直接跳往cond11块
    // 2：判断矩阵乘法结构
    // 2-1：cond11块-->cond21块-->cond31块-->body块(uncondbr + phi)
    // 2-1：body块中有两组gep，分别获取上面存储的两个数组基址下的相关地址
    // 2-1：获取的数组下标之间有联系（k相同）
    // 2-1：load得到结果做乘法
    // 2-1：乘法结果移到一个临时变量中，这个临时变量还用phi做过累加
    // 2-1：body块的false后继中也有一组gep获取结果数组地址，并将乘法结果存入，注意下标联系
    
    BasicBlock* entry = func->getEntry();

    UncondBrInstruction* br_inst1 = dynamic_cast<UncondBrInstruction*>(find_inst(Instruction::UNCOND, entry));
    if(!br_inst1) return false;
    BasicBlock* temp_bb1 = br_inst1->getBranch(); //B155
    debugmul("temp_bb1: " << temp_bb1->getNo());

    CondBrInstruction* cbr_inst1 = dynamic_cast<CondBrInstruction*>(find_inst(Instruction::COND, temp_bb1));
    if(!cbr_inst1) return false;
    BasicBlock* temp_bb2 = cbr_inst1->getFalseBranch(); //B166
    debugmul("temp_bb2: " << temp_bb2->getNo());

    CondBrInstruction* cbr_inst2 = dynamic_cast<CondBrInstruction*>(find_inst(Instruction::COND, temp_bb2));
    if(!cbr_inst2) return false;
    BasicBlock* temp_bb3 = cbr_inst2->getFalseBranch(); //B171
    debugmul("temp_bb3: " << temp_bb3->getNo());

    UncondBrInstruction* cbr_inst3 = dynamic_cast<UncondBrInstruction*>(find_inst(Instruction::UNCOND, temp_bb3));
    if(!cbr_inst3) return false;
    BasicBlock* temp_bb4 = cbr_inst3->getBranch(); //B156
    debugmul("temp_bb4: " << temp_bb4->getNo());

    UncondBrInstruction* br_inst4 = dynamic_cast<UncondBrInstruction*>(find_inst(Instruction::UNCOND, temp_bb4));
    if(!br_inst4) return false;
    BasicBlock* temp_bb5 = br_inst4->getBranch(); //B173
    debugmul("temp_bb5: " << temp_bb5->getNo());

    UncondBrInstruction* br_inst5 = dynamic_cast<UncondBrInstruction*>(find_inst(Instruction::UNCOND, temp_bb5));
    if(!br_inst5) return false;
    BasicBlock* transbb = br_inst5->getBranch(); //B179
    debugmul("transbb: " << transbb->getNo());

    int gep_nums = 0;
    GepInstruction* gep_inst0;
    GepInstruction* gep_inst1;
    GepInstruction* gep_inst2;
    GepInstruction* gep_inst3;
    for(auto inst = transbb->begin(); inst != transbb->end(); inst = inst->getNext()) {
        if(inst->isGep()) {
            switch (gep_nums) {
            case 0:
                gep_inst0 = dynamic_cast<GepInstruction*>(inst);
                gep_nums++;
                break;
            case 1:
                gep_inst1 = dynamic_cast<GepInstruction*>(inst);
                gep_nums++;
                break;
            case 2:
                gep_inst2 = dynamic_cast<GepInstruction*>(inst);
                gep_nums++;
                break;
            case 3:
                gep_inst3 = dynamic_cast<GepInstruction*>(inst);
                gep_nums++;
                break;
            default:
                return false;
            }
        }
    }

    if(!dynamic_cast<GepInstruction* >(gep_inst0)->getBase()->isGlobal()) return false;
    if(!dynamic_cast<GepInstruction* >(gep_inst2)->getBase()->isGlobal()) return false;

    
    //做转置的两个数组:
    if(gep_inst0->getUse()[1] != gep_inst3->getUse()[1]) return false;
    if(gep_inst1->getUse()[1] != gep_inst2->getUse()[1]) return false;

    // debugmul(gep_inst0->getBase()->getEntry()->toStr();
    // debugmul(SE1->toStr();
    // auto SE1 = gep_inst1->getBase()->getEntry();
    // auto SE2 = gep_inst2->getBase()->getEntry();

    if(gep_inst0->getDef() != gep_inst1->getBase()) return false;
    if(gep_inst2->getDef() != gep_inst3->getBase()) return false;

    //TODO: getUse()[0]可能会报错;
    auto op_load = gep_inst1->getDef()->getUse()[0]->getDef();
    if(op_load->getUse()[0]->getDef() != gep_inst3->getDef()) return false; //load => store.

    debugmul("transpose in: " << transbb->getNo());


    CondBrInstruction* cbr_inst6 = dynamic_cast<CondBrInstruction*>(find_inst(Instruction::COND, transbb));
    if(!cbr_inst6) return false;
    BasicBlock* temp_bb6 = cbr_inst6->getFalseBranch(); //B191
    debugmul("temp_bb6: " << temp_bb6->getNo());

    CondBrInstruction* cbr_inst7 = dynamic_cast<CondBrInstruction*>(find_inst(Instruction::COND, temp_bb6));
    if(!cbr_inst7) return false;
    BasicBlock* temp_bb7 = cbr_inst7->getFalseBranch(); //B196
    debugmul("temp_bb7: " << temp_bb7->getNo());

    UncondBrInstruction* br_inst8 = dynamic_cast<UncondBrInstruction*>(find_inst(Instruction::UNCOND, temp_bb7));
    if(!br_inst8) return false;
    BasicBlock* temp_bb8 = br_inst8->getBranch(); //B174
    debugmul("temp_bb8: " << temp_bb8->getNo());

    UncondBrInstruction* br_inst9 = dynamic_cast<UncondBrInstruction*>(find_inst(Instruction::UNCOND, temp_bb8));
    if(!br_inst9) return false;
    BasicBlock* temp_bb9 = br_inst9->getBranch(); //B198
    debugmul("temp_bb9: " << temp_bb9->getNo());

    UncondBrInstruction* br_inst10 = dynamic_cast<UncondBrInstruction*>(find_inst(Instruction::UNCOND, temp_bb9));
    if(!br_inst10) return false;
    BasicBlock* temp_bb10 = br_inst10->getBranch(); //B198
    debugmul("temp_bb9: " << temp_bb9->getNo());

    UncondBrInstruction* br_inst11 = dynamic_cast<UncondBrInstruction*>(find_inst(Instruction::UNCOND, temp_bb10));
    if(!br_inst11) return false;
    BasicBlock* matmul_bb = br_inst11->getBranch(); //B198
    debugmul("matmul_bb: " << matmul_bb->getNo());

    int gep_nums0 = 0;
    GepInstruction* gep_inst00;
    GepInstruction* gep_inst01;
    GepInstruction* gep_inst02;
    GepInstruction* gep_inst03;
    for(auto inst = matmul_bb->begin(); inst != matmul_bb->end(); inst = inst->getNext()) {
        if(inst->isGep()) {
            switch (gep_nums0) {
            case 0:
                gep_inst00 = dynamic_cast<GepInstruction*>(inst);
                gep_nums0++;
                break;
            case 1:
                gep_inst01 = dynamic_cast<GepInstruction*>(inst);
                gep_nums0++;
                break;
            case 2:
                gep_inst02 = dynamic_cast<GepInstruction*>(inst);
                gep_nums0++;
                break;
            case 3:
                gep_inst03 = dynamic_cast<GepInstruction*>(inst);
                gep_nums0++;
                break;
            default:
                return false;
            }
        }
    }

    if(!dynamic_cast<GepInstruction* >(gep_inst00)->getBase()->isGlobal()) return false;
    if(!dynamic_cast<GepInstruction* >(gep_inst02)->getBase()->isGlobal()) return false;

    if(gep_inst00->getDef() != gep_inst01->getBase()) return false;
    if(gep_inst02->getDef() != gep_inst03->getBase()) return false;

    if(gep_inst00->getBase()->getSymbolEntry() != gep_inst0->getBase()->getSymbolEntry()) return false;
    if(gep_inst02->getBase()->getSymbolEntry() != gep_inst2->getBase()->getSymbolEntry()) return false;

    if(gep_inst01->getUse()[1] != gep_inst02->getUse()[1]) return false;

    // debugmul(gep_inst00->getBase()->getEntry()->toStr();
    // debugmul(SE1->toStr();
    // if(gep_inst00->getBase()->getEntry() != SE1) return false;
    // if(gep_inst02->getBase()->getEntry() != SE2) return false;


    //TODO: getUse()[0]可能会报错;
    Instruction* _mul1 = gep_inst01->getDef()->getUse()[0]->getDef()->getUse()[0];
    Instruction* _mul2 = gep_inst03->getDef()->getUse()[0]->getDef()->getUse()[0];

    if(!_mul1->isBin() || !_mul2->isBin()) return false;
    if(!dynamic_cast<BinaryInstruction* >(_mul1)->isMul() || !dynamic_cast<BinaryInstruction* >(_mul2)->isMul()) return false;
    if(_mul1 != _mul2) return false;

    //结果累加:
    PhiInstruction* phi_inst1 = dynamic_cast<PhiInstruction*>(find_inst(Instruction::PHI, matmul_bb));
    if(!phi_inst1) return false;

    Operand* sum_op = phi_inst1->getSrc(matmul_bb);

    //TODO: getUse()[0]可能会报错;
    Operand* _sum_op = _mul1->getDef()->getUse()[0]->getDef();
    if(sum_op != _sum_op) return false;

    //下标循环:
    Instruction* phi_inst2 = phi_inst1->getNext();
    if(!phi_inst2 || !phi_inst2->isPhi()) return false;

    if(phi_inst2->getDef() != gep_inst01->getUse()[1]) return false;


    debugmul("matmul in: " << matmul_bb->getNo());

    //结果写回:
    CondBrInstruction* cbr_inst12 = dynamic_cast<CondBrInstruction*>(find_inst(Instruction::COND, matmul_bb));
    if(!cbr_inst12) return false;
    BasicBlock* wback_bb = cbr_inst12->getFalseBranch(); //B224
    debugmul("wback_bb: " << wback_bb->getNo());

    PhiInstruction* phi_inst3 = dynamic_cast<PhiInstruction*>(find_inst(Instruction::PHI, wback_bb));
    if(!phi_inst3) return false;

    Operand* __sum_op = phi_inst3->getSrc(wback_bb);
    if(sum_op != __sum_op) return false;

    int gep_nums00 = 0;
    GepInstruction* gep_inst000;
    GepInstruction* gep_inst001;
    for(auto inst = wback_bb->begin(); inst != wback_bb->end(); inst = inst->getNext()) {
        if(inst->isGep()) {
            switch (gep_nums00) {
            case 0:
                gep_inst000 = dynamic_cast<GepInstruction*>(inst);
                gep_nums00++;
                break;
            case 1:
                gep_inst001 = dynamic_cast<GepInstruction*>(inst);
                gep_nums00++;
                break;
            default:
                return false;
            }
        }
    }

    if(gep_inst000->getDef() != gep_inst001->getBase()) return false;

    if(!(gep_inst001->getDef()->getUse()[0])->isStore()) return false;
    if(dynamic_cast<StoreInstruction* >(gep_inst001->getDef()->getUse()[0])->getSrc() != phi_inst3->getDef()) return false;

    debugmul("true");

    // body块的指针:
    debugmul("matmul_bb: " << matmul_bb->getNo());
    // body块中四条gep的vector数组
    std::vector<GepInstruction* > geps;
    geps.push_back(gep_inst00);
    geps.push_back(gep_inst01);
    geps.push_back(gep_inst02);
    geps.push_back(gep_inst03);
    // body块的false后继
    debugmul("wback_bb: " << wback_bb->getNo());
    // c数组的两条gep指令
    debugmul("c_gep1: " << gep_inst000->getInstName());
    debugmul("c_gep2: " << gep_inst001->getInstName());
    std::vector<GepInstruction* > geps2;
    geps2.push_back(gep_inst000);
    geps2.push_back(gep_inst001);

    // 需要返回：
    // body块的指针
    // body块中四条gep的vector数组
    // body块的false后继
    // c数组的两条gep指令
    // cond11块指针
    // iter12块指针
    BasicBlock* iter12 = wback_bb->getSuccBlock()[1];// 这里不要让他get到自己
    assert(iter12);
    if(find_inst(Instruction::CMP, iter12) == nullptr)
    {
        iter12 = iter12->getSuccBlock()[0];
    }
    assert(iter12);
    assert(find_inst(Instruction::CMP, iter12));

    printf("checkfunc3 end\n");
    if(call_multithread)
        doMultithread2(func, matmul_bb, geps, wback_bb, geps2, temp_bb8, iter12);
    return true;
}

void setPreSucRelation(std::vector<BasicBlock*> bbs);

Operand* get_n(BasicBlock* bb);

int findN_frombody(BasicBlock* bb);

Operand* get_B(BasicBlock* bb);

Operand* get_iter(BasicBlock* bb);

Operand* get_iter2(BasicBlock* bb);

void install_pthread_sysfunc();

void tellPredModifySucc(BasicBlock* bb, BasicBlock* nextbb);

Type* gettype(Instruction* ins, int index);

BinaryInstruction* find_mul(BasicBlock* bb);

void Multithread::doMultithread(std::vector<BasicBlock*> blks, Function *func)
{
    debugmul("multithread::doMultithread begin");
    install_pthread_sysfunc();
    this->multithread_who = 1;
    BasicBlock* cond1 = blks[0];
    debugmul("cond1 = "<<cond1->getNo());
    BasicBlock* cond2 = blks[1];
    debugmul("cond2 = "<<cond2->getNo());
    BasicBlock* ifbb = blks[2];
    debugmul("ifbb = "<<ifbb->getNo());
    BasicBlock* cond3 = blks[3];
    debugmul("cond3 = "<<cond3->getNo());
    BasicBlock* bodybb = blks[4];
    debugmul("bodybb = "<<bodybb->getNo());
    BasicBlock* iter21 = blks[5];
    debugmul("iter21 = "<<iter21->getNo());
    BasicBlock* iter22 = blks[6];
    debugmul("iter22 = "<<iter22->getNo());
    BasicBlock* iter1 = blks[7];
    debugmul("iter1 = "<<iter1->getNo());
    // TODO: 加入转置循环, B矩阵转置为D矩阵，计算栈偏移
    // 1-1: 加个D数组的alloca
    // new alloca指令，新建D的1024*1024整型数组
    // 获取1024
    int N = findN_frombody(bodybb);
    Type* typeD = new ArrayType(TypeSystem::intType, N);
    Type* typeDD = new ArrayType(typeD, N);
    int scope = 0;
    SymbolEntry* seD = new IdentifierSymbolEntry(typeDD, "D", scope);
    // dynamic_cast<IdentifierSymbolEntry*>(seD)->setLocal();//?
    Operand *addr_D;
    SymbolEntry *addr_seD;
    Type *typeD2;
    typeD2 = new PointerType(seD->getType());
    addr_seD = new TemporarySymbolEntry(typeD2, SymbolTable::getLabel());
    addr_D = new Operand(addr_seD);
    addr_D->setOpdSe(seD);

    AllocaInstruction* allocaD = new AllocaInstruction(addr_D, seD);
    func->getEntry()->addAlloc(allocaD);
    
    // 新建块
    BasicBlock* tcond1 = new BasicBlock(func);
    debugmul("tcond1 = "<<tcond1->getNo());
    BasicBlock* tcond21 = new BasicBlock(func);
    debugmul("tcond21 = "<<tcond21->getNo());
    BasicBlock* tcond22 = new BasicBlock(func);
    debugmul("tcond22 = "<<tcond22->getNo());
    BasicBlock* tbody = new BasicBlock(func);
    debugmul("tbody = "<<tbody->getNo());
    BasicBlock* tend = new BasicBlock(func);
    debugmul("tend = "<<tend->getNo());

    // 1-2: 新建tcond1块
    // 获取n
    Operand* n = get_n(cond1);
    debugmul("n = "<<n->toStr());
    debugmul("n type = "<<n->getType()->toStr());
    // dst_load_intcond1 = load i32, i32* n
    Operand* dst_load_intcond1 = new Operand(new TemporarySymbolEntry(TypeSystem::intType, SymbolTable::getLabel()));
    LoadInstruction* load_intcond1 = new LoadInstruction(dst_load_intcond1, n, tcond1);
    // dst_cmp_intcond1 = icmp slt i32 0, dst_load_intcond1
    Operand* dst_cmp_intcond1 = new Operand(new TemporarySymbolEntry(TypeSystem::boolType, SymbolTable::getLabel()));
    CmpInstruction* cmp_intcond1 = new CmpInstruction(CmpInstruction::L, dst_cmp_intcond1,
        new Operand(new ConstantSymbolEntry(TypeSystem::intType, 0)),
        dst_load_intcond1, tcond1);
    // br i1 dst_cmp_intcond1, label tcond21, label tend
    CondBrInstruction* condbr_intcond1 = new CondBrInstruction(tcond21, tend, dst_cmp_intcond1, tcond1);

    // 1-3: 新建tcond21块
    // i1 = phi i32[0, tcond1], [i2, tcond22]
    Operand* i1 = new Operand(new TemporarySymbolEntry(TypeSystem::intType, SymbolTable::getLabel()));
    debugmul("i1 = "<<i1->toStr());
    Operand* i2 = new Operand(new TemporarySymbolEntry(TypeSystem::intType, SymbolTable::getLabel()));
    debugmul("i2 = "<<i2->toStr());
    PhiInstruction* phi_intcond1 = new PhiInstruction(i1, tcond21);
    phi_intcond1->addSrc(tcond1, new Operand(new ConstantSymbolEntry(TypeSystem::intType, 0)));
    phi_intcond1->addSrc(tcond22, i2);
    // dst_load_intcond21 = load i32, i32* n
    Operand* dst_load_intcond21 = new Operand(new TemporarySymbolEntry(TypeSystem::intType, SymbolTable::getLabel()));
    LoadInstruction* load_intcond21 = new LoadInstruction(dst_load_intcond21, n, tcond21);
    // dst_cmp_intcond21 = icmp slt i32 0, dst_load_intcond21
    Operand* dst_cmp_intcond21 = new Operand(new TemporarySymbolEntry(TypeSystem::boolType, SymbolTable::getLabel()));
    CmpInstruction* cmp_intcond21 = new CmpInstruction(CmpInstruction::L, dst_cmp_intcond21,
        new Operand(new ConstantSymbolEntry(TypeSystem::intType, 0)),
        dst_load_intcond21, tcond21);
    // br i1 dst_cmp_intcond21, label tbody, label tcond22
    CondBrInstruction* condbr_intcond21 = new CondBrInstruction(tbody, tcond22, dst_cmp_intcond21, tcond21);

    // 1-4: 新建tcond22块
    // i2 = add i32 i1, 1
    BinaryInstruction* add_intcond22 = new BinaryInstruction(BinaryInstruction::ADD, i2, i1,
        new Operand(new ConstantSymbolEntry(TypeSystem::intType, 1)), tcond22);
    // dst_load_intcond22 = load i32, i32* n
    Operand* dst_load_intcond22 = new Operand(new TemporarySymbolEntry(TypeSystem::intType, SymbolTable::getLabel()));
    LoadInstruction* load_intcond22 = new LoadInstruction(dst_load_intcond22, n, tcond22);
    // dst_cmp_intcond22 = icmp slt i32 i2, dst_load_intcond22
    Operand* dst_cmp_intcond22 = new Operand(new TemporarySymbolEntry(TypeSystem::boolType, SymbolTable::getLabel()));
    CmpInstruction* cmp_intcond22 = new CmpInstruction(CmpInstruction::L, dst_cmp_intcond22,
        i2, dst_load_intcond22, tcond22);
    // br i1 dst_cmp_intcond22, label tcond21, label tend
    CondBrInstruction* condbr_intcond22 = new CondBrInstruction(tcond21, tend, dst_cmp_intcond22, tcond22);

    // 1-5: 新建tbody块
    // j1 = phi i32[0, tcond21], [j2, tbody]
    Operand* j1 = new Operand(new TemporarySymbolEntry(TypeSystem::intType, SymbolTable::getLabel()));
    Operand* j2 = new Operand(new TemporarySymbolEntry(TypeSystem::intType, SymbolTable::getLabel()));
    PhiInstruction* phi_intbody = new PhiInstruction(j1, tbody);
    phi_intbody->addSrc(tcond21, new Operand(new ConstantSymbolEntry(TypeSystem::intType, 0)));
    phi_intbody->addSrc(tbody, j2);
    // dst_loadB_intbody = load [1024xi32]*, [1024xi32]** B
    Operand* addr_B = get_B(bodybb);
    debugmul("addr_B type = "<<addr_B->getType()->toStr());
    debugmul("addr_B label = "<<addr_B->toStr());
    Type* type_B1 = addr_B->getType();
    assert(dynamic_cast<PointerType*>(type_B1));
    Type* type_B2 = dynamic_cast<PointerType*>(type_B1)->getType();
    assert(type_B2);
    debugmul("type_B2 = "<<type_B2->toStr());
    // debugmul("type_B2->isPtr() = "<<type_B2->isPtr());
    // debugmul("type_B2->isArray() = "<<type_B2->isArray());
    // debugmul("type_B2->isPtr2Array() = "<<type_B2->isPtr2Array());
    assert(dynamic_cast<ArrayType*>(type_B2));
    Type* type_B3 = dynamic_cast<ArrayType*>(type_B2)->getElementType();
    debugmul("type_B3 = "<<type_B3->toStr());
    Operand* dst_loadB_intbody = new Operand(new TemporarySymbolEntry(type_B2, SymbolTable::getLabel()));
    LoadInstruction* loadB_intbody = new LoadInstruction(dst_loadB_intbody, addr_B, tbody);
    // dst_gepB1 = gep [1024xi32], [1024xi32]* dst_loadB_intbody, i32 j1
    Operand* dst_gepB1 = new Operand(new TemporarySymbolEntry(type_B2, SymbolTable::getLabel()));
    GepInstruction* gepB1 = new GepInstruction(dst_gepB1, dst_loadB_intbody, j1, tbody);
    gepB1->setparamFirst();
    debugmul("checkgep arr dst_loadB_intbody type = "<<dst_loadB_intbody->getType()->toStr());
    // dst_gepB2 = gep [1024xi32], [1024xi32]* dst_gepB1, i32 0, i32 i1
    Operand* dst_gepB2 = new Operand(new TemporarySymbolEntry(new PointerType(dynamic_cast<ArrayType*>(type_B3)->getElementType()), SymbolTable::getLabel()));
    GepInstruction* gepB2 = new GepInstruction(dst_gepB2, dst_gepB1, i1, tbody);
    debugmul("checkgep arr dst_gepB1 type = "<<dst_gepB1->getType()->toStr());
    // dst_loadBji = load i32, i32* dst_gepB2
    Operand* dst_loadBji = new Operand(new TemporarySymbolEntry(TypeSystem::intType, SymbolTable::getLabel()));
    LoadInstruction* loadBji_intbody = new LoadInstruction(dst_loadBji, dst_gepB2, tbody);
    // dst_gepD1 = gep [1024x[1024xi32]], [1024x[1024xi32]]* addr_D, i32 0, i32 i1
    Type* type_D1 = addr_D->getType();
    debugmul("type_D1 = "<<type_D1->toStr());
    assert(dynamic_cast<PointerType*>(type_D1));
    Type* type_D2 = dynamic_cast<PointerType*>(type_D1)->getType();
    debugmul("type_D2 = "<<type_D2->toStr());
    assert(dynamic_cast<ArrayType*>(type_D2));
    Type* type_D3 = dynamic_cast<ArrayType*>(type_D2)->getElementType();
    debugmul("type_D3 = "<<type_D3->toStr());
    ArrayType* type_D4 = dynamic_cast<ArrayType*>(type_D3);
    Operand* dst_gepD1 = new Operand(new TemporarySymbolEntry(new PointerType(type_D3), SymbolTable::getLabel()));
    GepInstruction* gepD1 = new GepInstruction(dst_gepD1, addr_D, i1, tbody);
    gepD1->setFirst();
    debugmul("checkgep arr addr_D type = "<<addr_D->getType()->toStr());
    // dst_gepD2 = gep [1024xi32], [1024xi32]* dst_gepD1, i32 0, i32 j1
    Operand* dst_gepD2 = new Operand(new TemporarySymbolEntry(new PointerType(type_D4->getElementType()), SymbolTable::getLabel()));
    GepInstruction* gepD2 = new GepInstruction(dst_gepD2, dst_gepD1, j1, tbody);
    debugmul("checkgep arr dst_gepD1 type = "<<dst_gepD1->getType()->toStr());
    // store i32 dst_loadBji, i32* dst_gepD2
    StoreInstruction* store_intbody = new StoreInstruction(dst_gepD2, dst_loadBji, tbody);
    // j2  = add i32 j1, 1
    BinaryInstruction* add_intbody = new BinaryInstruction(BinaryInstruction::ADD, j2, j1,
        new Operand(new ConstantSymbolEntry(TypeSystem::intType, 1)), tbody);
    // dst_loadn_intbody = load i32, i32* n
    Operand* dst_loadn_intbody = new Operand(new TemporarySymbolEntry(TypeSystem::intType, SymbolTable::getLabel()));
    LoadInstruction* loadn_intbody = new LoadInstruction(dst_loadn_intbody, n, tbody);
    // dst_cmp_intbody = icmp slt i32 j2, dst_loadn_intbody
    Operand* dst_cmp_intbody = new Operand(new TemporarySymbolEntry(TypeSystem::boolType, SymbolTable::getLabel()));
    CmpInstruction* cmp_intbody = new CmpInstruction(CmpInstruction::L, dst_cmp_intbody,
        j2, dst_loadn_intbody, tbody);
    // br i1 dst_cmp_intbody, label tbody, label tcond22
    CondBrInstruction* condbr_intbody = new CondBrInstruction(tbody, tcond22, dst_cmp_intbody, tbody);
    // 1-6: 新建tend块
    UncondBrInstruction* uncond_intend = new UncondBrInstruction(cond1, tend);

    // 1-7: 处理块之间前驱后继关系
    // 让cond1的前驱块跳转到tcond1
    tellPredModifySucc(cond1, tcond1);
    // BasicBlock* cond1_pre = cond1->getPredBlock()[0];
    // debugmul("cond1_pre = "<<cond1_pre->getNo());
    // assert(cond1_pre->getPredBlock()[0] == func->getEntry());
    // UncondBrInstruction* uncond_incond1pre = dynamic_cast<UncondBrInstruction*>(find_inst(Instruction::UNCOND, cond1_pre));
    // assert(uncond_incond1pre);
    // uncond_incond1pre->setBranch(tcond1);

    // 集中调整转置部分块的前驱、后继关系
    std::vector<BasicBlock*> t_newbbs;
    t_newbbs.push_back(tcond1);
    t_newbbs.push_back(tcond21);
    t_newbbs.push_back(tcond22);
    t_newbbs.push_back(tbody);
    t_newbbs.push_back(tend);
    setPreSucRelation(t_newbbs);

    // 首尾两个
    // cond1_pre->removeSucc(cond1);
    // cond1->removePred(cond1_pre);
    // cond1_pre->addSucc(tcond1);
    // tcond1->addPred(cond1_pre);
    // wzr:
    // tend->addSucc(cond1);
    // cond1->addPred(tend);
    
    // 1-8: 处理块内use-def关系

    // TODO: 矩阵乘法的最内层循环，k,i,j替换以及B矩阵替换成D
    // 获取k1, i1, j1
    Operand* k1 = get_iter(iter1);
    assert(k1);
    // Operand* i1 = get_iter(iter22);
    i1 = get_iter(iter22);
    debugmul("iter22 i1 = "<<i1->toStr());
    assert(i1);
    // Operand* j1 = get_iter(bodybb);
    j1 = get_iter(bodybb);
    assert(j1);
    // 替换
    std::vector<Instruction*> geps = find_insts(Instruction::GEP, bodybb);
    assert(geps.size() == 8);
    geps[0]->setOperand2(2, k1);
    geps[1]->setOperand2(2, i1);
    geps[2]->setOperand2(2, k1);
    geps[3]->setOperand2(2, j1);
    geps[4]->setOperand2(2, i1);
    geps[5]->setOperand2(2, j1);
    geps[6]->setOperand2(2, k1);
    geps[7]->setOperand2(2, i1);
    // B->D
    // LoadInstruction* load_before_gep4 = dynamic_cast<LoadInstruction*>(geps[4]->getPrev());
    // assert(load_before_gep4);
    // load_before_gep4->setOperand2(1, addr_D);
    geps[4]->setOperand2(1, addr_D);
    // wzr: 若D矩阵的类型有些问题，在这里调整
    dynamic_cast<GepInstruction*>(geps[4])->setFirst();
    dynamic_cast<GepInstruction*>(geps[4])->setparamFirst_false();
    // TODO: 删除if分支
    // cond2 --> ifbb 变为 cond2 --> cond3
    CondBrInstruction* condbr_incond2 = dynamic_cast<CondBrInstruction*>(find_inst(Instruction::COND, cond2));
    assert(condbr_incond2);
    assert(condbr_incond2->getTrueBranch() == ifbb);
    condbr_incond2->setTrueBranch(cond3);
    cond2->removeSucc(ifbb);
    ifbb->removePred(cond2);
    cond2->addSucc(cond3);
    cond3->addPred(cond2);
    // PhiInstruction* phi_incond3 = dynamic_cast<PhiInstruction*>(find_inst(Instruction::phi, cond3));
    // if(phi_incond3)
    // {
    //     phi_incond3->changeSrcBlock2(ifbb, cond2);
    // }
    debugmul("delete if1");
    // iter22 --> ifbb 变为 iter22 --> cond3
    CondBrInstruction* condbr_initer22 = dynamic_cast<CondBrInstruction*>(find_inst(Instruction::COND, iter22));
    assert(condbr_initer22);
    assert(condbr_initer22->getTrueBranch() == ifbb);
    condbr_initer22->setTrueBranch(cond3);
    iter22->removeSucc(ifbb);
    ifbb->removePred(iter22);
    iter22->addSucc(cond3);
    cond3->addPred(iter22);
    // if(phi_incond3)
    // {
    //     phi_incond3->changeSrcBlock2(ifbb, iter22);
    // }
    // cond2->removePred(ifbb);
    // ifbb->removeSucc(cond2);
    // ifbb->removeSucc(cond3);
    // cond3->removePred(ifbb);
    debugmul("delete if2");
    // ifbb 和iter21 删除所有关系
    ifbb->deleteAllRelation();
    func->remove(ifbb);
    iter21->deleteAllRelation();
    func->remove(iter21);
    debugmul("delete if3");

    // 删除ifbb后，原三重循环的i变量迭代出现一些交接问题
    // wzr: 这个写法不具有普遍性，仅仅针对个别样例
    // 强行设置i1
    PhiInstruction* phi_incond2 = nullptr;
    if(find_inst(Instruction::PHI, cond2))
    {
        debugmul("set phi dst");
        phi_incond2 = dynamic_cast<PhiInstruction*>(find_inst(Instruction::PHI, cond2));
        phi_incond2->setOperand2(0, i1);
    }
    assert(phi_incond2);
    if(find_inst(Instruction::CMP, cond2))
    {
        debugmul("set cmp op1");
        find_inst(Instruction::CMP, cond2)->setOperand2(1, i1);
    }

    // 转移phi
    cond2->remove(phi_incond2);
    cond3->insertFront(phi_incond2);
    debugmul("cond1->getSuccBlock()[0] = "<<cond1->getSuccBlock()[0]->getNo());
    debugmul("cond3->getPredBlock()[0] = "<<cond3->getPredBlock()[0]->getNo());
    phi_incond2->changeSrcBlock2(cond1->getSuccBlock()[0], cond3->getPredBlock()[0]);
    phi_incond2->changeSrcBlock2(iter21, iter22);
    auto phi_it = phi_incond2->getSrcs().begin();
    if(phi_it->second->isConst()){
        phi_it ++;
    }
    debugmul("phi_it->second = "<<phi_it->second->toStr());
    debugmul("get_iter2(iter22) = "<<get_iter2(iter22)->toStr());
    phi_incond2->changeSrcOperand(phi_it->second, get_iter2(iter22));
    // return;

    
    // TODO: 加入多线程跳转语句，调整iter,自旋，加入多线程汇编
    // 1: 加入bl pthread_create
    SymbolEntry* se_pthread_create = identifiers->lookup("pthread_create");
    debugmul("se_pthread_create = "<<se_pthread_create->toStr());
    std::vector<Operand*> ops_pcreate;
    CallInstruction* call_pcreate = new CallInstruction(
        nullptr, se_pthread_create, ops_pcreate, nullptr);
    cond1->insertFront(call_pcreate);
    // 2: 移入r10
    CmpInstruction* cmp_incond1 = dynamic_cast<CmpInstruction*>(find_inst(Instruction::CMP, cond1));
    assert(cmp_incond1);
    Operand* threadid = new Operand(new TemporarySymbolEntry(TypeSystem::intType, SymbolTable::getLabel()));
    cmp_incond1->setOperand2(1, threadid);
    cmp_incond1->setaddMov(true);
    blk_needmodify.push_back(cond1);
    // getchar();
    blk_needmodify_label.push_back(1);
    // cmp_incond1->getPrev().set
    // 3: 插入bl pthread_join
    SymbolEntry* se_pthread_join = identifiers->lookup("pthread_join");
    debugmul("se_pthread_join = "<<se_pthread_join->toStr());
    std::vector<Operand*> ops_pjoin;
    CallInstruction* call_pjoin = new CallInstruction(
        nullptr, se_pthread_join, ops_pjoin, nullptr);
    CondBrInstruction* condbr_initer1 = dynamic_cast<CondBrInstruction*>(find_inst(Instruction::COND, iter1));
    assert(condbr_initer1);
    BasicBlock* iter1_succ = condbr_initer1->getFalseBranch();
    debugmul("iter1_succ = "<<iter1_succ->getNo());
    iter1_succ->insertFront(call_pjoin);
    // 4: adder 1变为线程数(4)
    CmpInstruction* cmp_initer1 = dynamic_cast<CmpInstruction*>(find_inst(Instruction::CMP, iter1));
    assert(cmp_initer1);
    BinaryInstruction* bin_initer1 = dynamic_cast<BinaryInstruction*>(cmp_initer1->getOp()[1]->getDef());
    assert(bin_initer1);
    bin_initer1->setOperand2(2, new Operand(new ConstantSymbolEntry(TypeSystem::intType, THREADNUM)));
    // 5: 最内块自旋？
    // 6: 加入汇编函数本身

    // end
    this->mUnit->setmultithread_on(true);
    
    // 
    std::vector<Instruction*> allocas = find_insts(Instruction::ALLOCA, func->getEntry());
    Instruction* last_alloca = allocas[allocas.size() - 1];
    func->getEntry()->insertAfter(allocaD, last_alloca);
    debugmul("multithread::doMultithread end");
    return;
}

void Multithread::doMultithread2(Function* func, BasicBlock* bodybb, std::vector<GepInstruction*> geps_inbodybb, 
    BasicBlock* iter22, std::vector<GepInstruction*> geps_initer22, BasicBlock* cond11, BasicBlock* iter12)
{
    debugmul("doMultithread2 begin");
    unit->set_target_func_unit(dynamic_cast<IdentifierSymbolEntry*>(func->getSymPtr()));
    this->multithread_who = 2;
    // BasicBlock* bodybb = nullptr;
    // BasicBlock* iter22 = nullptr;
    // BasicBlock* cond11 = nullptr;
    // BasicBlock* iter12 = nullptr;
    // std::vector<Instruction*> geps_inbodybb;
    // std::vector<Instruction*> geps_initer22;
    // 转换b数组的下标
    Operand* arr1 = geps_inbodybb[0]->getOp()[1];
    debugmul("arr1 = "<<arr1->toStr());
    geps_inbodybb[2]->setOperand2(1, arr1);
    Operand* index21 = geps_inbodybb[2]->getOp()[2];
    debugmul("index21 = "<<index21->toStr());
    Operand* index22 = geps_inbodybb[3]->getOp()[2];
    debugmul("index22 = "<<index22->toStr());
    geps_inbodybb[2]->setOperand2(2, index22);
    geps_inbodybb[3]->setOperand2(2, index21);
    // 累加对象从temp换成c[i][j]
    BinaryInstruction* mul_inbody = find_mul(bodybb);
    assert(find_mul);
    Operand* dst_mul_inbody = mul_inbody->getOp()[0];
    StoreInstruction* store_initer22 = dynamic_cast<StoreInstruction*>(find_inst(Instruction::STORE, iter22));
    iter22->remove(store_initer22);
    // GepInstruction* geps_initer22_1_copy = geps_initer22[1]->copy1();
    iter22->remove(geps_initer22[0]);
    iter22->remove(geps_initer22[1]);
    debugmul("bk");
    // return;

    Type* type31 = gettype(geps_inbodybb[1]->getNext(), 1);
    Operand* dst_load_arr3 = new Operand(new TemporarySymbolEntry(TypeSystem::intType, SymbolTable::getLabel()));
    LoadInstruction* load_arr3 = new LoadInstruction(dst_load_arr3, geps_initer22[1]->getOp()[0]);
    BinaryInstruction* add_to_arr3 = new BinaryInstruction(BinaryInstruction::ADD, dst_load_arr3, 
        dst_load_arr3, dst_mul_inbody);
    StoreInstruction* store_to_arr3 = new StoreInstruction(geps_initer22[1]->getOp()[0], dst_load_arr3);
    bodybb->insertAfter(store_to_arr3, mul_inbody);
    bodybb->insertAfter(add_to_arr3, mul_inbody);
    bodybb->insertAfter(load_arr3, mul_inbody);
    
    bodybb->insertAfter(geps_initer22[1]->copy1(), mul_inbody);
    bodybb->insertAfter(geps_initer22[0]->copy1(), mul_inbody);
    // ！！！请先通过正确性测试再进行下面的步骤
    // cond11中加入对pthread_create的调用
    // iter12末尾或后继中加入对pthread_join的调用，并调整累加器为线程数
    // 加入多线程汇编函数本身
    
    debugmul("cond11 = "<<cond11->getNo());
    debugmul("iter12 = "<<iter12->getNo());
    if(identifiers->lookup("pthread_create") == nullptr
        || identifiers->lookup("pthread_join") == nullptr)
    {
        install_pthread_sysfunc();
    }
    // 1: 加入bl pthread_create
    SymbolEntry* se_pthread_create = identifiers->lookup("pthread_create");
    debugmul("se_pthread_create = "<<se_pthread_create->toStr());
    std::vector<Operand*> ops_pcreate;
    CallInstruction* call_pcreate = new CallInstruction(
        nullptr, se_pthread_create, ops_pcreate, nullptr);
    cond11->insertFront(call_pcreate);
    // 2: 移入r10
    // CmpInstruction* cmp_incond11 = dynamic_cast<CmpInstruction*>(find_inst(Instruction::CMP, cond11));
    // assert(cmp_incond11);
    // Operand* threadid = new Operand(new TemporarySymbolEntry(TypeSystem::intType, SymbolTable::getLabel()));
    // cmp_incond11->setOperand2(1, threadid);
    // cmp_incond11->setaddMov(true);
    blk_needmodify.push_back(cond11);
    blk_needmodify_label.push_back(2);
    // getchar();
    // cmp_incond11->getPrev().set
    // 3: 插入bl pthread_join
    SymbolEntry* se_pthread_join = identifiers->lookup("pthread_join");
    debugmul("se_pthread_join = "<<se_pthread_join->toStr());
    std::vector<Operand*> ops_pjoin;
    CallInstruction* call_pjoin = new CallInstruction(
        nullptr, se_pthread_join, ops_pjoin, nullptr);
    CondBrInstruction* condbr_initer12 = dynamic_cast<CondBrInstruction*>(find_inst(Instruction::COND, iter12));
    assert(condbr_initer12);
    BasicBlock* iter12_succ = condbr_initer12->getFalseBranch();
    debugmul("iter12_succ = "<<iter12_succ->getNo());
    iter12_succ->insertFront(call_pjoin);
    // 4: adder 1变为线程数(4)
    CmpInstruction* cmp_initer12 = dynamic_cast<CmpInstruction*>(find_inst(Instruction::CMP, iter12));
    assert(cmp_initer12);
    BinaryInstruction* bin_initer12 = dynamic_cast<BinaryInstruction*>(cmp_initer12->getOp()[1]->getDef());
    assert(bin_initer12);
    bin_initer12->setOperand2(2, new Operand(new ConstantSymbolEntry(TypeSystem::intType, THREADNUM)));
    // 6: 加入汇编函数本身
    this->mUnit->setmultithread_on(true);
    debugmul("multithread_who1 = "<<multithread_who);
    
    // end
    debugmul("multithread2 end");
}

BinaryInstruction* find_mul(BasicBlock* bb)
{
    std::vector<Instruction*> binarys = find_insts(Instruction::BINARY, bb);
    for(auto bin = binarys.begin(); bin != binarys.end(); bin ++)
    {
        BinaryInstruction* bin_b = dynamic_cast<BinaryInstruction*>(*bin);
        assert(bin_b);
        if(bin_b->getOpcode() == BinaryInstruction::MUL)
        {
            return bin_b;
        }
    }
    return nullptr;
}

Type* gettype(Instruction* ins, int index)
{
    // debugmul("ins name = "<<ins->getInstName());
    // debugmul("ins ope size = "<<ins->getOp().size());
    // debugmul("index = "<<index);
    assert(!(index < 0 || index >= ins->getOp().size()));
    Operand* ope = ins->getOp()[index];
    return ope->getType();   
}

Operand* get_B(BasicBlock* bb)
{   
    // wzr: 这种写法太浪费遍历次数了，遍历获取bodybb中gep完全相同的过程重复了3次
    std::vector<Instruction*> geps = find_insts(Instruction::GEP, bb);
    if(geps.size() != 8)return 0;
    LoadInstruction* load_before_gep4 = dynamic_cast<LoadInstruction*>(geps[4]->getPrev());
    Operand* addr_B = load_before_gep4->getOp()[1];
    debugmul("addr_B"<<addr_B->toStr());
    return addr_B;
}

int findN_frombody(BasicBlock* bb)
{
    GepInstruction* gep = dynamic_cast<GepInstruction*>(find_inst(Instruction::GEP, bb));
    assert(gep);
    Type* type_dst_gep = gep->getOp()[0]->getType();
    debugmul("type_dst_gep = "<<type_dst_gep->toStr());
    PointerType* poinertype = dynamic_cast<PointerType*>(type_dst_gep);
    assert(poinertype);
    Type* pointerrype2 = poinertype->getType();
    ArrayType* arraytype = dynamic_cast<ArrayType*>(pointerrype2);
    assert(arraytype);
    debugmul("arraytype = "<<arraytype->toStr());
    int N = arraytype->getLength();
    debugmul("N = "<<N);
    return N;
}

Operand* get_n(BasicBlock* bb)
{
    CmpInstruction* cmp = dynamic_cast<CmpInstruction*>(find_inst(Instruction::CMP, bb));
    assert(cmp);
    Operand* n_content = cmp->getOp()[2];
    LoadInstruction* n_def = dynamic_cast<LoadInstruction*>(n_content->getDef());
    assert(n_def);
    Operand* n_ptr = n_def->getOp()[1];
    return n_ptr;
}

Operand* get_iter(BasicBlock* bb)
{   
    CmpInstruction* cmp = dynamic_cast<CmpInstruction*>(find_inst(Instruction::CMP, bb));
    if(!cmp)return nullptr;
    Operand* it2 = cmp->getOp()[1];
    BinaryInstruction* bin = dynamic_cast<BinaryInstruction*>(it2->getDef());
    if(!bin)return nullptr;
    Operand* it1 = bin->getOp()[1];
    return it1;
}

Operand* get_iter2(BasicBlock* bb)
{   
    CmpInstruction* cmp = dynamic_cast<CmpInstruction*>(find_inst(Instruction::CMP, bb));
    if(!cmp)return nullptr;
    Operand* it2 = cmp->getOp()[1];
    BinaryInstruction* bin = dynamic_cast<BinaryInstruction*>(it2->getDef());
    if(!bin)return nullptr;
    Operand* it1 = bin->getOp()[0];
    return it1;
}

void setPreSucRelation(std::vector<BasicBlock*> t_newbbs)
{
    for(int i = 0; i < t_newbbs.size(); i ++)
    {
        // printf("newbb[%d].no = %d\n", i, t_newbbs[i]->getNo());
        // Instruction* last = get_lastInstruction(t_newbbs[i]);
        Instruction* condbr = find_inst(Instruction::COND, t_newbbs[i]);
        Instruction* uncondbr = find_inst(Instruction::UNCOND, t_newbbs[i]);
        assert(condbr != nullptr || uncondbr != nullptr);
        // assert(last->isCond() || last->isUncond());
        if(condbr != nullptr && uncondbr == nullptr)
        {
            CondBrInstruction* last_cond = dynamic_cast<CondBrInstruction*>(condbr);
            t_newbbs[i]->addSucc(last_cond->getTrueBranch());
            t_newbbs[i]->addSucc(last_cond->getFalseBranch());
            last_cond->getTrueBranch()->addPred(t_newbbs[i]);
            last_cond->getFalseBranch()->addPred(t_newbbs[i]);
        }
        else if(uncondbr != nullptr && condbr == nullptr)
        {
            UncondBrInstruction* last_uncond = dynamic_cast<UncondBrInstruction*>(uncondbr);
            t_newbbs[i]->addSucc(last_uncond->getBranch());
            last_uncond->getBranch()->addPred(t_newbbs[i]);
        }
        else
        {
            assert(0 && "shoudn't reach here!");
        }
    }
}

void install_pthread_sysfunc()
{
    std::vector<Type*> vec;
    std::vector<SymbolEntry*> vec1;
    Type* funcType = new FunctionType(TypeSystem::voidType, vec, vec1);
    SymbolTable* st = identifiers;
    while(st->getPrev())
        st = st->getPrev();
    SymbolEntry* se = new IdentifierSymbolEntry(funcType, "pthread_create", st->getLevel(), -1, true);
    ((IdentifierSymbolEntry*)se)->setIntParamNo(0);
    ((IdentifierSymbolEntry*)se)->setFloatParamNo(0);
    st->install("pthread_create", se);

    // SymbolEntry* se_pthread_create = identifiers->lookup("pthread_create");
    // printf("se_pthread_create = %s\n", se_pthread_create);
    // getchar();

    std::vector<Type*> vec_join;
    std::vector<SymbolEntry*> vec_join1;
    Type* funcType_join = new FunctionType(TypeSystem::voidType, vec_join, vec_join1);
    SymbolTable* st_join = identifiers;
    while(st_join->getPrev())
        st_join = st_join->getPrev();
    SymbolEntry* se_join = new IdentifierSymbolEntry(funcType_join, "pthread_join", st_join->getLevel(), -1, true);
    ((IdentifierSymbolEntry*)se_join)->setIntParamNo(0);
    ((IdentifierSymbolEntry*)se_join)->setFloatParamNo(0);
    st_join->install("pthread_join", se_join);
}

// tell bb's predblock, modify their succ from bb to nextbb
void tellPredModifySucc(BasicBlock* bb, BasicBlock* nextbb)
{
    for(BasicBlock* pred : bb->getPredBlock())
    {
        debugmul("cond1 pred = "<<pred->getNo());
        debugmul("nextbb = "<<nextbb->getNo());
        pred->removeSucc(bb);
        bb->removePred(pred);
        pred->addSucc(nextbb);
        nextbb->addPred(pred);
        setBranch_trueBranch(nextbb, pred);
    }
}

void Multithread::pass2()
{
    debugmul("multithread_who = "<<multithread_who);
    auto label_it = blk_needmodify_label.begin();
    for(auto bb = blk_needmodify.begin(); bb != blk_needmodify.end(); bb ++)
    {
        for(auto mfunc = mUnit->begin(); mfunc != mUnit->end(); mfunc++)
        {
            for(auto mbb = (*mfunc)->begin(); mbb != (*mfunc)->end(); mbb ++)
            {
                debugmul("(*mbb)->getNo() = "<<(*mbb)->getNo());
                debugmul("(*bb)->getNo() = "<<(*bb)->getNo());
                if((*mbb)->getNo() == (*bb)->getNo())
                {
                    debugmul("if((*mbb)->getNo() == (*bb)->getNo()) in");
                    std::vector<MachineInstruction *> inst_list = (*mbb)->getInsts();
                    if(*label_it == 1)
                    {
                        for(int i = 0; i < inst_list.size(); i ++)
                        {
                            MovMInstruction* find_mov = dynamic_cast<MovMInstruction*>(inst_list[i]);
                            if(find_mov)
                            {
                                debugmul("find_mov");
                                debugmul("find_mov->getUse()[0]->getReg() = "<<find_mov->getUse()[0]->getReg());
                                if(i>0)debugmul("inst_list[i - 1]->isMov() = "<<inst_list[i - 1]->isMov());
                                #ifdef DEBUG_SWITCH_VEC
                                    // if(i>0 && inst_list[i - 1]->getUse()[0])inst_list[i - 1]->getUse()[0]->output_to_console();
                                    // if(inst_list[i]->getDef()[0])inst_list[i]->getDef()[0]->output_to_console();
                                #endif
                                if(find_mov->getUse()[0]->getReg() == 10
                                    && i > 0 && inst_list[i - 1]->isMov()
                                    /*&& inst_list[i - 1]->getUse()[0] == inst_list[i]->getDef()[0]*/)
                                {
                                    debugmul("begin switch");
                                    // dynamic_cast<MovMInstruction*>(inst_list[i - 1])->output_to_console();
                                    // dynamic_cast<MovMInstruction*>(inst_list[i])->output_to_console();
                                    auto temp = inst_list[i - 1];
                                    (*mbb)->getInsts()[i - 1] = inst_list[i];
                                    (*mbb)->getInsts()[i] = temp;
                                    debugmul("end switch");
                                    // dynamic_cast<MovMInstruction*>(inst_list[i - 1])->output_to_console();
                                    // dynamic_cast<MovMInstruction*>(inst_list[i])->output_to_console();
                                }
                            }
                        }
                    }
                    else if(*label_it == 2)
                    {
                        debugmul("multithread2 asm r10 begin");
                        MovMInstruction* last_mov = dynamic_cast<MovMInstruction*>(inst_list[inst_list.size() - 2]);
                        assert(last_mov);
                        MachineOperand* dst = last_mov->getUse()[0];
                        debugmul("dst = "<<dst->toStr());
                        MovMInstruction* movr10 = new MovMInstruction(*mbb, MovMInstruction::MOV, dst, new MachineOperand(MachineOperand::REG, 10));
                        // int size = (*mbb)->getInsts().size();
                        (*mbb)->getInsts().insert((*mbb)->getInsts().end() - 2, movr10);
                        debugmul("multithread2 asm r10 end");
                    }
                    else
                    {
                        printf("no multithread did, why here?\n");
                    }
                }
            }
        }
        label_it++;
    }
}