      
#include "../include/SSADestruction.h"
#include <vector>
using namespace std;

#define DEBUG_SWITCH_SSAD 0
#if DEBUG_SWITCH_SSAD
#define printssad(str) \
    std::cout << str << "\n";
#else
#define printssad(str) //
#endif

bool while_with_if(Function* func);

void SSADestruction::pass() {
    printssad("==========start SSADestruction=========");
    std::vector<Function*> func = unit->get_functions();
    for(auto f = func.begin(); f != func.end(); f++){
        printssad("=========start func SSADestruction=========");
        printssad((*f)->getSymPtr()->toStr());
        // if((*f)->getSymPtr()->toStr() == "@pseudo_sha1"){
        if((*f)->getwhile_with_if_and_specialassign()){
            pass(*f);
        }else{
            passSpecial(*f);
        }
        printssad("=========start func SSADestruction=========");
    }
    printssad("==========end SSADestruction=========");
}

// 需要在每一条phi指令的来源块中多插入一个块来进行变量之间的保存和替代
void SSADestruction::pass(Function* function) {
    vector<BasicBlock*> blocks(function->begin(), function->end());
    for (auto it = blocks.begin(); it != blocks.end(); it++) {
        map<BasicBlock*, vector<Instruction*>> copyInss;
        map<BasicBlock*, vector<Instruction*>> newcopyInss;
        vector<Instruction*> phiremove;
        if (!(*it)->begin()->isPhi())
            continue;
        for(auto inst = (*it)->begin(); inst != (*it)->end(); inst = inst->getNext()){
            if(!inst->isPhi()) break;
            auto phi = dynamic_cast<PhiInstruction*>(inst);
            std::map<BasicBlock*, Operand*> srcs = phi->getSrcs();
            printssad("Get phi instruction in block: "<<(*it)->getNo());
            for(auto src: srcs){
                BasicBlock* b = src.first; 
                Operand* op = src.second;
                auto addin = new BinaryInstruction(BinaryInstruction::ADD, phi->getDef(), op,
                                new Operand(
                                    new ConstantSymbolEntry(phi->getDef()->getType(), 0)));
                copyInss[b].push_back(addin); 
                op->addUse(addin);
                phi->getDef()->addUse(addin);
                printssad("get src block no: "<<b->getNo());
            }
            phiremove.push_back(phi);
        }
        if(copyInss.empty()){
            continue;
        }
        printssad("============start insert block===============");
        // 在每一个phi指令的跳转块中插入对应的块用来存放phi指令的替代指令
        for(auto bbcopyinst: copyInss){
            BasicBlock* block = bbcopyinst.first;
            vector<Instruction*> ins = copyInss[block];
            auto succlist = block->getSuccBlock();
            auto prevlist = (*it)->getPredBlock();
            if(find(succlist.begin(), succlist.end(), *it) == succlist.end()){
                printssad("block src: "<<block->getNo()<<" continue");
                continue;
            }
            if(find(prevlist.begin(), prevlist.end(), block) == prevlist.end()){
                printssad("block src: "<<block->getNo()<<" continue");
                continue;
            }
            auto newBlock = new BasicBlock(function);
            printssad("block src: "<<block->getNo()<<" add succ: "<<newBlock->getNo());
            block->removeSucc(*it);
            block->addSucc(newBlock);
            newBlock->addPred(block);
            newBlock->addSucc(*it);
            (*it)->removePred(block);
            (*it)->addPred(newBlock);
            // 修改之前的跳转指令
            auto endIns = block->rbegin();
            printssad("change block and newblock success!");
            if(endIns->isCond()){
                auto in = dynamic_cast<CondBrInstruction*>(endIns);
                if(in->getTrueBranch()->getNo() == (*it)->getNo())
                    in->setTrueBranch(newBlock);
                else
                    in->setFalseBranch(newBlock);
                printssad("branch cond replace: "<< (*it)->getNo() << " => " << newBlock->getNo());
            }
            else if(endIns->isUncond()){
                auto in = dynamic_cast<UncondBrInstruction*>(endIns);
                if(in->getBranch()->getNo() == (*it)->getNo()){
                    in->setBranch(newBlock);
                }
                printssad("branch uncond replace: "<< (*it)->getNo() << " => " << newBlock->getNo());
            }
            auto uncond = new UncondBrInstruction((*it), newBlock);
            // 修改插入的add指令
            for (auto in : ins)
                newBlock->insertBefore(in, uncond);
            printssad("insert add instructions success!");
        }
        printssad("============end insert block===============");
        for(auto phi: phiremove){
            (*it)->remove(phi);
        }
    }
}

void SSADestruction::passSpecial(Function* function) {
    //  std::cout<<"开始phi lower\n";
    map<BasicBlock*, vector<Instruction*>> copyInss;
    vector<BasicBlock*> blocks(function->begin(), function->end());
    // for (auto it = function->begin(); it != function->end(); it++) {
    for (auto it = blocks.begin(); it != blocks.end(); it++) {
        // std::cout<<"block:"<<(*it)->getNo()<<std::endl;
        if (!(*it)->begin()->isPhi())
            continue;
        copyInss.clear();
        vector<Instruction*> phiremove;
        // std::cout<<"phiremove.size() = "<<phiremove.size()<<"\n";
        for(auto inst = (*it)->begin(); inst != (*it)->end(); inst = inst->getNext()){
            if(!inst->isPhi()) break;
            // std::cout<<"inst->getdef = "<<inst->getDef()->toStr()<<"\n";
            if(inst->getDef()->isConst())
            {
                // std::cout<<"in phiremove.push_back(inst);\n";
                phiremove.push_back(inst);
                continue;
            }
            // std::cout<<"I'm phi\n";
            auto phi = dynamic_cast<PhiInstruction*>(inst);
            std::map<BasicBlock*, Operand*> srcs = phi->getSrcs();
            auto temp = new TemporarySymbolEntry(TypeSystem::intType, SymbolTable::getLabel());
            // auto tempop = new Operand(temp);
            // // fix: 判断phi指令的src块是否和phi指令为同一个块
            // // 这会导致所有的phi指令都会建一个新的块
            for(auto src: srcs){
                BasicBlock* b = src.first;
                Operand* op = src.second;
                // new MovInstruction();
                // copyInss[b].push_back(new StoreInstruction(phi->getDef(), op));
                // fix: 建立use-def链
                auto addin = new BinaryInstruction(BinaryInstruction::ADD, phi->getDef(), op,
                                new Operand(
                                    new ConstantSymbolEntry(phi->getDef()->getType(), 0)));
                copyInss[b].push_back(addin); 
                op->addUse(addin);
                phi->getDef()->addUse(addin);
                // 需要注意的是如果phi指令remove, 对应的def变为第一条add指令
            }   
            // new StoreInstruction(phi->getDef(), tempop, (*it));
            phiremove.push_back(phi);
            // (*it)->remove(phi);
        }
        // fix: 如果存在和phi指令所在块同样的替换块的话
        // fix: 需要找到一个块内的所有指令而非整个函数的
        if(copyInss.count((*it)) != 0 ){
            // std::cout<<"加入关键边"<<std::endl;
            vector<Instruction*> ins = copyInss[(*it)];
            // 插入一个新块
            auto newBlock = new BasicBlock(function);
            if((*it)->existsInSucc((*it)))
            (*it)->removeSucc((*it));
            if((*it)->existsInPred((*it)))
            (*it)->removePred((*it));

            if(!(*it)->existsInPred((newBlock)))
            (*it)->addPred(newBlock);
            if(!(*it)->existsInSucc((newBlock)))
            (*it)->addSucc(newBlock);

            if(!newBlock->existsInSucc((*it)))
            newBlock->addSucc(*it);
            if(!newBlock->existsInPred((*it)))
            newBlock->addPred(*it);

            // 修改之前的跳转指令
            auto endIns = (*it)->rbegin();
            if(endIns->isCond()){
                auto in = dynamic_cast<CondBrInstruction*>(endIns);
                in->setTrueBranch(newBlock);
            }
            new UncondBrInstruction((*it), newBlock);
            // 修改插入的add指令
            for(auto i: ins){
                copyInss[newBlock].push_back(i);
            }
        }
        copyInss.erase((*it));

        // std::cout<<"phiremove.size() = "<<phiremove.size()<<"\n";
        for(auto phi: phiremove){
            // std::cout<<"removing phi = "<<phi->getDef()->toStr()<<"\n";
            // phi->getDef()->setDef(phi->getDef()->getUse()[0]);
            (*it)->remove(phi);
        }

        // std::cout<<"插入phi替换指令"<<std::endl;

        // 插入phi指令替换指令
        // for (auto it2: copyInss) {
        //     auto block = it2.first;
        //     auto& ins = it2.second;
        //     // 插入phi指令在条件跳转之前
        //     Instruction* endIns = block->rbegin();
        //     // std::cout<<"插入的指令数目"<<ins.size()<<std::endl;
        //     if(endIns->isUncond()){
        //         auto inst = block->begin();
        //         for( ; inst != block->end(); inst = inst->getNext()){
        //             // std::cout<<inst->getInstName();
        //         }
        //         for (auto in : ins){
        //             // std::cout<<"hh";
        //             block->insertBefore(in, block->begin());
        //             // std::cout<<in->getInstName();
        //         }
        //     }
        //     else if(endIns->isCond()){
        //         auto seIns = endIns->getPrev();
        //         for (auto in : ins)
        //             block->insertBefore(in, seIns);
        //     }
        //     else{
        //         for (auto in : ins)
        //             block->insertBefore(in, endIns);
        //     }

        // }
        for (auto it2: copyInss) {
            auto block = it2.first;
            auto& ins = it2.second;
            // 插入phi指令在条件跳转之前
            auto endIns = block->rbegin();
            //std::cout<<"插入的指令数目"<<ins.size()<<std::endl;
            if(endIns->isUncond()){
                for (auto in : ins){
                    // std::cout<<"插入 "<<in->getInstName() 
                    //   << " 到 " << endIns->getInstName() << " 之前."<<std::endl;    
                    block->insertBefore(in, endIns);
                }
            }
            else if(endIns->isCond()){
                auto seIns = endIns->getPrev();
                for (auto in : ins)
                    block->insertBefore(in, seIns);
            }
            else{
                for (auto in : ins)
                    block->insertBefore(in, endIns);
            }
        }
    }
}

bool while_with_if(Function* func)
{
    return false;
}

    