%option noyywrap
%option yylineno
%{
    #define YY_NO_UNPUT
    #define YY_NO_INPUT
    #include "parser.h"
    #include <ostream>
    #include <fstream>
    #include <iostream>
    using namespace std;
    extern FILE *yyin; 
    extern FILE *yyout;
    int offsets = 0;
    int newoffsets = 0;

    void update_loc(){
        offsets = newoffsets;
        newoffsets = offsets + strlen(yytext);
    }
    #define YY_USER_ACTION update_loc();  
    
%}
DECIMIAL ([1-9][0-9]*|0)
OCTAL (0[0-7]+)
HEXADECIMAL (0[xX][0-9A-Fa-f]+)
ID [[:alpha:]_][[:alpha:][:digit:]_]*
EOL (\r\n|\n|\r)
WHITE [\t ]

BLOCKCOMMENTBEGIN \/\*
BLOCKCOMMENTELEMENT .
BLOCKCOMMENTEND \*\/
%x BLOCKCOMMENT
LINECOMMENT \/\/[^\n]*

DECIMAL_FLOAT ((([0-9]*[.][0-9]*([eE][+-]?[0-9]+)?)|([0-9]+[eE][+-]?[0-9]+))[fLlL]?)
HEXADECIMAL_FLOAT (0[xX](([0-9A-Fa-f]*[.][0-9A-Fa-f]*([pP][+-]?[0-9]+)?)|([0-9A-Fa-f]+[pP][+-]?[0-9]+))[fLlL]?)

%%

"+" {
    return ADD;
}
"-" {
    return SUB;
}
"*" {
    return MUL;
}
"/" {
    return DIV;
}

"==" {
    return EQUAL;
}
"!=" {
    return NOTEQUAL;
}
"=" {
    return ASSIGN;
}
"<" {
    return LESS;
}
"<=" {
    return LESSEQUAL;
}
">" {
    return GREATER;
}
">=" {
    return GREATEREQUAL;
}


"%" {
    return MOD;
}
"&&" {
    return AND;
}
"||" {
    return OR;
}

"int" {
    return INT;
}
"float" {
    return FLOAT;
}
"void" {
    return VOID;
}
"const" {
    return CONST;
}
"if" {
    return IF;
};
"else" {
    return ELSE;
};
"return" {
    return RETURN;
}
"while" {
    return WHILE;
}
"break" {
    return BREAK;
}
"continue" {
    return CONTINUE;
}
"!" {
    return NOT;
}
";" {
    return SEMICOLON;
}
"(" {
    return LPAREN;
}
")" {
    return RPAREN;
}
"{" {
    return LBRACE;
}
"}" {
    return RBRACE;
}
"[" {
    return LBRACKET;
}
"]" {
    return RBRACKET;
}
"," {
    return COMMA;
}

{DECIMAL_FLOAT} {
    yylval.FloatType = atof(yytext);
    return DECIMAL_FLOAT;
}
{HEXADECIMAL_FLOAT} {
    yylval.FloatType = atof(yytext);
    return HEXADECIMAL_FLOAT;
}
{DECIMIAL} {
    yylval.IntType = atoi(yytext);
    return DECIMIAL;
}
{OCTAL} {
    int temp;
    sscanf(yytext, "%o", &temp);
    yylval.IntType = temp;
    return OCTAL;
}
{HEXADECIMAL} {
    int temp;
    sscanf(yytext, "%x", &temp);
    yylval.IntType = temp;
    return HEXADECIMAL;
}

"getint" {
    char *token = new char[strlen(yytext) + 1];
    strcpy(token, yytext);
    yylval.StrType = token;
    return SYSY_FUNC_NOPARAM;
}
"getch" {
    char *token = new char[strlen(yytext) + 1];
    strcpy(token, yytext);
    yylval.StrType = token;
    return SYSY_FUNC_NOPARAM;
}
"getfloat" {
    char *token = new char[strlen(yytext) + 1];
    strcpy(token, yytext);
    yylval.StrType = token;
    return SYSY_FUNC_NOPARAM;
}
"getarray" {
    char *token = new char[strlen(yytext) + 1];
    strcpy(token, yytext);
    yylval.StrType = token;
    return SYSY_FUNC_WITHPARAM;
}
"getfarray" {
    char *token = new char[strlen(yytext) + 1];
    strcpy(token, yytext);
    yylval.StrType = token;
    return SYSY_FUNC_WITHPARAM;
}

"putint" {
    char *token = new char[strlen(yytext) + 1];
    strcpy(token, yytext);
    yylval.StrType = token;
    return SYSY_FUNC_WITHPARAM;
}
"putch" {
    char *token = new char[strlen(yytext) + 1];
    strcpy(token, yytext);
    yylval.StrType = token;
    return SYSY_FUNC_WITHPARAM;
}
"putfloat" {
    char *token = new char[strlen(yytext) + 1];
    strcpy(token, yytext);
    yylval.StrType = token;
    return SYSY_FUNC_WITHPARAM;
}
"putarray" {
    char *token = new char[strlen(yytext) + 1];
    strcpy(token, yytext);
    yylval.StrType = token;
    return SYSY_FUNC_WITHPARAM;
}
"putfarray" {
    char *token = new char[strlen(yytext) + 1];
    strcpy(token, yytext);
    yylval.StrType = token;
    return SYSY_FUNC_WITHPARAM;
}
"putf" {
    char *token = new char[strlen(yytext) + 1];
    strcpy(token, yytext);
    yylval.StrType = token;
    return SYSY_FUNC_WITHPARAM;
}

"starttime" {
    char *token = new char[6 + strlen(yytext) + 1];
    char* sysy = (char *)"_sysy_";
    strcpy(token, sysy);
    strcpy(token + 6, yytext);
    yylval.StrType = token;
    return SYSY_FUNC_NOPARAM;
}
"stoptime" {
    char *token = new char[6 + strlen(yytext) + 1];
    char* sysy = (char*)"_sysy_";
    strcpy(token, sysy);
    strcpy(token + 6, yytext);
    yylval.StrType = token;
    return SYSY_FUNC_NOPARAM;
}

{ID} {
    char *token;
    token = new char[strlen(yytext) + 1];
    strcpy(token, yytext);
    yylval.StrType = token;

    return ID;
}

<*>{EOL} {
    newoffsets = 0;
}

{WHITE}
{LINECOMMENT} {}
{BLOCKCOMMENTBEGIN} {BEGIN BLOCKCOMMENT;}
<BLOCKCOMMENT>{BLOCKCOMMENTELEMENT} {}
<BLOCKCOMMENT>{BLOCKCOMMENTEND} {BEGIN INITIAL;}


%%
