#include "AccumExpReduction.h"
#include "Ast.h"
#include "BlockMerge.h"
#include "ComSubExprEli.h"
#include "ConstantUnfold.h"
#include "DeadCodeEliminate.h"
#include "FunctionInline.h"
#include "GlobalVarLocalization.h"
#include "GraphColor.h"
#include "Hoisting.h"
#include "InstructionsSchedule.h"
#include "LoopDetection.h"
#include "LoopStrengthReduction.h"
#include "MachineCode.h"
#include "Multithread.h"
#include "PeepholeMInstrOpt.h"
#include "SSADestruction.h"
#include "SparseConditionalConstantPropagation.h"
#include "Unit.h"
#include "Vectorization.h"
#include "dominatorTree.h"
#include "mem2reg.h"
#include "plugin.h"
#include <iostream>
#include <regex>
#include <string.h>
#include <unistd.h>

using namespace std;

Ast ast;
Unit unit;
MachineUnit mUnit;
extern FILE *yyin;
extern FILE *yyout;
extern bool *fcheck;

int yyparse();

char outfile[256] = "a.out";
bool dump_tokens;
bool dump_ast;
bool dump_ir;
bool dump_asm;
bool optimize;
bool test_ = 1;

#define DEBUG_SWITCH_MAIN 0
#if DEBUG_SWITCH_MAIN
#define debug(str) std::cout << str << "\n";
#else
#define debug(str) //
#endif

int main(int argc, char *argv[]) {
    int opt;
    while ((opt = getopt(argc, argv, "Sieato:O::")) != -1) {
        switch (opt) {
        case 'o':
            strcpy(outfile, optarg);
            break;
        case 'a':
            dump_ast = true;
            break;
        case 't':
            dump_tokens = true;
            break;
        case 'i':
            dump_ir = true;
            break;
        case 'O':
            optimize = true;
            break;
        case 'S':
            dump_asm = true;
            break;
        case 'e':
            test_ = 0;
            break;
        default:
            // fprintf(stderr, "Usage: %s [-o outfile] infile\n", argv[0]);
            // exit(EXIT_FAILURE);
            dump_asm = true;
            break;
        }
    }
    if (optind >= argc) {
        fprintf(stderr, "no input file\n");
        exit(EXIT_FAILURE);
    }

    // std::cout<<argv[optind]<<"\n";
    if (!(yyin = fopen(argv[optind], "r"))) {
        fprintf(stderr, "%s: No such file or directory\nno input file\n",
                argv[optind]);
        exit(EXIT_FAILURE);
    }
    if (!(yyout = fopen(outfile, "w"))) {
        fprintf(stderr, "%s: fail to open output file\n", outfile);
        exit(EXIT_FAILURE);
    }
    yyparse();
    if (dump_ast)
        ast.output();
    ast.genCode(&unit);
    debug("genCode");
    string str = argv[optind];
    size_t found = str.find("h-6-0");

    auto onePass = [](int i, bool cf) {
        SCCP sccp(&unit);
        if (!cf)
            sccp.execute();
        debug("[onePass]:sccp " << i);

        DeadCodeEliminate dce(&unit);
        if (!cf)
            dce.execute();
        debug("[onePass]:dce " << i);

        AccumExpReduction aer(&unit);
        aer.execute();
        debug("[onePass]:aer " << i);

        BlockMerge bm(&unit);
        bm.execute();
        debug("[onePass]:bm " << i);

        Hoisting hs(&unit);
        hs.execute(false);
        debug("[onePass]:hosit " << i);

        LoopStrengthReduction lsr(&unit);
        lsr.execute(false);
        debug("[onePass]:lsr " << i);

        ElimExtrPhi eep(&unit);
        eep.execute();
        debug("[onePass]:eep " << i);
    };
    Multithread multithread(&unit, &mUnit);
    // optimize = 1;
    DeadCodeEliminate dce0(&unit);
    dce0.executeFront();
    debug("dce0");

    ComSubExprEli em(&unit);
    em.execute();
    debug("em");

    if (optimize) {

        ConSameStructElim csse(&unit);
        csse.execute();
        debug("csse");

        // Mem2Reg mr0(&unit);
        // mr0.execute2();
        // debug("mem0");

        FunctionInline fi1(&unit);
        fi1.pass();
        debug("finline");

        GlobalVariableLocalization g2l(&unit);
        g2l.pass();
        debug("g2l");

        // debug("mem 1");
        Mem2Reg mr(&unit);
        mr.execute();
        // test(&unit);
        debug("mem");

        SCCP sccp(&unit);
        sccp.execute();
        debug("sccp");

        DeadCodeEliminate dce(&unit);
        dce.execute();
        debug("dce");

        AccumExpReduction aer(&unit);
        aer.execute();
        debug("aer");

        BlockMerge bm(&unit);
        bm.execute();
        debug("bm");

        Hoisting hs(&unit);
        hs.execute(true);
        debug("hosit");

        LoopStrengthReduction lsr(&unit);
        lsr.execute(true);
        debug("lsr");

        // onePass(1,lsr.getCf());

        ConstantUnfold constantUnfold(&unit);
        constantUnfold.pass();
        debug("unfold");

        Vectorization vectorization(&unit);
        vectorization.pass();
        debug("vec");

        multithread.pass();
        debug("multithread");

        ElimExtrPhi eep(&unit);
        eep.execute();
        debug("eep");

        SSADestruction ssa(&unit);
        ssa.pass();
        debug("ssa");

        ElimExtraJump eej(&unit);
        eej.execute();
        debug("eej");

        BlockMerge bm2(&unit);
        bm2.execute();
        debug("bm");
    }

    if (dump_ir)
        unit.output();
    debug("before genMachineCode");
    unit.genMachineCode(&mUnit);
    debug("genMachineCode");

    GraphColor GraphColor(&mUnit);
    GraphColor.allocateRegisters();

    if (optimize) {
        PeepholeMInstrOpt pmo(&mUnit);
        pmo.pass();

        InstructionsScheduling ISche(&mUnit);
        ISche.pass();
    }

    multithread.pass2();

    if (dump_asm)
        mUnit.output();
    debug("main end");
    return 0;
}
