#include "mem2reg.h"

#define DEBUG_SWITCH_M2R 0
#if DEBUG_SWITCH_M2R
#define printmr(str) std::cout << str << "\n";
#else
#define printmr(str) //
#endif

void Mem2Reg::execute() {
    printmr("==========start Mem2Reg=========");
    this->dom = new dominatorTree(unit);
    this->dom->execute();
    std::vector<Function *> func = unit->get_functions();
    for (auto f = func.begin(); f != func.end(); f++) {
        printmr("=========start func mem2reg=========");
        printmr((*f)->getSymPtr()->toStr());
        execute(*f);
        printmr("=========end func mem2reg=========");
    }
    printmr("==========end Mem2Reg=========");
}

bool Mem2Reg::checkFunc(SymbolEntry *se) {
    FunctionType* funcType = (FunctionType *)(se->getType());
    if (((funcType)->getParamsType())
            .size() > 5)
        return false;
    std::vector<Type *> paramlist =
        funcType->getParamsType();
    int arrParamCnt = 0;
    int intParamCnt = 0;
    int floatParamCnt = 0;
    for (auto param : paramlist) {
        if (param->isArray()) {
            arrParamCnt++;
        } else if (param->isFloat()) {
            floatParamCnt++;
        } else {
            intParamCnt++;
        }
    }
    // std::cout<<paramlist.size()<<" "<<intParamCnt<<" "<<arrParamCnt<<"\n";
    if (paramlist.size() == 4 && intParamCnt == 1 && arrParamCnt == 3){
        flag = 0;
        return false;
    }
    return true;
}

void Mem2Reg::execute2() {
    printmr("==========start Mem2Reg=========");
    this->dom = new dominatorTree(unit);
    this->dom->execute();
    std::vector<Function *> func = unit->get_functions();
    for (auto f = func.begin(); f != func.end(); f++) {
        printmr("=========start func mem2reg=========");
        printmr((*f)->getSymPtr()->toStr());
        if (!checkFunc((*f)->getSymPtr()))
            continue;

        execute(*f, 1);
        printmr("=========end func mem2reg=========");
    }
    printmr("==========end Mem2Reg=========");
}

// 0329m fix: 需要再次构建前驱后继关系
// 0408m fix: 需要再次构建def-use链
void Mem2Reg::execute(Function *f, bool p) {
    std::unordered_map<AllocaInstruction *, int>
        alloca_ids; // 把alloca映射到整数，后面有好几个vector用这个做下标
    std::vector<AllocaInstruction *> allocas;
    int paramCnt = 0;
    printmr("=========找到所有的alloca指令=========");
    // 找到所有的alloca指令
    for (auto it = f->begin(); it != f->end(); it++) {
        for (auto i = (*it)->begin(); i != (*it)->end(); i = i->getNext()) {
            if (i->isAlloc()) {
                // 这里有一些判断条件
                // 非函数参数:
                // 非数组
                // if((dynamic_cast<TemporarySymbolEntry*>(dynamic_cast<AllocaInstruction*>(i)->getSymbolEntry()))->isG2L())
                // {
                //     alloca_ids.insert({dynamic_cast<AllocaInstruction*>(i),
                //     (int)alloca_ids.size()});
                //     allocas.push_back(dynamic_cast<AllocaInstruction*>(i));
                //     printmr("=========isG2L=========");
                //     continue;
                // }
                if ((dynamic_cast<AllocaInstruction *>(i)->getSymbolEntry())
                        ->isTemporary()) {
                    auto se = dynamic_cast<TemporarySymbolEntry *>(
                        dynamic_cast<AllocaInstruction *>(i)->getSymbolEntry());
                    if (se->isG2L()) {
                        alloca_ids.insert({dynamic_cast<AllocaInstruction *>(i),
                                           (int)alloca_ids.size()});
                        allocas.push_back(
                            dynamic_cast<AllocaInstruction *>(i)); // se
                    }
                } else {
                    auto se = dynamic_cast<IdentifierSymbolEntry *>(
                        dynamic_cast<AllocaInstruction *>(i)->getSymbolEntry());

                    if (se->isLocal())
                        printmr("=========isLocal=========");
                    if (se->isParam())
                        printmr("=========isParam=========");
                    if (se->isParam2Local())
                        printmr("=========isParam2Local=========");
                    if (se->getType()->isInt())
                        printmr("=========isInt=========");
                    // if(se->getType()->isFloat())
                    // printmr("=========isFloat=========");
                    // printmr("=========isFloat2=========");  || (se->isParam()
                    // && !se->getType()->isArray())

                    if (p) {
                        if ((se->isParam() && !se->getType()->isArray()) &&
                            paramCnt < 4) {
                            alloca_ids.insert(
                                {dynamic_cast<AllocaInstruction *>(i),
                                 (int)alloca_ids.size()});
                            allocas.push_back(
                                dynamic_cast<AllocaInstruction *>(i)); // se
                            paramCnt++;
                        }
                    } else if ((se->isLocal() || se->isParam2Local() ||
                                (!se->isParam() &&
                                 !se->getType()->isArray())) &&
                               (se->getType()->isInt() ||
                                se->getType()->isFloat())) { //
                        alloca_ids.insert({dynamic_cast<AllocaInstruction *>(i),
                                           (int)alloca_ids.size()});
                        allocas.push_back(
                            dynamic_cast<AllocaInstruction *>(i)); // se
                        // std::cout<<dynamic_cast<AllocaInstruction*>(i)->getSymbolEntry()->toStr()<<"\n";
                    }
                }
            }
        }
    }
    printmr("allocas的大小: " << allocas.size());

    printmr("=========找到所有的store指令=========");
    // 找到所有的store指令
    std::vector<std::vector<BasicBlock *>> alloca_defs(alloca_ids.size());
    for (auto it = f->begin(); it != f->end(); it++) {
        BasicBlock *bb = (*it);
        for (auto i = bb->begin(); i != bb->end(); i = i->getNext()) {
            if (i->isStore()) {
                // 这里需要在alloca_ids中找到store指令对应的alloca指令
                // 找一下定义dynamic_cast<StoreInstruction*>(i)->getSrc()的指令即可
                auto it = alloca_ids.find(dynamic_cast<AllocaInstruction *>(
                    dynamic_cast<StoreInstruction *>(i)
                        ->getAddr()
                        ->getDef())); // addr
                if (it != alloca_ids.end()) {
                    alloca_defs[it->second].push_back(
                        bb); // 将每一个alloca对应的块插入
                } // alloca_defs存储的是每一个alloca变量定义的块
            }
        }
    }
    printmr("=========mem2reg算法阶段1：放置phi节点=========");

    // mem2reg算法阶段1：放置phi节点
    // worklist定义在循环外面，只是为了减少申请内存的次数
    std::vector<BasicBlock *> worklist; // 用stack还是queue在这里没有本质区别
    std::unordered_map<PhiInstruction *, int>
        phis; // 记录加入的phi属于的alloca的id

    for (int id = 0; id < allocas.size(); id++) {
        f->clear_all_vis(); // 对于一个新的alloca指令清空所有块的visit值,
                            // 标记所有块都未被访问过
        for (BasicBlock *bb : alloca_defs[id]) {
            worklist.push_back(bb);
        }
        // 将所有alloca变量定义的块加入到工作集当中
        while (!worklist.empty()) {
            BasicBlock *x = worklist.back();
            worklist.pop_back();
            // 这里的处理有点粗糙, 暂时先这样
            auto se = allocas[id]->getSymbolEntry();
            // phi 指令需要公共的变量作为最后值的存放位置
            // Operand* temp =new Operand(new
            // TemporarySymbolEntry(se->getType(),
            //                                     SymbolTable::getLabel()));
            // 遍历定义了alloca变量的块的支配边界
            for (auto y : dom->getDominanceFrontier(x)) {
                if (!y->isVisited()) {
                    y->setvisit(true);
                    // 这里插入的是对应alloca类型的变量
                    // FIX: create new operand
                    Operand *temp = new Operand(new TemporarySymbolEntry(
                        se->getType(), SymbolTable::getLabel()));
                    PhiInstruction *inst = new PhiInstruction(temp);
                    // std::cout << se->toStr() << std::endl;

                    if (se->isTemporary() &&
                        dynamic_cast<TemporarySymbolEntry *>(se)->isG2L()) {
                        se = temp->getSymbolEntry();
                        // std::cout << se->toStr() << se << std::endl;
                    }

                    temp->setOpdSe(se);                            // for hoist
                    inst->setIdentse((IdentifierSymbolEntry *)se); // for dce
                    inst->push_back_opse(se);                      // for cse
                    y->insertFront(inst);
                    phis.insert({inst, id}); // 插入alloca的id与生成的phi指令
                    worklist.push_back(y);
                }
            }
        }
    }
    printmr("=========mem2reg算法阶段2：变量重命名=========");
    // mem2reg算法阶段2：变量重命名，即删除Load，把对Load结果的引用换成对寄存器的引用，把Store改成寄存器赋值
    std::vector<IdentifierSymbolEntry *> params1;
    std::vector<IdentifierSymbolEntry *> params2;
    std::map<IdentifierSymbolEntry *, Operand *> params2op;
    std::map<Operand *, Operand *> paramMap; // 0817

    std::vector<std::pair<BasicBlock *, std::vector<Operand *>>> worklist2{
        {f->getEntry(),
         std::vector<Operand *>(alloca_ids.size(),
                                new Operand(new ConstantSymbolEntry(
                                    TypeSystem::intType,
                                    0)))}}; // 这里暂时处理为nullptr,
                                            // 之后将其进行更改为对应的未定义类
    std::map<BasicBlock *, Operand *> srcs;
    f->clear_all_vis();
    while (!worklist2.empty()) {
        BasicBlock *bb = worklist2.back().first;
        std::vector<Operand *> values = std::move(worklist2.back().second);
        // std::cout << "First: " <<  values[1]->toStr() << std::endl;
        worklist2.pop_back();
        if (!bb->isVisited()) {
            bb->setvisit(true);
            for (auto i = bb->begin(); i != bb->end();) {
                // 如果是alloca指令, 直接移除
                if (i->isAlloc()) {
                    if (find(allocas.begin(), allocas.end(),
                             dynamic_cast<AllocaInstruction *>(i)) !=
                        allocas.end()) {
                        if ((dynamic_cast<AllocaInstruction *>(i)
                                 ->getSymbolEntry())
                                ->isTemporary() &&
                            dynamic_cast<TemporarySymbolEntry *>(
                                dynamic_cast<AllocaInstruction *>(i)
                                    ->getSymbolEntry())
                                ->isG2L()) {
                            bb->remove(i);
                            i = i->getNext();
                            continue;
                        }

                        auto se = dynamic_cast<IdentifierSymbolEntry *>(
                            dynamic_cast<AllocaInstruction *>(i)
                                ->getSymbolEntry());
                        if (se->getType()->isArray()) {
                            // std::cout << "alloc: " <<
                            // dynamic_cast<AllocaInstruction*>(i)->getDef()->toStr()
                            // << std::endl; params1.push_back(se);
                            // params2.push_back(se);
                            i = i->getNext();
                            continue;
                        }

                        // 需要先将指令的前驱后继关系都取消
                        // 但是store和load指令是根据对应的def变量找到的alloca指令
                        if (se->isParam())
                            paramMap[i->getDef()] = nullptr;
                        bb->remove(i);
                        // delete static_cast<AllocaInstruction *>(i);
                    }

                } else if (i->isStore()) {
                    auto x = dynamic_cast<StoreInstruction *>(i);
                    auto it = alloca_ids.find(dynamic_cast<AllocaInstruction *>(
                        dynamic_cast<StoreInstruction *>(i)
                            ->getAddr()
                            ->getDef())); // addr
                    if (it != alloca_ids.end()) {
                        if (dynamic_cast<AllocaInstruction *>(
                                dynamic_cast<StoreInstruction *>(i)
                                    ->getAddr()
                                    ->getDef())
                                ->getSymbolEntry()
                                ->isTemporary()) {
                            values[it->second] = x->getSrc();
                            x->getDef()->removeUse(x);
                            bb->remove(i);
                            delete x;
                            i = i->getNext();
                            continue;
                        }
                        auto se = dynamic_cast<IdentifierSymbolEntry *>(
                            (dynamic_cast<AllocaInstruction *>(
                                 dynamic_cast<StoreInstruction *>(i)
                                     ->getAddr()
                                     ->getDef()))
                                ->getSymbolEntry());
                        // auto index = find(params1.begin(), params1.end(),
                        // se); if(index != params1.end()) {
                        //     params1.erase(index);
                        //     params2op[se] = x->getSrc();
                        //     values[it->second] = x->getSrc();
                        //     // std::cout << "  store: " << it->second << " "
                        //     <<
                        //     (dynamic_cast<AllocaInstruction*>(dynamic_cast<StoreInstruction*>(i)->getAddr()->getDef()))
                        //                     // ->getDef()->toStr() << " : "
                        //                     << x->getSrc()->toStr() <<
                        //                     std::endl;
                        //     i = i->getNext();
                        //     continue;
                        // }
                        auto val = x->getSrc();
                        if (paramMap.find(x->getAddr()) != paramMap.end() &&
                            paramMap[x->getAddr()] == nullptr) {
                            // std::cout << x->getSrc()
                            //                  ->getSymbolEntry()
                            //                  ->getType()
                            //                  ->toStr()
                            //           << "\n";
                            auto dst = new Operand(new TemporarySymbolEntry(
                                x->getSrc()->getSymbolEntry()->getType(),
                                SymbolTable::getLabel()));
                            auto zero = new ConstantSymbolEntry(
                                x->getSrc()->getSymbolEntry()->getType(), 0);
                            auto bin = new BinaryInstruction(
                                BinaryInstruction::ADD, dst, x->getSrc(),
                                new Operand(zero));
                            // std::cout<<"param: "<<x->getSrc()->toStr()<<"dst:
                            // "<<dst->toStr()<<"\n";
                            bin->push_back_opse(x->getSrc()->getSymbolEntry());
                            bin->push_back_opse(zero);
                            bb->insertAfter(bin, x);
                            val = dst;
                            paramMap[x->getAddr()] = dst;
                        }

                        // 将对应的alloca的变量/地址, 替换为store的新变量/地址
                        values[it->second] = val;
                        x->getDef()->removeUse(x);
                        bb->remove(i);
                        delete x;
                    }
                } else if (i->isLoad()) {
                    // 找到对应的load使用的alloca变量 ->
                    // 这里是不是也可以找地址呢?
                    LoadInstruction *x = dynamic_cast<LoadInstruction *>(i);
                    auto loadalloc =
                        dynamic_cast<LoadInstruction *>(i)->getAddr()->getDef();
                    auto it = alloca_ids.find(
                        dynamic_cast<AllocaInstruction *>(loadalloc));
                    // 先定义后使用
                    if (it != alloca_ids.end()) {
                        if (dynamic_cast<AllocaInstruction *>(
                                dynamic_cast<LoadInstruction *>(i)
                                    ->getAddr()
                                    ->getDef())
                                ->getSymbolEntry()
                                ->isTemporary()) {
                            x->getAddr()->removeUse(x);
                            x->getDef()->replaceAllUseWith(values[it->second]);
                            bb->remove(i);
                            delete x;
                            i = i->getNext();
                            continue;
                        }

                        auto se = dynamic_cast<IdentifierSymbolEntry *>(
                            (dynamic_cast<AllocaInstruction *>(
                                 dynamic_cast<LoadInstruction *>(i)
                                     ->getAddr()
                                     ->getDef()))
                                ->getSymbolEntry());
                        // auto index = find(params2.begin(), params2.end(),
                        // se); if(index != params2.end()) {
                        //     if(values[it->second] == params2op[se]) {
                        //         values[it->second] = x->getDst();
                        //         // std::cout << "   load: " << it->second <<
                        //         "     " <<
                        //         it->first->getDst()->getSymbolEntry()->toStr()
                        //         << " : " << values[it->second]->toStr() <<
                        //         std::endl; i = i->getNext(); continue;
                        //     }
                        // }

                        // 需要替换所有用到load->def的变量为对应alloca的变量
                        // fix: 这里的addr和operand如何转换
                        x->getAddr()->removeUse(x);
                        // 这里已经进行了use链的重构
                        // std::cout << "replaceAllUseWith to: " << it->second
                        // << "     " << values[it->second]->toStr() <<
                        // std::endl;
                        x->getDef()->replaceAllUseWith(values[it->second]);
                        bb->remove(i);
                        delete x; // 跟i->deleteValue()作用是一样的(但是跟delete
                                  // i不一样)，逻辑上节省了一次dispatch的过程
                    }
                } else if (i->isPhi()) {
                    auto x = dynamic_cast<PhiInstruction *>(i);
                    auto it = phis.find(
                        x); // 也许程序中本来就存在phi，但是phis不一定包含了所有的phi
                    if (it != phis.end()) {
                        // 将对应的alloca变量的operand替换为phi指令的
                        values[it->second] = x->getDef();
                        // std::cout << "   phi: " << it->second << " : " <<
                        // x->getDef()->toStr() << std::endl;
                    }
                }
                i = i->getNext();
                // std::cout << "Final: " <<  values[1]->toStr() << std::endl;
                // //std::cout<<"hhh"<<std::endl;
            }
            // 遍历当前块的所有后续块
            // 找的方式是从当前块的后继块开始找
            // std::cout << "emplace_back: " <<  values[1]->toStr() <<
            // std::endl; std::cout << bb->getNumOfSucc() << std::endl;
            for (auto x = bb->succ_begin(); x != bb->succ_end(); x++) {
                // 比如从21块的后继节点开始找: 22
                if ((*x)) {
                    BasicBlock *succ_bb = *x;
                    worklist2.emplace_back(succ_bb,
                                           values); // 将后继块加入到工作集当中
                    // 在后继块当中找首指令为phi的指令
                    for (auto i = succ_bb->begin(); i != succ_bb->end();
                         i = i->getNext()) {
                        if (i->isPhi()) {
                            auto p = dynamic_cast<PhiInstruction *>(i);
                            auto it = phis.find(p);
                            if (it != phis.end()) {
                                // 找到前驱块里面的最后一个值赋值给phi
                                // 前驱块当中有它本身存在
                                p->addSrc(bb, values[it->second]);
                            }
                        } else {
                            break; // PhiInstruction一定是在指令序列的最前面，所以遇到第一个非PhiInstruction的指令就可以break了
                        }
                    }
                }
            }
        }
    }
    // 0413m
    // 粗略实现: 默认当前块当中的所有phi指令都已经填充完毕了
    // 找到一个块当中的所有phi指令
    printmr("=========mem2reg算法阶段3: 调整phi指令的顺序=========");
    std::list<PhiInstruction *> phi_queue;
    for (auto it = f->begin(); it != f->end(); it++) {
        auto bb = *it;
        phi_queue.clear();
        for (auto instphi = bb->begin(); instphi != bb->end();
             instphi = instphi->getNext()) {
            if (!instphi || !instphi->isPhi())
                break;
            auto phi = dynamic_cast<PhiInstruction *>(instphi);
            std::map<BasicBlock *, Operand *> srcs = phi->getSrcs();
            std::list<PhiInstruction *>::iterator minPos = phi_queue.end(), pos;
            for (auto src : srcs) {
                if (src.second->getDef() && src.second->getDef()->isPhi()) {
                    auto beforephi =
                        dynamic_cast<PhiInstruction *>(src.second->getDef());
                    pos = std::find(phi_queue.begin(), phi_queue.end(),
                                    beforephi);
                    if (std::distance(phi_queue.begin(), pos) <
                        std::distance(phi_queue.begin(), minPos)) {
                        minPos = pos;
                    }
                }
            }
            phi_queue.insert(minPos, phi);
            bb->remove(instphi);
        }
        for (auto phi = phi_queue.rbegin(); phi != phi_queue.rend(); phi++) {
            bb->insertFront(*phi);
        }
    }
}